/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name: super.c
**
*******************************************************************************
******************************************************************************/

#if defined (__FreeBSD__)
#include <gfs_fbsd.h>
#endif

#include "gfs.h"
#include "blklist.h"
/* #include "dcache.h" */
#include "dio.h"
/* #include "dio_linux.h" */
#include "glock.h"
#include "file.h"
#include "fs.h"
/* #include "inode.h" */
#include "locking.h"
#include "super.h"
#if 0
#include <../gfs.h>
#include <super.h>
#include <glock.h>
/* #include <dio.h> */
#endif

/******************************************************************************
*******************************************************************************
**
** sget()
**
** Description:
**   Get the superblock and resource index off of the disk.
**     Grabs memory that must be freed for sdp->sd_rgd.
**
** Input(s):
**   sdp - the filesystem
**   silent - if this is TRUE, don't complain about bad magic numbers
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/
int
sget(gfs_sbd_t *sdp, int silent){
  int blocksize = 0;
  buffer_head_t *bh;
  char *buf, *p;
  int x;


  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering sget()\n"););


  ASSERT(GFS_BASIC_BLOCK <= PAGE_SIZE);
 
#if defined(__linux__)
  blocksize = get_hardblocksize(sdp->sd_vsb->s_dev);
#endif

  if (blocksize != GFS_BASIC_BLOCK)
  {
    IFDEBUG(DEBUG_SUPER3,
	    printk(DEBUG_HOW "GFS:  Changing hardblocksize from %d to %u\n",
		   blocksize, GFS_BASIC_BLOCK););
#if defined(__linux__)
    set_blocksize(sdp->sd_vsb->s_dev, GFS_BASIC_BLOCK);
#endif
  }

  /*  Fill in some fake values to allow gfs_dread() to work  */

  sdp->sd_sb.sb_bsize = GFS_BASIC_BLOCK;
  sdp->sd_fsb2bb_shift = 0;

  bh = gfs_dread(sdp, GFS_SB_ADDR, GFS_SB_LOCK);

  if (!bh)
  {
    printk(KERN_ERR "GFS:  can't read superblock (dread)\n");
    IFDEBUG(DEBUG_SUPER1,
	    printk(DEBUG_HOW "Exiting sget()\n"););
    return(-1);
  }

  ASSERT(sizeof(gfs_sb_t) <= GFS_BASIC_BLOCK);

  gfs_sb_in(&sdp->sd_sb, BH_DATA(bh));

  gfs_dforget(bh);

  if (sdp->sd_sb.sb_magic != GFS_MAGIC_SB)
  {
    if (!silent)
	  /* printk(KERN_ERR "GFS:  bad GFS filesystem on %s\n", kdevname(sdp->sd_vsb->s_dev)); */
		IFDEBUG(DEBUG_SUPER1, 
	    printk(DEBUG_HOW "Exiting sget()\n"););
    return(-1);
  }

  /*  Only allow clients with the same major and minor version number to interact  */

  if ((sdp->sd_sb.sb_version & 0x00FFFF00) != (GFS_VERSION & 0x00FFFF00) || 
      sdp->sd_sb.sb_sbsize != sizeof(gfs_sb_t) ||
      sdp->sd_sb.sb_risize != sizeof(gfs_rindex_t) ||
      sdp->sd_sb.sb_rgsize != sizeof(gfs_rgrp_t) ||
      sdp->sd_sb.sb_disize != sizeof(gfs_dinode_t) ||
      sdp->sd_sb.sb_desize != sizeof(gfs_dirent_t) ||
      sdp->sd_sb.sb_lfsize != sizeof(gfs_leaf_t))
  {
    printk(KERN_ERR "GFS:  code version %d.%d.%d is incompatible with filesystem version %d.%d.%d\n",
	   GFSMajorVersion, GFSMinorVersion, GFSUpdateLevel,
	   (sdp->sd_sb.sb_version >> 16) & 0xFF,
	   (sdp->sd_sb.sb_version >> 8) & 0xFF,
	   (sdp->sd_sb.sb_version) & 0xFF);
    IFDEBUG(DEBUG_SUPER1,
	    printk(DEBUG_HOW "Exiting sget()\n"););
    return(-1);
  }

  if (sdp->sd_sb.sb_bsize > PAGE_SIZE)
  {
    printk(KERN_ERR "GFS:  the blocksize (%u) is too big for the page size (%u)\n",
	   sdp->sd_sb.sb_bsize, PAGE_SIZE);
    return(-1);
  }

  if (sdp->sd_sb.sb_bsize != GFS_BASIC_BLOCK)
  {
    IFDEBUG(DEBUG_SUPER3,
	    printk(DEBUG_HOW "GFS:  Changing hardblocksize from %u to %d\n",
		   GFS_BASIC_BLOCK, sdp->sd_sb.sb_bsize););
#if defined(__linux__)
    set_blocksize(sdp->sd_vsb->s_dev, sdp->sd_sb.sb_bsize);
#endif
  }

  /*  If we already have the resource groups, don't get them again.  */

  if (sdp->sd_rgd)
  {
    IFDEBUG(DEBUG_SUPER1,
	    printk(DEBUG_HOW "Exiting sget()\n"););
    return(0);
  }

  /*  Alloc memory for the Resource index  */

  sdp->sd_rgd = (gfs_rgrpd_t *)gmalloc(sdp->sd_sb.sb_rgrps * sizeof(gfs_rgrpd_t));
  if (!sdp->sd_rgd)
  {
    printk(KERN_ERR "GFS:  out of memory\n");
    return(-1);
  }
  memset(sdp->sd_rgd, 0, sdp->sd_sb.sb_rgrps * sizeof(gfs_rgrpd_t));

  /*  Get temp space for the superblock  */

  buf = (char *)gmalloc(sdp->sd_sb.sb_length * sdp->sd_sb.sb_bsize);
  if (!buf)
  {
    printk(KERN_ERR "GFS:  out of memory\n");
    gfree(sdp->sd_rgd, sdp->sd_sb.sb_rgrps * sizeof(gfs_rgrpd_t));
    return(-ENOMEM);
  }

  /*  Read in the whole superblock  */

  for (x = 0; x < sdp->sd_sb.sb_length; x++)
  {
    bh = gfs_dread(sdp, GFS_SB_ADDR + x, GFS_SB_LOCK);
    if (!bh)
    {
      printk(KERN_ERR "GFS:  error reading extended superblock\n");
      gfree(buf, sdp->sd_sb.sb_length * sdp->sd_sb.sb_bsize);
      gfree(sdp->sd_rgd, sdp->sd_sb.sb_rgrps * sizeof(gfs_rgrpd_t));
      IFDEBUG(DEBUG_SUPER1,
	      printk(DEBUG_HOW "Exiting sget(ERROR)\n"););
      return(-1);
    }
    memcpy(buf + x * sdp->sd_sb.sb_bsize, BH_DATA(bh), sdp->sd_sb.sb_bsize);
    gfs_dforget(bh);
  }

  /*  Copy in the resource index  */

  p = buf + sizeof(gfs_sb_t);
  for (x = 0; x < sdp->sd_sb.sb_rgrps; x++)
  {
    gfs_rindex_in(&sdp->sd_rgd[x].rd_ri, p);
    p += sizeof(gfs_rindex_t);
  }

  /*  Free the temp space  */

  gfree(buf, sdp->sd_sb.sb_length * sdp->sd_sb.sb_bsize);


  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting sget()\n"););

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** sput()
**
** Description:
**   Put the superblock and resource index on to the disk.
**     Since the resource index doesn't change, we only need to 
**     write back the SuperBlock.
**
** Input(s):
**   sdp - the filesystem
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/
int
sput(gfs_sbd_t *sdp){
  buffer_head_t *bh;

  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering sput()\n"););


  sdp->sd_sb.sb_time = CURRENT_TIME;

  bh = gfs_dread(sdp, GFS_SB_ADDR, GFS_SB_LOCK);
  if (!bh)
  {
    printk(KERN_ERR "GFS:  can't read the superblock\n");
    IFDEBUG(DEBUG_SUPER1,
            printk(DEBUG_HOW "Exiting sput()\n"););
    return(-1);
  }

  gfs_sb_out(&sdp->sd_sb, BH_DATA(bh));

  gfs_dwrite(bh, DW_NOKEEP);

  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting sput()\n"););
  
  return(0);
}


/******************************************************************************
*******************************************************************************
**
** mountbit_find()
**
** Description:
**   This routine sets mount IDs in the bitmap entry.
**
** Input(s):
**   sdp - the filesystem
**
** Returns:
**   This machines ID or -1 on failure
**
*******************************************************************************
******************************************************************************/

int
gfs_mountbit_find(gfs_sbd_t *sdp)
{
  int id = -1;
  char byte, mask;
  int i, j;


  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering mountbit_find()\n"););


  for (i = 0; i < GFS_MOUNT_BITS; i++)
  {
    byte = sdp->sd_sb.sb_mountids[i];
      
    mask = 1;

    for (j = 0; j < 8; j++)
    {
      if (!(mask & byte))
      {
	id = 8 * i + j;
	sdp->sd_sb.sb_mountids[i] |= mask;
	break;
      }

      mask <<= 1;
    }    
  
    if (id >= 0)
      break;
  }

  sdp->sd_id = id;


  IFDEBUG(DEBUG_SUPER2,
	  printk(DEBUG_HOW "sdp->sd_id = %d\n", sdp->sd_id););

  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting mountbit_find()\n"););

  return(id);
}


/******************************************************************************
*******************************************************************************
**
** mountbit_value()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int
gfs_mountbit_value(uint8 *bitmap, unsigned int no)
{
  IFDEBUG(DEBUG_FS1,
	  printk(DEBUG_HOW "Doing mountbit_value()\n"););

  return(bitmap[no / 8] & (1 << (no % 8)));
}


/******************************************************************************
*******************************************************************************
**
** mountbit_set()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void
gfs_mountbit_set(uint8 *bitmap, unsigned int no, int value)
{
  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering mountbit_set()\n"););

  if (value)
    bitmap[no / 8] |= 1 << (no % 8);
  else
    bitmap[no / 8] &= ~(1 << (no % 8));

  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting mountbit_set()\n"););
}

#if 0
/******************************************************************************
*******************************************************************************
**
** gfs_inoded()
**
** Description:
**   A kernel thread that frees the GFS inodes that gfs_clear_inode() gives it.
**
** Input(s):
**   data - a pointer to the gfs_sbd_t structure for the filesystem
**
** Returns:
**   an integer that I don't understand
**
*******************************************************************************
******************************************************************************/

static int gfs_inoded(void *data)
{
  gfs_sbd_t *sdp = (gfs_sbd_t *)data;
  gfs_blklist_t *blks;
  gfs_diroff_t *dop;

  gfs_inode_t *ip;
  uint64 ino;
  int error;
  int sync;
  unsigned long flags;


  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Entering gfs_inoded()\n"););

  ASSERT(sdp);


  /*  Set up the thread  */

  lock_kernel();

  exit_mm(current);

  current->session = 1;
  current->pgrp = 1;

  spin_lock_irqsave(&current->sigmask_lock, flags);
  sigfillset(&current->blocked);
  recalc_sigpending(current);
  spin_unlock_irqrestore(&current->sigmask_lock, flags);

  current->fs->umask = 0;
  
  sprintf(current->comm, "gfs_inoded");

  unlock_kernel();


  /*  Wake up the process who started us  */

  GFS_SEMA_UP(&sdp->sd_id_lock);


  while (1)
  {
    GFS_SEMA_DOWN_INTERRUPTIBLE(&sdp->sd_id_work);

    if (sdp->sd_id_stop)
    {
      IFDEBUG(DEBUG_SUPER2,
	      printk(DEBUG_HOW "  Caught signal in gfs_inoded\n"););
      break;
    }


    IFDEBUG(DEBUG_SUPER1,
	    printk(DEBUG_HOW "Entering gfs_inoded() core\n"););

    /*  We have to worry about Glock/sd_id_mutex deadlock here.
	We can't grab a Glock while holding the sd_id_mutex, so:
	get the mutex, find an inode to release, release the 
	mutex, grab the appropriate Glock, and then get the mutex
	again.  */

    GFS_SEMA_DOWN(&sdp->sd_id_mutex);

    if (gfs_list_empty(&sdp->sd_id_list))
    {
      printk(KERN_ERR "GFS:  gfs_inoded woken up for nothing\n");
      GFS_SEMA_UP(&sdp->sd_id_mutex);
      continue;
    }
    ip = gfs_list_entry(sdp->sd_id_list.prev, gfs_inode_t, i_list);
    ino = ip->i_num;

    GFS_SEMA_UP(&sdp->sd_id_mutex);


    blks = NULL;
    sync = 0;


    /*  Grab the Glock for that inode  */

    error = gfs_glock(sdp, gfs_ino2gl(sdp, ino), NULL, 0);
    if (error)
    {
      printk(KERN_ERR "GFS: can't get a lock in gfs_inoded\n");
      ASSERT(0);  /*  What is the right thing to do here?  */
    }


    GFS_SEMA_DOWN(&sdp->sd_id_mutex);

    ip = gfs_list_entry(sdp->sd_id_list.prev, gfs_inode_t, i_list);

    ASSERT(ip->i_num == ino);


    /*  Take the inode out of the list  */

    gfs_list_del(&ip->i_list);

    /*  We need to do our own cache update here  */

    if (!gfs_buffer_uptodate(ip->i_didata))
    {
      gfs_dreread(sdp, 1, &ip->i_didata, TRUE);
      gfs_dinode_in(&ip->i_di, BH_DATA(ip->i_didata));
    }

    /*  Write back the dinode  */

    IFDEBUG(DEBUG_SUPER3,
	    int x;
	    printk(DEBUG_HOW "gfs_inoded(%"FMT_64"u)\n", ip->i_num);
	    printk(DEBUG_HOW "Before:  ip->i_di.di_nopen = %d\n", ip->i_di.di_nopen);
	    printk(DEBUG_HOW "Before:  ip->i_di.di_open = ");
	    for (x = 0; x < GFS_MOUNT_BITS; x++)
	      printk("%.2X ", ip->i_di.di_open[x]);
	    printk("\n");); 

    ip->i_di.di_nopen--;

    ASSERT(mountbit_value(ip->i_di.di_open, sdp->sd_id));
    mountbit_set(ip->i_di.di_open, sdp->sd_id, 0);

    IFDEBUG(DEBUG_SUPER3,
	    int x;
	    printk(DEBUG_HOW "After:  ip->i_di.di_nopen = %d\n", ip->i_di.di_nopen);
	    printk(DEBUG_HOW "After:  ip->i_di.di_open = ");
	    for (x = 0; x < GFS_MOUNT_BITS; x++)
	      printk("%.2X ", ip->i_di.di_open[x]);
	    printk("\n"););

    if (ip->i_di.di_nopen == 0 && ip->i_di.di_nlink == 0)
    {
      blks = gfs_difree(ip);

      /*  Zero out the dinode magic number  */

      ip->i_di.di_magic = 0;
      gfs_dinode_out(&ip->i_di, BH_DATA(ip->i_didata));
      gfs_dwrite(ip->i_didata, DW_NOKEEP);

      sync = GL_SYNC;
    }
    else
    {
      gfs_dinode_out(&ip->i_di, BH_DATA(ip->i_didata));
      gfs_dwrite(ip->i_didata, DW_NOKEEP);
    }

    sdp->sd_id_count--;

    GFS_SEMA_UP(&sdp->sd_id_mutex);

    gfs_gunlock(sdp, gfs_ino2gl(sdp, ino), sync);


    /*  Free any blocks from gfs_difree()  */

    if (blks)
    {
      gfs_blkfree(sdp, &blks);
      if (blks)
      {
	GARBAGE_BLOCKS();
	gfs_blklist_free(blks);
      }
    }

    /*  Free up directory offsets  */

    while (ip->i_diroff)
    {
      dop = ip->i_diroff;
      ip->i_diroff = dop->do_next;
      gfree(dop, sizeof(gfs_diroff_t));
    }

    gfree(ip, sizeof(gfs_inode_t));    

    IFDEBUG(DEBUG_SUPER1,
	    printk(DEBUG_HOW "Exiting gfs_inoded() core\n"););
  }


  GFS_SEMA_UP(&sdp->sd_id_lock);

  IFDEBUG(DEBUG_DIO1,
	  printk(DEBUG_HOW "Exiting gfs_inoded()\n"););

  return(0);
}
#endif

/******************************************************************************
*******************************************************************************
**
** clear_out_glock()
**
** Description:
**   Remove all the gfs_bhdata_t structures associated with a Glock
**
** Input(s):
**   gl - the Glock structure to be cleaned
**
*******************************************************************************
******************************************************************************/

void 
gfs_clear_out_glock(gfs_glock_t *gl)
{
  gfs_list_head_t *head, *next;
  gfs_bhdata_t *bd;


  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering clear_out_glock()\n"););


  GFS_SEMA_DOWN(&gl->gl_buflock);

  head = &(gl->gl_bufs);

  next = head->next;
  while (1)
  {
    gfs_list_head_t * tmp = next;

    next = next->next;
    if (tmp == head)
      break;
    bd = gfs_list_entry(tmp, gfs_bhdata_t, bd_list);

    gfs_list_del(&bd->bd_list);

#if defined(__linux__)
    if (bd->bd_bh)
    {
      bd->bd_bh->b_rm_fn = NULL;
      bd->bd_bh->b_pdata = NULL;
      mark_buffer_uptodate(bd->bd_bh, FALSE);
    }
#endif

    gfree(bd, sizeof(gfs_bhdata_t));
  }
    
  GFS_SEMA_UP(&gl->gl_buflock);


  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting clear_out_glock()\n"););
}


/******************************************************************************
*******************************************************************************
**
** clear_out_gla()
**
** Description:
**   Free all the Glock structures and their buffer head structures.
**
** Input(s):
**   sdp - the superblock data for this filesystem
**
*******************************************************************************
******************************************************************************/

void
gfs_clear_out_gla(gfs_sbd_t *sdp)
{
  int x;
  gfs_list_head_t *head, *next;
  gfs_glock_t *gl;


  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering clear_out_gla()\n"););


  for (x = 0; x < GFS_GHASH_SIZE; x++)
  {
    GFS_SEMA_DOWN(&(sdp->sd_gla_lock[x]));

    head = &(sdp->sd_gla[x]);

    next = head->next;
    while (1)
    {
      gfs_list_head_t *tmp = next;

      next = next->next;
      if (tmp == head)
	break;
      gl = gfs_list_entry(tmp, gfs_glock_t, gl_next);

      gfs_list_del(&gl->gl_next);

      ASSERT(gl->gl_count == 0);

      if (gl->gl_held)
      {
	gl->gl_release = CB_NEED_E;
	gfs_demote_glock(sdp, gl);
      }
      if (gl->gl_lockdata)
	sdp->sd_lockstruct->release(sdp->sd_lockdata, &gl->gl_lockdata);

      gfs_clear_out_glock(gl);

      gfree(gl, sizeof(gfs_glock_t));
    }

    GFS_SEMA_UP(&(sdp->sd_gla_lock[x]));
  }


  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting clear_out_gla()\n"););
}

#if 0
/******************************************************************************
*******************************************************************************
**
** gfs_read_super()
**
** Description:
**   Read in the super block, root inode, and generaly-all-around mount the
**     filesystem.
**   Note that this implementation updates the root dinode on the disk before
**     updating the superblock.
**
** Input(s):
**   sb - the VFS super block
**   data - the option data passed in with the mount -o flag
**   silent - if TRUE, don't complain if this isn't a GFS filesystem
**
** Returns:
**   The superblock, or NULL on failure
**
*******************************************************************************
******************************************************************************/

struct super_block *gfs_read_super(struct super_block *sb, void *data, int silent)
{
  gfs_sbd_t *sdp;
  int x, error;
  gfs_dcached_t *dcd;


  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering gfs_read_super()\n"););


  MOD_INC_USE_COUNT;



  /*  Make sure there are no dirty buffers lying around  */

  shrink_dcache_sb(sb);
  sync_dev(sb->s_dev);
  invalidate_buffers(sb->s_dev);


  lock_super(sb);

  sdp = (gfs_sbd_t *)gmalloc(sizeof(gfs_sbd_t));
  if (!sdp)
  {
    printk(KERN_ERR "GFS:  out of memory\n");
    goto fail;
  }
  memset(sdp, 0, sizeof(gfs_sbd_t));

  vsb2sbd(sb) = sdp;
  sdp->sd_vsb = sb;


  /*  Init semaphores  */

  for (x = 0; x < GFS_GHASH_SIZE; x++)
  {
    INIT_GFS_LIST_HEAD(&sdp->sd_gla[x]);
    GFS_SEMA_INIT(&sdp->sd_gla_lock[x]);
  }
  INIT_GFS_LIST_HEAD(&sdp->sd_gl_lru);
  GFS_SEMA_INIT(&sdp->sd_gl_lru_lock);

  for (x = 0; x < GFS_IHASH_SIZE; x++)
  {
    INIT_GFS_LIST_HEAD(&sdp->sd_ihash[x]);
    GFS_SEMA_INIT(&sdp->sd_ihash_lock[x]);
  }
  

  if (gfs_mount_lockproto(sdp, (char *)data))
    goto fail_sbd_free;


  error = gfs_glock(sdp, GFS_SB_LOCK, NULL, 0);
  if (error < 0)
  {
    printk(KERN_ERR "GFS:  can't get root glock\n");
    goto fail_lock_unmount;
  }


  if (sget(sdp, silent) < 0)
    goto fail_gla_free;


  /*  Set mount information  */

  if (mountbit_find(sdp) < 0)
  {
    printk(KERN_ERR "GFS:  too many mounts on this filesystem\n");
    goto fail_ri_free;
  }

  sdp->sd_sb.sb_mounts++;
  sdp->sd_mrg = sdp->sd_id % sdp->sd_sb.sb_rgrps;
  

#if defined(__linux__)
  sdp->sd_fsb2bb_shift = 0;
#endif  /*  linux  */
#if defined(__sgi__) || defined(__FreeBSD__)
  sdp->sd_fsb2bb_shift = ffs(sdp->sd_sb.sb_bsize / GFS_BASIC_BLOCK) - 1;
#endif  /*  sgi  */
  sdp->sd_diptrs = (sdp->sd_sb.sb_bsize - sizeof(gfs_dinode_t)) / sizeof(uint64);
  sdp->sd_inptrs = sdp->sd_sb.sb_bsize / sizeof(uint64);


  /*  Setup up Virtual Super Block  */

  sb->s_blocksize = sdp->sd_sb.sb_bsize;
  sb->s_blocksize_bits = ffs(sb->s_blocksize) - 1;
  sb->s_magic = GFS_SUPER_MAGIC;
  sb->s_op = &gfs_sops;

  /* uses gmalloc, need to be initialized for leaf_merging code (not active yet) */
  /*
  if (gfs_init_tunevals(sdp, 1))
    goto fail_ri_free;
  */

  /*  Get the root inode  */

  sdp->sd_rooti = iget(sb, sdp->sd_sb.sb_rootdi);
  if (!sdp->sd_rooti)
  {
    printk(KERN_ERR "GFS:  can't get root inode\n");
    goto fail_tune_free;
  }

  sb->s_root = d_alloc_root(sdp->sd_rooti, NULL);
  if (!sb->s_root)
  {
    printk(KERN_ERR "GFS:  can't get root dentry\n");
    goto fail_iput;
  }

  dcd = (gfs_dcached_t *)gmalloc(sizeof(gfs_dcached_t));
  if (!dcd)
    goto fail_dput;
    
  dcd->dc_vn = vi2ino(sdp->sd_rooti)->i_vn;
  dcd->dc_time = CURRENT_TIME;
  
  vdc2dcd(sb->s_root) = dcd;

  sb->s_root->d_op = &gfs_dops;


  /*  Start up the buffer thread  */

  GFS_SEMA_INIT_LKD(&sdp->sd_bufd_lock);

  if (kernel_thread(gfs_bufd, sdp, 0) < 0)
    goto fail_dput;

  GFS_SEMA_DOWN(&sdp->sd_bufd_lock);


  /*  Start up the inode thread  */

  GFS_SEMA_INIT(&sdp->sd_id_mutex);
  INIT_GFS_LIST_HEAD(&sdp->sd_id_list);
  GFS_SEMA_INIT_LKD(&sdp->sd_id_work);
  GFS_SEMA_INIT_LKD(&sdp->sd_id_lock);

  if (kernel_thread(gfs_inoded, sdp, 0) < 0)
    goto fail_bufd;

  GFS_SEMA_DOWN(&sdp->sd_id_lock);


  /*  Write back the super block  */
  
  if (sput(sdp) < 0)
    goto fail_inoded;

  gfs_gunlock(sdp, GFS_SB_LOCK, 0);

  unlock_super(sb);


  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting gfs_read_super()\n"););

  return(sb);



 fail_inoded:
  sdp->sd_id_stop = 1;
  GFS_SEMA_UP(&sdp->sd_id_work);
  GFS_SEMA_DOWN(&sdp->sd_id_lock);

 fail_bufd:
  sdp->sd_bufd_stop = 1;
  GFS_SEMA_DOWN(&sdp->sd_bufd_lock);

 fail_dput:
  dput(sb->s_root);
  goto fail_ri_free;

 fail_iput:
  iput(sdp->sd_rooti);

 fail_tune_free:
  gfree(sdp->tune, sizeof(gfs_tune_t));

 fail_ri_free:
  gfree(sdp->sd_rgd, sdp->sd_sb.sb_rgrps * sizeof(gfs_rgrpd_t));

 fail_gla_free:
  gfs_gunlock(sdp, GFS_SB_LOCK, 0);
  clear_out_gla(sdp);

 fail_lock_unmount:
  gfs_unmount_lockproto(sdp); 

 fail_sbd_free:
  gfree(sdp, sizeof(gfs_sbd_t));

 fail:
  unlock_super(sb);
  MOD_DEC_USE_COUNT;
  sb->s_dev = NODEV;

  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting gfs_read_super()\n"););

  return(NULL);
}
#endif

#if 0
/******************************************************************************
*******************************************************************************
**
** gfs_read_inode()
**
** Description:
**   Read in a inode off the disk
**
** Input(s):
**   inode - the inode struct (use the i_ino field to get the inode number)
**
*******************************************************************************
******************************************************************************/

void gfs_read_inode(struct inode *inode)
{
  gfs_inode_t *ip;
  gfs_sbd_t *sdp = vsb2sbd(inode->i_sb);
  uint64 glock;
  int error;


  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering gfs_read_inode()\n"););


  ip = (gfs_inode_t *)gmalloc(sizeof(gfs_inode_t));
  if (!ip)
  {
    printk(KERN_ERR "GFS:  out of memory\n");
    goto fail;
  }
  memset(ip, 0, sizeof(gfs_inode_t));

  vi2ino(inode) = ip;
  ip->i_vino = inode;

  ip->i_sbd = sdp;
  ip->i_num = inode->i_ino;
  GFS_SEMA_INIT(&ip->i_sema);
  ip->i_diroff = NULL;

  /*  Aquire glock  */

  glock = gfs_ino2gl(sdp, ip->i_num);

  if ((error = gfs_glock(sdp, glock, NULL, 0)) < 0)
    goto fail_ino_free;

  ASSERT(gfs_ihashlookup(sdp, ip->i_num) == NULL);

  /*  Read in dinode  */

  ip->i_didata = gfs_dread(sdp, ip->i_num, glock);
  if (!ip->i_didata)
    goto fail_gunlock;

  gfs_dinode_in(&ip->i_di, BH_DATA(ip->i_didata));

  if (ip->i_di.di_magic != GFS_MAGIC_DI)
  {
    printk("GFS:  ip->i_di.di_magic (%"FMT_64"u) != GFS_MAGIC_DI\n",
	   ip->i_num);
    {
      uint32 x;
      for (x = 0; x < sdp->sd_sb.sb_bsize; x++)
      {
	printk("%.2X ", ((unsigned char *)(BH_DATA(ip->i_didata)))[x]);
	if (x % 16 == 15)
	  printk("\n");
      }
      if (x % 16 != 0)
	printk("\n");
    }
    ASSERT(0);
  }
  ASSERT(ip->i_di.di_addr == ip->i_num);


  IFDEBUG(DEBUG_SUPER3,
	  int x;
	  printk(DEBUG_HOW "gfs_read_inode(%"FMT_64"u)\n", ip->i_num);
	  printk(DEBUG_HOW "Before:  ip->i_di.di_nopen = %d\n", ip->i_di.di_nopen);
	  printk(DEBUG_HOW "Before:  ip->i_di.di_open = ");
	  for (x = 0; x < GFS_MOUNT_BITS; x++)
	    printk("%.2X ", ip->i_di.di_open[x]);
	  printk("\n"););

  ip->i_di.di_nopen++;

  ASSERT(!mountbit_value(ip->i_di.di_open, sdp->sd_id));
  mountbit_set(ip->i_di.di_open, sdp->sd_id, 1);

  IFDEBUG(DEBUG_SUPER3,
	  int x;
	  printk(DEBUG_HOW "After:  ip->i_di.di_nopen = %d\n", ip->i_di.di_nopen);
	  printk(DEBUG_HOW "After:  ip->i_di.di_open = ");
	  for (x = 0; x < GFS_MOUNT_BITS; x++)
	    printk("%.2X ", ip->i_di.di_open[x]);
	  printk("\n"););


  /*  Write dinode back out  */

  gfs_dinode_out(&ip->i_di, BH_DATA(ip->i_didata));
  gfs_dwrite(ip->i_didata, 0);

  gfs_inode_attr_in(ip);

  gfs_ihashadd(ip);

  gfs_gunlock(sdp, glock, 0);


  if (S_ISREG(inode->i_mode))
    inode->i_op = &gfs_file_iops;
  else if (S_ISDIR(inode->i_mode))
    inode->i_op = &gfs_dir_iops;
  else if (S_ISLNK(inode->i_mode))
    inode->i_op = &gfs_symlink_iops;
  else if (S_ISBLK(inode->i_mode))
    inode->i_op = &gfs_blkdev_inode_operations;
  else if (S_ISCHR(inode->i_mode))
    inode->i_op = &gfs_chrdev_inode_operations;
/*
  else if (S_ISFIFO(inode->i_mode))
    init_fifo(inode);
*/
  else
    panic("GFS:  unknown file type:  %d\n", inode->i_mode);


  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting gfs_read_inode()\n"););

  return;


 fail_gunlock:
  gfs_gunlock(sdp, glock, 0);

 fail_ino_free:
  gfree(ip, sizeof(gfs_inode_t));

 fail:
  make_bad_inode(inode);

  printk("GFS:  making bad inode\n");

  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting gfs_read_inode()\n"););
}
#endif

/******************************************************************************
*******************************************************************************
**
** gfs_put_inode()
**
** Description:
**   Stole a great idea from NFS to keep open inodes to a minimum.
**
** Input(s):
**   inode - the inode to be put
**
*******************************************************************************
******************************************************************************/

void gfs_put_inode(struct inode *inode)
{
  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering gfs_put_inode()\n"););
#if 0
  if (inode->i_count == 1)
    inode->i_nlink = 0;
#endif

  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting gfs_put_inode()\n"););
}

#if 0
/******************************************************************************
*******************************************************************************
**
** gfs_notify_change()
**
** Description:
**   The VFS layer wants to change one or more of an inodes attributes.  Write
**   that change out to disk.
**
** Input(s):
**   dentry - the inode's dentry
**   attr - the attributes to change
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_notify_change(struct dentry *dentry, struct iattr *attr)
{
  gfs_inode_t *ip = vi2ino(dentry->d_inode);
  gfs_sbd_t *sdp = ip->i_sbd;
  int error = 0;
  gfs_blklist_t *blks = NULL;
  int sync = 0;


  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering gfs_notify_change()\n"););


  error = inode_change_ok(dentry->d_inode, attr);
  if (!error)
  {
    if ((error = gfs_glock(sdp, ip->i_di.di_glock, ip, 0)) < 0)
    {
      IFDEBUG(DEBUG_SUPER1,
	      printk(DEBUG_HOW "Exiting gfs_notify_change()\n"););
      return(error);
    }

    /*  Linux handles truncates in a way that is really bad for GFS.
	So, we implement truncate here.  */

    if (attr->ia_valid & ATTR_SIZE)
    {
      if (ip->i_di.di_type != FILE_REG || ip->i_di.di_nopen != 1)
      {
	gfs_gunlock(sdp, ip->i_di.di_glock, 0);
	IFDEBUG(DEBUG_SUPER1,
		printk(DEBUG_HOW "Exiting gfs_notify_change()\n"););
	return(-EPERM);
      }

      if (attr->ia_size != ip->i_di.di_bsize)
      {
	error = gfs_truncate(ip, attr->ia_size, &blks);
	sync = GL_SYNC;
      }
    }

    if (!error)
    {
      inode_setattr(dentry->d_inode, attr);
      gfs_inode_attr_out(ip);
    }

    gfs_dinode_out(&ip->i_di, BH_DATA(ip->i_didata));
    gfs_dwrite(ip->i_didata, 0);
    
    gfs_gunlock(sdp, ip->i_di.di_glock, sync);

    if (blks)
    {
      gfs_blkfree(sdp, &blks);
      if (blks)
      {
	GARBAGE_BLOCKS();
	gfs_blklist_free(blks);
      }
    }
  }


  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting gfs_notify_change()\n"););

  return(error);
}
#endif

#if 0
/******************************************************************************
*******************************************************************************
**
** gfs_statfs()
**
** Description:
**   Gather and return stats about the filesystem
**
** Input(s):
**   sb - the filesystem's superblock
**   statfsbuf - the buffer to return the data in
**   size - the size of the buffer
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_statfs(struct super_block *sb, struct statfs *statfsbuf, int size)
{
  gfs_sbd_t *sdp;
  gfs_rgrpd_t *rgd;
  struct statfs sfs;

  int i;

  uint64 blocks = 0;
  uint64 dinodes = 0;
  uint64 free = 0;
  int error;


  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering gfs_statfs()\n"););


  if (!sb || !(sdp = vsb2sbd(sb)))
    return(-EIO);

  memset(&sfs, 0, sizeof(struct statfs));

  for (i = 0; i < sdp->sd_sb.sb_rgrps; i++)
  {
    rgd = &sdp->sd_rgd[i];

    error = gfs_glock(sdp, rgd->rd_ri.ri_glock, NULL, GL_SHARED);
    if (error)
    {
      IFDEBUG(DEBUG_SUPER1,
	      printk(DEBUG_HOW "Exiting gfs_statfs()\n"););
      return(error);      
    }

    /*  Read/Reread the RG block  */

    error = gfs_rgrp_read(sdp, rgd);
    if (error)
    {
      gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);
      IFDEBUG(DEBUG_SUPER1,
	      printk(DEBUG_HOW "Exiting gfs_statfs()\n"););
      return(error);
    }
      
    IFDEBUG(DEBUG_SUPER2,
	    printk(DEBUG_HOW "%.4u:  %"FMT_64"u\t%"FMT_64"u\t%"FMT_64"u\n",
		   i,
		   rgd->rd_rg.rg_dinodes,
		   rgd->rd_ri.ri_data - rgd->rd_rg.rg_free,
		   rgd->rd_ri.ri_data););

    blocks += rgd->rd_ri.ri_data;
    dinodes += rgd->rd_rg.rg_dinodes;
    free += rgd->rd_rg.rg_free;

    gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);
  }


  sfs.f_type = GFS_SUPER_MAGIC;
  sfs.f_bsize = sdp->sd_sb.sb_bsize;
  sfs.f_blocks = blocks;
  sfs.f_bfree = free;
  sfs.f_bavail = free;
  sfs.f_files = dinodes;
  sfs.f_ffree = free;
  sfs.f_namelen = GFS_FNAMESIZE;


  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting gfs_statfs()\n"););

  return(copy_to_user(statfsbuf, &sfs, (size < sizeof(struct statfs)) ? size : sizeof(struct statfs)));
}
#endif

#if 0
/******************************************************************************
*******************************************************************************
**
** gfs_remount_fs()
**
** Description:
**   Remount the filesystem
**
** Input(s):
**   sb - the filesystem's superblock
**   flags - some flags
**   options - some options
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int gfs_remount_fs(struct super_block *sb, int *flags, char *options)
{
  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering gfs_remount_fs()\n"););

  printk(KERN_ERR "GFS:  Hoser!!!  Somebody called remountfs()\n");

  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting gfs_remount_fs()\n"););

  return(-ENOSYS);
}
#endif

#if 0
/******************************************************************************
*******************************************************************************
**
** gfs_clear_inode()
**
** Description:
**   Deallocate a inode when the VFS is done with it.
**
**   gfs_clear_inode() can't acquire any Glocks.  If it does, deadlock can
**   occur.  Our solution is to pass the inodes to the gfs_inoded thread and
**   it releases them.
**
** Input(s):
**   inode - the inode to be freed
**
*******************************************************************************
******************************************************************************/

void gfs_clear_inode(struct inode *inode)
{
  gfs_inode_t *ip;
  gfs_sbd_t *sdp;


  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering gfs_clear_inode()\n"););


  if (!inode || !(ip = vi2ino(inode)))
  {
    IFDEBUG(DEBUG_SUPER1,
	    printk(DEBUG_HOW "Exiting gfs_clear_inode()\n"););
    return;
  }

  if (!(sdp = vsb2sbd(inode->i_sb)))
  {
    printk(KERN_ERR "GFS:  No superblock when clearing inode %"FMT_64"u  (Damn Crud Puppies)\n",
	   ip->i_num);
    goto cleanup;
  }


  gfs_ihashdel(ip);

  GFS_SEMA_DOWN(&sdp->sd_id_mutex);

  gfs_list_add(&ip->i_list, &sdp->sd_id_list);

  sdp->sd_id_count++;

  GFS_SEMA_UP(&sdp->sd_id_mutex);


  GFS_SEMA_UP(&sdp->sd_id_work);


 cleanup:
  vi2ino(inode) = NULL;

  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting gfs_clear_inode()\n"););
}
#endif 

#if 0
/******************************************************************************
*******************************************************************************
**
** gfs_put_super()
**
** Description:
**   Unmount the filesystem.
**
** Input(s):
**   sb - the filesystem's superblock
**
*******************************************************************************
******************************************************************************/

void gfs_put_super(struct super_block *sb)
{
  gfs_rgrpd_t *rgd;
  gfs_sbd_t *sdp = vsb2sbd(sb);
  int x, y, error;
  gfs_inode_t *ip;
  buffer_head_t *bh;



  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Entering gfs_put_super()\n"););


  do
  {
    for (x = 0; x < GFS_IHASH_SIZE; x++)
    {
      GFS_SEMA_DOWN(&sdp->sd_ihash_lock[x]);

      if (!gfs_list_empty(&sdp->sd_ihash[x]))
	ip = gfs_list_entry(sdp->sd_ihash[x].next, gfs_inode_t, i_list);
      else 
	ip = NULL;

      GFS_SEMA_UP(&sdp->sd_ihash_lock[x]);

      if (ip)
	break;
    }

    if (ip)
    {
      IFDEBUG(DEBUG_SUPER2,
	      printk(DEBUG_HOW "  Calling gfs_clear_inode() on %"FMT_64"u\n", ip->i_num););
      gfs_clear_inode(ip->i_vino);
    }
  } while (ip);


  /*  Wait for all the gfs_clear_inode()s to take effect  */

  while (1)
  {
    GFS_SEMA_DOWN(&sdp->sd_id_mutex);

    if (gfs_list_empty(&sdp->sd_id_list))
    {
      GFS_SEMA_UP(&sdp->sd_id_mutex);
      break;
    }

    GFS_SEMA_UP(&sdp->sd_id_mutex);

    schedule();
  }


  /*  Kill off the inode thread  */

  IFDEBUG(DEBUG_SUPER2,
	  printk(DEBUG_HOW "  killing gfs_inoded...\n"););

  sdp->sd_id_stop = 1;
  GFS_SEMA_UP(&sdp->sd_id_work);
  GFS_SEMA_DOWN(&sdp->sd_id_lock);


  /*  Kill off the buffer thread  */

  IFDEBUG(DEBUG_SUPER2,
	  printk(DEBUG_HOW "  killing gfs_bufd...\n"););

  sdp->sd_bufd_stop = 1;
  GFS_SEMA_DOWN(&sdp->sd_bufd_lock);


  /*  Write back SuperBlock  */

  error = gfs_glock(sdp, GFS_SB_LOCK, NULL, 0);
  ASSERT(error == 0);

  bh = gfs_dread(sdp, GFS_SB_ADDR >> sdp->sd_fsb2bb_shift, GFS_SB_LOCK);
  ASSERT(bh);
  gfs_sb_in(&sdp->sd_sb, BH_DATA(bh));

  sdp->sd_sb.sb_time = CURRENT_TIME;
  sdp->sd_sb.sb_mounts--;

  ASSERT(mountbit_value(sdp->sd_sb.sb_mountids, sdp->sd_id));
  mountbit_set(sdp->sd_sb.sb_mountids, sdp->sd_id, 0);

  gfs_sb_out(&sdp->sd_sb, BH_DATA(bh));
  gfs_dwrite(bh, DW_NOKEEP);

  gfs_gunlock(sdp, GFS_SB_LOCK, 0);


  /*  Take apart glock structures and buffer lists  */

  clear_out_gla(sdp);

  /*  Unmount the locking protocol  */

  gfs_unmount_lockproto(sdp);


  /*  Get rid of rgrp blocks  */

  for (x = 0; x < sdp->sd_sb.sb_rgrps; x++)
  {
    rgd = &sdp->sd_rgd[x];
    if (rgd->rd_bh)
    {
      for (y = 0; y < rgd->rd_ri.ri_length; y++)
	gfs_dforget(rgd->rd_bh[y]);
      gfree(rgd->rd_bh, rgd->rd_ri.ri_length * sizeof(buffer_head_t *));
      gfree(rgd->rd_bits, rgd->rd_ri.ri_length * sizeof(gfs_bitmap_t));
    }
  }


  /*  Free More Stuff  */


  /* uncomment next line when gfs_init_tunevals is uncommented in read_super */
  /* gfree(sdp->tune, sizeof(gfs_tune_t)); */
  gfree(sdp->sd_rgd, sdp->sd_sb.sb_rgrps * sizeof(gfs_rgrpd_t));
  gfree(sdp, sizeof(gfs_sbd_t));

  vsb2sbd(sb) = NULL;
  
  MOD_DEC_USE_COUNT; 

  IFDEBUG(DEBUG_SUPER1,
	  printk(DEBUG_HOW "Exiting gfs_put_super()\n"););
}
#endif



#if 0
struct super_operations gfs_sops =
{
  gfs_read_inode,
  NULL,  /*  write_inode  */
  gfs_put_inode,
  NULL,  /*  delete_inode  */
  gfs_notify_change,
  gfs_put_super,
  NULL,  /*  write_super  */
  gfs_statfs,
  gfs_remount_fs,
  gfs_clear_inode,
  NULL  /*  umount_begin  */
};
#endif
