/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name: dcache.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "blklist.h"
#include "dcache.h"
#include "dir.h"
#include "glock.h"


/******************************************************************************
*******************************************************************************
**
** gfs_drevalidate()
**
** Description:
**   Check to make sure the lookup necessary to arrive at this inode from its
**   parent is still good.
**
** Input(s):
**   The dentry to lookup
**
** Returns:
**   1 if the dentry is ok, 0 if it isn't
**
*******************************************************************************
******************************************************************************/

int gfs_drevalidate(struct dentry *dentry, int flags)
{
  gfs_inode_t *dip;
  gfs_dcached_t *dcd = vdc2dcd(dentry);
  gfs_sbd_t *sdp;

  uint64 ino;
  int error;


  IFDEBUG(DEBUG_DCACHE1,
	  printk(DEBUG_HOW "Entering gfs_drevalidate()\n"););


  if (IS_ROOT(dentry))
    goto out_valid;

  if (!dentry->d_inode || is_bad_inode(dentry->d_inode))
    goto out_invalid;


  dip = vi2ino(dentry->d_parent->d_inode);
  sdp = dip->i_sbd;

  error = gfs_glock(sdp, dip->i_di.di_glock, dip, GL_SHARED);
  if (error) 
  {
    IFDEBUG(DEBUG_DCACHE1,
	    printk(DEBUG_HOW "Exiting gfs_drevalidate()\n"););
    return(error);
  }

  if (dip->i_vn == dcd->dc_vn &&
      CURRENT_TIME - dcd->dc_time < GFS_GLOCK_ROLLOVER)
  {
    gfs_gunlock(sdp, dip->i_di.di_glock, 0);
    dcd->dc_vn = dip->i_vn;
    dcd->dc_time = CURRENT_TIME;
    goto out_valid;
  }

  error = gfs_dir_search(dip, &dentry->d_name, &ino);
  
  gfs_gunlock(sdp, dip->i_di.di_glock, 0);

  if (error)
    goto out_invalid;

  if (ino == vi2ino(dentry->d_inode)->i_num)
  {
    dcd->dc_vn = dip->i_vn;
    dcd->dc_time = CURRENT_TIME;
    goto out_valid;
  }


 out_invalid:
  IFDEBUG(DEBUG_DCACHE1,
	  printk(DEBUG_HOW "Exiting gfs_drevalidate()\n"););
  return(0);

 out_valid:
  IFDEBUG(DEBUG_DCACHE1,
	  printk(DEBUG_HOW "Exiting gfs_drevalidate()\n"););
  return(1);
}


/******************************************************************************
*******************************************************************************
**
** gfs_drelease()
**
** Description:
**   Free the GFS specific data from a dentry about to be freed.
**
** Input(s):
**   The dentry to be freed.
**
*******************************************************************************
******************************************************************************/

void gfs_drelease(struct dentry *dentry)
{
  IFDEBUG(DEBUG_DCACHE1,
	  printk(DEBUG_HOW "Entering gfs_drelease()\n"););

  if (vdc2dcd(dentry))
    gfree(vdc2dcd(dentry), sizeof(gfs_dcached_t));
  else
    printk(KERN_ERR "GFS:  dentry with NULL d_fsdata\n");

  IFDEBUG(DEBUG_DCACHE1,
	  printk(DEBUG_HOW "Exiting gfs_drelease()\n"););
}


struct dentry_operations gfs_dops = 
{
  gfs_drevalidate,
  NULL,  /*  d_hash  */
  NULL,  /*  d_compare  */
  NULL,  /*  d_delete  */
  gfs_drelease,
  NULL   /*  d_iput  */
};
