/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name: inode.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "blklist.h"
#include "bmap.h"
#include "dcache.h"
#include "dio.h"
#include "dir.h"
#include "glock.h"
#include "file_linux.h"
#include "fs.h"
#include "inode.h"





/******************************************************************************
*******************************************************************************
**
** gfs_inode_attr_in()
**
** Description:
**   Copy attributes from the dinode into the VFS inode
**
** Input(s):
**   ip - the inode
**
*******************************************************************************
******************************************************************************/

void gfs_inode_attr_in(gfs_inode_t *ip)
{
  struct inode *ino = ip->i_vino;

  ENTER(GFS_INODE_ATTR_IN, DEBUG_INODE1); 
 
  ino->i_mode = ip->i_di.di_mode & S_IALLUGO;

  switch (ip->i_di.di_type)
  {
  case FILE_REG:
    ino->i_mode |= S_IFREG;
    break;
  case FILE_DIR:
    ino->i_mode |= S_IFDIR;
    break;
  case FILE_LNK:
    ino->i_mode |= S_IFLNK;
    break;
  case FILE_BLK:
    ino->i_mode |= S_IFBLK;
    ino->i_rdev = MKDEV(ip->i_di.di_major, ip->i_di.di_minor);
    break;
  case FILE_CHR:
    ino->i_mode |= S_IFCHR;
    ino->i_rdev = MKDEV(ip->i_di.di_major, ip->i_di.di_minor);
    break;
  case FILE_FIFO:
    ino->i_mode |= S_IFIFO;
    break;
  default:
    printk(KERN_ERR "GFS:  D'oh!  dinode %"FMT_64"u has bad type %d\n",
	   ip->i_num, ip->i_di.di_type);
    break;
  };

  ino->i_nlink = ip->i_di.di_nlink;
  ino->i_uid = ip->i_di.di_uid;
  ino->i_gid = ip->i_di.di_gid;

#ifdef __alpha__
  ino->i_size = ip->i_di.di_bsize;
#else 
  ino->i_size = (ip->i_di.di_bsize & 0xFFFFFFFF80000000ll) ? -1 : ip->i_di.di_bsize;
#endif  /*  __alpha__  */

  ino->i_atime = ip->i_di.di_atime;
  ino->i_mtime = ip->i_di.di_mtime;
  ino->i_ctime = ip->i_di.di_ctime;
  ino->i_blksize = 1 << 20;
  ino->i_blocks = ip->i_di.di_blocks * vsb2sbd(ino->i_sb)->sd_sb.sb_bsize / 512;
  ino->i_version = ++event;

  EXIT(GFS_INODE_ATTR_IN, DEBUG_INODE1);
}


/******************************************************************************
*******************************************************************************
**
** gfs_inode_attr_out()
**
** Description:
**   Copy attributes out of the VFS inode into the dinode
**
**   Only copy out the attributes that we want the VFS layer
**   to be able to modify.
**
** Input(s):
**   ip - the inode
**
*******************************************************************************
******************************************************************************/

void gfs_inode_attr_out(gfs_inode_t *ip)
{
  struct inode *ino = ip->i_vino;

  ENTER(GFS_INODE_ATTR_OUT, DEBUG_INODE1);
 
  ip->i_di.di_mode = ino->i_mode & S_IALLUGO;
  ip->i_di.di_uid = ino->i_uid;
  ip->i_di.di_gid = ino->i_gid;
  ip->i_di.di_atime = ino->i_atime;
  ip->i_di.di_mtime = ino->i_mtime;
  ip->i_di.di_ctime = ino->i_ctime;

  EXIT(GFS_INODE_ATTR_OUT, DEBUG_INODE1);
}
 


/******************************************************************************
*******************************************************************************
**
** gfs_lookupi()
**
** Description:
**   Look up a filename in a directory and return its inode
**
**   Be sure do do a d_add() if this inode is going to kept of any length of
**   time.
**
** Input(s):
**   dip - the directory to search
**   name - the name of the inode we are looking for
**
** Output(s):
**   inode - the inode of the new file
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

static int gfs_lookupi(gfs_inode_t *dip, struct qstr *name, struct inode **inode, int flags)
{
  gfs_sbd_t *sdp = dip->i_sbd;
  gfs_inode_t *ip = NULL;

  int error;
  uint64 ino, ino2;
  uint64 glock;

  gfs_list_head_t *tmp;

  ENTER(GFS_LOOKUPI, DEBUG_INODE1);

  *inode = NULL;

  if (name->len > GFS_FNAMESIZE)
  {
    EXIT(GFS_LOOKUPI, DEBUG_INODE1);
    
    return(-ENAMETOOLONG);
  }

  ASSERT(!gfs_filecmp(name, ".", 1));
  ASSERT(!gfs_filecmp(name, "..", 2));


 restart:

  error = gfs_glock(sdp, dip->i_di.di_glock, dip, GL_SHARED);
  if (error)
    goto fail;

  if ((error = permission(dip->i_vino, MAY_EXEC)) || 
      (error = gfs_dir_search(dip, name, &ino)))
  {
    gfs_gunlock(sdp, dip->i_di.di_glock, 0);
    goto fail;
  }

  glock = gfs_ino2gl(sdp, ino);


  /*  Acquire the second lock  */

  if (glock <= dip->i_di.di_glock)  /*  <= to prevent recursive lock promotion  */
  {
    gfs_gunlock(sdp, dip->i_di.di_glock, 0);
    
    error = gfs_glock(sdp, glock, NULL, 0);
    if (error)
      goto fail;

    error = gfs_glock(sdp, dip->i_di.di_glock, dip, GL_SHARED);
    if (error)
    {
      gfs_gunlock(sdp, glock, 0);
      goto fail;
    }

    if ((error = permission(dip->i_vino, MAY_EXEC)) || 
	(error = gfs_dir_search(dip, name, &ino2)))
    {
      gfs_gunlock(sdp, dip->i_di.di_glock, 0);
      gfs_gunlock(sdp, glock, 0);
      goto fail;
    }

    if (ino != ino2)
    {
      gfs_gunlock(sdp, dip->i_di.di_glock, 0);
      gfs_gunlock(sdp, glock, 0);
      goto restart;
    }
  }
  else
  {
    error = gfs_glock(sdp, glock, NULL, 0);
    if (error)
    {
      gfs_gunlock(sdp, dip->i_di.di_glock, 0);
      goto fail;
    }
  }


  /*  Prevent an inode from being created while gfs_inoded
      is deleting it.
      Note!!!  This section isn't SMP safe.  */

  GFS_SEMA_DOWN(&sdp->sd_id_mutex);

  for (tmp = sdp->sd_id_list.next; tmp != &sdp->sd_id_list; tmp = tmp->next)
  {
    ip = gfs_list_entry(tmp, gfs_inode_t, i_list);
    if (ip->i_num == ino)
      break;
    else 
      ip = NULL;
  }

  GFS_SEMA_UP(&sdp->sd_id_mutex);

  if (ip)
  {
    IFDEBUG(DEBUG_PARANOIA,
	    printk(DEBUG_HOW "deadlock in lookup\n"););
    gfs_gunlock(sdp, glock, 0);
    gfs_gunlock(sdp, dip->i_di.di_glock, 0);
    error = -EDEADLOCK;
    goto fail;
  }


  *inode = iget(sdp->sd_vsb, ino);

  if (!*inode)
  {
    gfs_gunlock(sdp, glock, 0);
    gfs_gunlock(sdp, dip->i_di.di_glock, 0);
    error = -EACCES;
    goto fail;
  }


  gfs_gunlock(sdp, glock, 0);
  gfs_gunlock(sdp, dip->i_di.di_glock, 0);

  EXIT(GFS_LOOKUPI, DEBUG_INODE1);
 
  return(0);


 fail:
  EXIT(GFS_LOOKUPI, DEBUG_INODE1);
 
  return(error);
}


/******************************************************************************
*******************************************************************************
**
** gfs_lookup()
**
** Description:
**   Look up a filename in a directory and return its inode
**
**   Called by the VFS layer,  Lock dir and call gfs_lookupi()
**
** Input(s):
**   dir - the directory we are searching in
**   dentry - the dentry of the new inode
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

struct dentry *gfs_lookup(struct inode *dir, struct dentry *dentry)
{
  gfs_inode_t *dip = vi2ino(dir);
  struct inode *inode = NULL;
  gfs_dcached_t *dcd;

  int error;

  ENTER(GFS_LOOKUP, DEBUG_INODE1);
 
  IFDEBUG(DEBUG_INODE2,
	  printk(DEBUG_HOW "  %lu, %s\n", dir->i_ino, dentry->d_name.name););

 restart:

  error = gfs_lookupi(dip, &dentry->d_name, &inode, 0);


  if (error == -EDEADLOCK)
  {
    IFDEBUG(DEBUG_DEADLOCK,
	    printk(DEBUG_HOW "  lookup/clear_inode deadlock, retrying...\n"););
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(HZ >> 3);
    goto restart;
  }
  else if (error && error != -ENOENT)
  {
    IFDEBUG(DEBUG_INODE2,
	    printk(DEBUG_HOW "  error = %d\n", error););
    EXIT(GFS_LOOKUP, DEBUG_INODE1);
  
    return(ERR_PTR(error));
  }


  /*  Set up the dcache data and operations  */

  if (!vdc2dcd(dentry))
  {
    dcd = (gfs_dcached_t *)gmalloc(sizeof(gfs_dcached_t));
    
    dcd->dc_vn = dip->i_vn;
    dcd->dc_time = CURRENT_TIME;

    vdc2dcd(dentry) = dcd;
  }

  dentry->d_op = &gfs_dops;


  d_add(dentry, inode);


  IFDEBUG(DEBUG_INODE2,
	  printk(DEBUG_HOW "  error = %d\n", error););

  EXIT(GFS_LOOKUP, DEBUG_INODE1);
 
  return(NULL);
}


/******************************************************************************
*******************************************************************************
**
** gfs_createi()
**
** Description:
**   Create a new file
**
**   If the return value is 0, the glocks on both the directory and the new
**   file are held.
**
** Input(s):
**   dip - the directory to create the file in
**   name - the name of the new file
**   flag - the flags that the file was created with O_EXCL, etc...
**   mode - the files permissions
**   is_dir - whether or not we are creating a directory
**
** Output(s):
**   inode_num - if this is not NULL, the new dinode number
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

static int gfs_createi(gfs_inode_t *dip, struct qstr *name, int flag, int mode,
		       uint64 *inode_num, int is_dir)
{
  gfs_sbd_t *sdp = dip->i_sbd;
  gfs_blklist_t *blk;
  gfs_dinode_t di;
  buffer_head_t *bh;

  uint32 rgrp;
  int error;
  uint64 glock, ino, ino2;
  int new, have_dir;

  ENTER(GFS_CREATEI, DEBUG_INODE1);
 
  if (inode_num)
    *inode_num = 0;

  if (name->len > GFS_FNAMESIZE)
  {
    error = -ENAMETOOLONG;
    goto fail;
  }


 restart:

  error = gfs_glock(sdp, dip->i_di.di_glock, dip, 0);
  if (error)
    goto fail;
  have_dir = TRUE;


  error = permission(dip->i_vino, MAY_WRITE | MAY_EXEC);
  if (error)
    goto fail_unlock_dir;

  /*  Don't create entries in an unlinked directory  */

  if (dip->i_di.di_nlink == 0)
  {
    error = -EPERM;
    goto fail_unlock_dir;
  }


  error = gfs_dir_search(dip, name, &ino);
  switch (error)
  {
  case 0:
    if (flag & O_EXCL)
    {
      error = -EEXIST;
      goto fail_unlock_dir;
    }
    else
      new = FALSE;

    break;
  case -ENOENT:
    new = TRUE;

    error = gfs_blkdalloc(dip, &blk, 0xFFFFFFFF,
			  0xFFFFFFFFFFFFFFFFll, is_dir);

    if (error)
      goto fail_unlock_dir;

    ino = blk->blk_no;

    break;
  default:
    goto fail_unlock_dir;
    break;
  };


  glock = gfs_ino2gl(sdp, ino);


  if (glock >= dip->i_di.di_glock)
  {
    error = gfs_glock(sdp, glock, NULL, 0);
    if (error)
      goto fail_blkfree;
  }
  else
  {
    gfs_gunlock(sdp, dip->i_di.di_glock, 0);
    have_dir = FALSE;

    error = gfs_glock(sdp, glock, NULL, 0);
    if (error)
      goto fail_blkfree;
    
    error = gfs_glock(sdp, dip->i_di.di_glock, dip, 0);
    if (error)
    {
      gfs_gunlock(sdp, glock, 0);
      goto fail_blkfree;
    }
    have_dir = TRUE;

    error = permission(dip->i_vino, MAY_WRITE | MAY_EXEC);
    if (error)
      goto fail_unlock_both;

    /*  Don't create entries in an unlinked directory  */

    if (dip->i_di.di_nlink == 0)
    {
      error = -EPERM;
      goto fail_unlock_both;
    }
    
    error = gfs_dir_search(dip, name, &ino2);
    switch (error)
    {
    case 0:
      if (new)
      {
	gfs_blkfree(sdp, &blk);
	if (blk)
	{
	  GARBAGE_BLOCKS();
	  gfs_blklist_free(blk);
	}

	if (flag & O_EXCL)
	{
	  new = FALSE;
	  error = -EEXIST;
	  goto fail_unlock_both;
	}
	else
	{
	  gfs_gunlock(sdp, dip->i_di.di_glock, 0);
	  gfs_gunlock(sdp, glock, 0);
	  goto restart;
	}
      }

      if (ino != ino2)
      {
	gfs_gunlock(sdp, dip->i_di.di_glock, 0);
	gfs_gunlock(sdp, glock, 0);
	goto restart;
      }

      break;
    case -ENOENT:
      if (!new)
      {
	gfs_gunlock(sdp, dip->i_di.di_glock, 0);
	gfs_gunlock(sdp, glock, 0);
	goto restart;
      }

      break;
    default:
      goto fail_unlock_both;
      break;
    };
  }


  if (!new)
  {
    if (inode_num)
      *inode_num = ino;
    
    EXIT(GFS_CREATEI, DEBUG_INODE1);
   
    return(0);
  }


  bh = gfs_dgetblk(sdp, ino, glock);
  if (!bh)
  {
    error = -EIO;
    goto fail_unlock_both;
  }
  gfs_mark_uptodate(bh, TRUE);
  memset(BH_DATA(bh), 0, sdp->sd_sb.sb_bsize);


  memset(&di, 0, sizeof(gfs_dinode_t));

  di.di_magic = GFS_MAGIC_DI;

  di.di_addr = blk->blk_no;

  di.di_mode = mode & S_IALLUGO;
  di.di_uid = current->fsuid;
  di.di_gid = current->fsgid;
  di.di_nlink = 1;
  /*di.di_bsize = 0;*/
  di.di_blocks = 1;
  di.di_atime = di.di_mtime = di.di_ctime = CURRENT_TIME;
  /*di.di_major = 0;*/
  /*di.di_minor = 0;*/

  for (rgrp = 0; rgrp < sdp->sd_sb.sb_rgrps; rgrp++)
    if (blk->blk_no < sdp->sd_rgd[rgrp].rd_ri.ri_data1 + sdp->sd_rgd[rgrp].rd_ri.ri_data)
      break;
  ASSERT(rgrp < sdp->sd_sb.sb_rgrps);

  di.di_rgrp = rgrp;

  di.di_rgrp_meta = di.di_rgrp_data = rgrp;
  di.di_last_meta = di.di_last_data = blk->blk_no - sdp->sd_rgd[rgrp].rd_ri.ri_data1;

  di.di_glock = glock;
  di.di_type = FILE_REG;
  /*di.di_height = 0;*/
  di.di_stamp = (uint16)(gfs_random() >> 16);

  /*di.di_nopen = 0;*/
  /*memset(di.di_open, 0, GFS_MOUNT_BITS);*/

  di.di_regime = GFS_DIR_LINEAR;
  /*di.di_depth = 0;*/
  /*di.di_entries = 0;*/


  gfs_dinode_out(&di, BH_DATA(bh));
  gfs_dwrite(bh, DW_NOKEEP);


  if ((error = gfs_dir_add(dip, name, blk->blk_no)))
    goto fail_unlock_both;


  gfs_blklist_free(blk);


  if (inode_num)
    *inode_num = ino;

  EXIT(GFS_CREATEI, DEBUG_INODE1);
 
  return(0);



 fail_unlock_both:
  gfs_gunlock(sdp, glock, 0);

 fail_blkfree:
  if (new)
  {
    gfs_blkfree(sdp, &blk);
    if (blk)
    {
      GARBAGE_BLOCKS();
      gfs_blklist_free(blk);
    }
  }

 fail_unlock_dir:
  if (have_dir)
    gfs_gunlock(sdp, dip->i_di.di_glock, 0);

 fail:
  EXIT(GFS_CREATEI, DEBUG_INODE1);
 
  return(error);  
}


/******************************************************************************
*******************************************************************************
**
** gfs_create()
**
** Description:
**   Create a file
**
**   FIX ME!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
**
**   The "flag" variable passed into the creat() syscall should be propagated
**   all the way down to this function.  Without it, there is a race condition
**   with the O_EXCL flag.
**
** Input(s):
**   dir - the directory to create the file in
**   dentry - the dentry of the new file
**   mode - the permissions of the new file
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_create(struct inode *dir, struct dentry *dentry, int mode)
{
  gfs_inode_t *dip = vi2ino(dir);
  gfs_sbd_t *sdp = dip->i_sbd;
  struct inode *inode = NULL;
  
  uint64 inode_num;
  int error;

  ENTER(GFS_CREATE, DEBUG_INODE1);

 restart:

  error = gfs_createi(dip, &dentry->d_name, 0, mode, &inode_num, 0);
  if (error)
  {
    EXIT(GFS_CREATE, DEBUG_INODE1);
  
    return(error);
  }

  
  error = gfs_lookupi(dip, &dentry->d_name, &inode, 0);
  if (error)
  {
    gfs_gunlock(sdp, gfs_ino2gl(sdp, inode_num), 0);
    gfs_gunlock(sdp, dip->i_di.di_glock, 0);

    if (error == -EDEADLOCK)
      goto restart;

    EXIT(GFS_CREATE, DEBUG_INODE1);
   
    return(error);    
  }


  gfs_gunlock(sdp, vi2ino(inode)->i_di.di_glock, 0);
  gfs_gunlock(sdp, dip->i_di.di_glock, 0);


  dir->i_version = ++event;

  d_instantiate(dentry, inode);


  EXIT(GFS_CREATE, DEBUG_INODE1);
  
  return(error);
}


/******************************************************************************
*******************************************************************************
**
** gfs_link()
**
** Description:
**   Link the inode in "old_dentry" into the directory "dir" with the
**     name in "dentry".
**
** Input(s):
**   old_dentry - the old inode to link
**   dir - link into this directory
**   dentry - give it this name
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_link(struct dentry *old_dentry, struct inode *dir, struct dentry *dentry)
{
  gfs_inode_t *dip = vi2ino(dir);
  gfs_sbd_t *sdp = dip->i_sbd;
  gfs_inode_t *ip = vi2ino(old_dentry->d_inode);
  gfs_lockop_t lock[2];

  int error;

  ENTER(GFS_LINK, DEBUG_INODE1);
 
  memset(lock, 0, 2 * sizeof(gfs_lockop_t));

  lock[0].lock = dip->i_di.di_glock;
  lock[0].ip = dip;

  lock[1].lock = ip->i_di.di_glock;
  lock[1].ip = ip;


  error = gfs_glock_m(sdp, 2, lock);
  if (error)
  {
    EXIT(GFS_LINK, DEBUG_INODE1);
 
    return(error);
  }
    

  if ((error = permission(dir, MAY_WRITE | MAY_EXEC)))
  {
    gfs_gunlock_m(sdp, 2, lock);
    EXIT(GFS_LINK, DEBUG_INODE1);
   
    return(error);
  }

  error = gfs_dir_search(dip, &dentry->d_name, NULL);

  if (error == 0)
  {
    gfs_gunlock_m(sdp, 2, lock);
    EXIT(GFS_LINK, DEBUG_INODE1);
    
   
    return(-EEXIST);
  }
  if (error != -ENOENT)
  {
    gfs_gunlock_m(sdp, 2, lock);
    EXIT(GFS_LINK, DEBUG_INODE1);
   
    return(error);
  }

  error = gfs_change_nlink(ip, +1);
  if (error)
  {
    gfs_gunlock_m(sdp, 2, lock);
    EXIT(GFS_LINK, DEBUG_INODE1);
   
    return(error);
  }


  error = gfs_dir_add(dip, &dentry->d_name, ip->i_num);
  if (error)
  {
    gfs_change_nlink(ip, -1);
    gfs_gunlock_m(sdp, 2, lock);
    EXIT(GFS_LINK, DEBUG_INODE1);
   
    return(error);
  }


  gfs_gunlock_m(sdp, 2, lock);

  ip->i_vino->i_count++;

  d_instantiate(dentry, ip->i_vino);

  EXIT(GFS_LINK, DEBUG_INODE1);
 
  return(error);
}


/******************************************************************************
*******************************************************************************
**
** unlinki()
**
** Description:
**   Unlink a file.
**   Assumes Glocks on both dip and ip are held.
**
** Input(s):
**   dip - the directory holding the file to be removed
**   dname - the name of the file to be removed
**   ip - the file to be removed
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

static int unlinki(gfs_inode_t *dip, struct qstr *dname, gfs_inode_t *ip)
{
  int error;

  ENTER(UNLINKI, DEBUG_INODE1);

  if ((error = gfs_dir_del(dip, dname)))
  {
    EXIT(UNLINKI, DEBUG_INODE1);
   
    return(error);
  }

  /*  gfs_change_nlink() won't fail on decrement.  */

  gfs_change_nlink(ip, -1);

  EXIT(UNLINKI, DEBUG_INODE1);
 
  return(0);
}


/******************************************************************************
*******************************************************************************
**
** gfs_unlink()
**
** Description:
**   Unlink a file.  Call unlinki()
**
** Input(s):
**   dir - the directory containing the file
**   dentry - the file 
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_unlink(struct inode *dir, struct dentry *dentry)
{
  gfs_inode_t *dip = vi2ino(dir);
  gfs_sbd_t *sdp = dip->i_sbd;
  gfs_inode_t *ip = vi2ino(dentry->d_inode);
  gfs_lockop_t lock[2];

  int error;

  ENTER(GFS_UNLINK, DEBUG_INODE1);
 
  memset(lock, 0, 2 * sizeof(gfs_lockop_t));

  lock[0].lock = dip->i_di.di_glock;
  lock[0].ip = dip;

  lock[1].lock = ip->i_di.di_glock;
  lock[1].ip = ip;


  error = gfs_glock_m(sdp, 2, lock);
  if (error)
  {
    EXIT(GFS_UNLINK, DEBUG_INODE1);
    
   
    return(error);
  }

  
  if (!gfs_drevalidate(dentry, 0))
  {
    gfs_gunlock_m(sdp, 2, lock);
    EXIT(GFS_UNLINK, DEBUG_INODE1);
    
    return(-ENOENT);
  }


  if ((error = permission(dir, MAY_WRITE | MAY_EXEC)))
  {
    gfs_gunlock_m(sdp, 2, lock);
    EXIT(GFS_UNLINK, DEBUG_INODE1);
   
    return(error);
  }


  error = unlinki(dip, &dentry->d_name, ip);


  gfs_gunlock_m(sdp, 2, lock);


  if (!error)
    d_delete(dentry);

  EXIT(GFS_UNLINK, DEBUG_INODE1);
 
  return(error);
}


/******************************************************************************
*******************************************************************************
**
** gfs_symlink()
**
** Description:
**   Create a symlink
**
** Input(s):
**   dir - the directory to create the symlink in
**   dentry - the dentry to put the symlink in
**   symname - the thing the symlink points to
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_symlink(struct inode *dir, struct dentry *dentry, const char *symname)
{
  gfs_inode_t *dip = vi2ino(dir);
  gfs_sbd_t *sdp = dip->i_sbd;
  buffer_head_t *bh;
  gfs_dinode_t dinode;
  struct inode *inode = NULL;

  uint64 inode_num;
  int error;

  ENTER(GFS_SYMLINK, DEBUG_INODE1);
 


  /*  Must be stuffed with a null terminator for gfs_follow_link()  */

  if (strlen(symname) > sdp->sd_sb.sb_bsize - sizeof(gfs_dinode_t) - 1)
  {
    EXIT(GFS_SYMLINK, DEBUG_INODE1);
    
    return(-ENAMETOOLONG);
  }


  error = gfs_createi(dip, &dentry->d_name, O_EXCL, 0777, &inode_num, 0);
  if (error)
  {
    EXIT(GFS_SYMLINK, DEBUG_INODE1);
   
    return(error);
  }


  bh = gfs_dread(sdp, inode_num, gfs_ino2gl(sdp, inode_num));

  gfs_dinode_in(&dinode, BH_DATA(bh));

  dinode.di_type = FILE_LNK;
  dinode.di_bsize = strlen(symname);

  gfs_dinode_out(&dinode, BH_DATA(bh));

  memcpy(BH_DATA(bh) + sizeof(gfs_dinode_t), symname, dinode.di_bsize);

  gfs_dwrite(bh, DW_NOKEEP);

  
  inode = iget(sdp->sd_vsb, inode_num);
  if (!inode)
  {
    gfs_gunlock(sdp, dinode.di_glock, 0);
    gfs_gunlock(sdp, dip->i_di.di_glock, 0);
    EXIT(GFS_SYMLINK, DEBUG_INODE1);
   
    return(-EACCES);
  }


  gfs_gunlock(sdp, vi2ino(inode)->i_di.di_glock, 0);
  gfs_gunlock(sdp, dip->i_di.di_glock, 0);


  dir->i_version = ++event;

  d_instantiate(dentry, inode);

  EXIT(GFS_SYMLINK, DEBUG_INODE1);
 

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** gfs_mkdir()
**
** Description:
**   Make a directory
**
** Input(s):
**   dir - the directory to create the new directory in
**   dentry - the dentry for the new directory
**   mode - the permissions on the new directory
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_mkdir(struct inode *dir, struct dentry *dentry, int mode)
{
  gfs_inode_t *dip = vi2ino(dir);
  gfs_sbd_t *sdp = dip->i_sbd;
  buffer_head_t *bh;
  gfs_dinode_t *dinode;
  gfs_dirent_t *dent;
  struct inode *inode = NULL;
  gfs_blklist_t *blk;

  uint64 inode_num;
  int error;

  ENTER(GFS_MKDIR, DEBUG_INODE1);
 


  blk = (gfs_blklist_t *)gmalloc(sizeof(gfs_blklist_t));
  if (!blk)
  {
    EXIT(GFS_MKDIR, DEBUG_INODE1);
   
    return(-ENOMEM);
  }


  error = gfs_createi(dip, &dentry->d_name, O_EXCL, mode, &inode_num, 1);
  if (error)
  {
    gfree(blk, sizeof(gfs_blklist_t));
    EXIT(GFS_MKDIR, DEBUG_INODE1);
   
    return(error);
  }

  bh = gfs_dread(sdp, inode_num, gfs_ino2gl(sdp, inode_num));
 
  dinode = (gfs_dinode_t *) BH_DATA(bh);
  dinode->di_type = cpu_to_be16(FILE_DIR);
  dinode->di_bsize = cpu_to_be64(sdp->sd_sb.sb_bsize - sizeof(gfs_dinode_t));
  dinode->di_nlink = cpu_to_be32(2);
  dinode->di_entries = 0;

  error = gfs_dirent_alloc(bh, 1, &dent);
  if (error)
    goto fail;

  dent->de_hash = cpu_to_be32(0x0ED4E242);
  dent->de_ino = dinode->di_addr; /* already big endian */
  memcpy(dent->de_name, ".", 1);
  dinode->di_entries = cpu_to_be32(1);

  error = gfs_dirent_alloc(bh, 2, &dent);
  if (error)
    goto fail;

  dent->de_hash = cpu_to_be32(0x9608161C);
  dent->de_ino = cpu_to_be64(dip->i_num);
  memcpy(dent->de_name, "..", 2);
  dinode->di_entries = cpu_to_be32(2);

  
  gfs_dwrite(bh, DW_NOKEEP);


  error = gfs_change_nlink(dip, +1);
  if (error)
    goto fail;

  gfree(blk, sizeof(gfs_blklist_t));


  inode = iget(sdp->sd_vsb, inode_num);
  if (!inode)
  {
    gfs_gunlock(sdp, dinode->di_glock, 0);
    gfs_gunlock(sdp, dip->i_di.di_glock, 0);
    EXIT(GFS_MKDIR, DEBUG_INODE1);
  
    return(-EACCES);
  }


  gfs_gunlock(sdp, vi2ino(inode)->i_di.di_glock, 0);
  gfs_gunlock(sdp, dip->i_di.di_glock, 0);


  dir->i_version = ++event;

  d_instantiate(dentry, inode);

  EXIT(GFS_MKDIR, DEBUG_INODE1);
 

  return(0);



 fail:
  gfs_dir_del(dip, &dentry->d_name);
    
  blk->blk_no = inode_num;
  blk->blk_len = 1;
  blk->blk_next = NULL;

  gfs_blkfree(sdp, &blk);
  if (blk)
  {
    GARBAGE_BLOCKS();
    gfs_blklist_free(blk);
  }

  gfs_gunlock(sdp, dinode->di_glock, 0);
  gfs_gunlock(sdp, dip->i_di.di_glock, 0);

  EXIT(GFS_MKDIR, DEBUG_INODE1);
 
  return(error);    
}


/******************************************************************************
*******************************************************************************
**
** rmdiri()
**
** Description:
**   Remove a directory
**   Assumes Glocks on dip and ip are held
**
** Input(s):
**   dip - the directory holding the directory to be removed
**   dname - the name of the dirctory to be removed
**   ip - the directory to be removed
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

static int rmdiri(gfs_inode_t *dip, struct qstr *dname, gfs_inode_t *ip)
{
  struct qstr name;

  int error;

  ENTER(RMDIRI, DEBUG_INODE1);
 
  /*  Check to make sure the directory is empty!!!!  */

  if (ip->i_di.di_entries > 2)
  {
    EXIT(RMDIRI, DEBUG_INODE1);
    
  
    return(-EEXIST);
  }


  if ((error = gfs_dir_del(dip, dname)))
  {
    EXIT(RMDIRI, DEBUG_INODE1);
  
    return(error);
  }
  
  gfs_change_nlink(dip, -1);

  name.len = 1;
  name.name = ".";
  gfs_dir_del(ip, &name);

  name.len = 2;
  name.name = "..";
  gfs_dir_del(ip, &name);

  ip->i_di.di_nlink = 0;

  gfs_inode_attr_in(ip);
  gfs_dinode_out(&ip->i_di, BH_DATA(ip->i_didata));
  gfs_dwrite(ip->i_didata, 0);

  EXIT(RMDIRI, DEBUG_INODE1);
 

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** gfs_rmdir()
**
** Description:
**   Remove a directory.  Call rmdiri()
**
** Input(s):
**   dir - the directory containing the directory to remove
**   dentry - the dentry for the direcotry to remove
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_rmdir(struct inode *dir, struct dentry *dentry)
{
  gfs_inode_t *dip = vi2ino(dir);
  gfs_sbd_t *sdp = dip->i_sbd;
  gfs_inode_t *ip = vi2ino(dentry->d_inode);
  gfs_lockop_t lock[2];

  int error;

  ENTER(GFS_RMDIR, DEBUG_INODE1);
 


  memset(lock, 0, 2 * sizeof(gfs_lockop_t));

  lock[0].lock = dip->i_di.di_glock;
  lock[0].ip = dip;

  lock[1].lock = ip->i_di.di_glock;
  lock[1].ip = ip;


  error = gfs_glock_m(sdp, 2, lock);
  if (error)
  {
    EXIT(GFS_RMDIR, DEBUG_INODE1);
   
    return(error);
  }


  if (!gfs_drevalidate(dentry, 0))
  {
    gfs_gunlock_m(sdp, 2, lock);
    EXIT(GFS_RMDIR, DEBUG_INODE1);
  
    return(-ENOENT);
  }


  error = permission(dir, MAY_WRITE | MAY_EXEC);
  if (error)
  {
    gfs_gunlock_m(sdp, 2,lock);
    EXIT(GFS_RMDIR, DEBUG_INODE1);
  
    return(error);    
  }


  if (ip->i_di.di_type != FILE_DIR)
  {
    gfs_gunlock_m(sdp, 2, lock);
    EXIT(GFS_RMDIR, DEBUG_INODE1);
   
    return(-ENOTDIR);    
  }

  if (dentry->d_count > 1)  /*  Monkey see (ext2 do this), Monkey do.  */
    shrink_dcache_parent(dentry);


  error = rmdiri(dip, &dentry->d_name, ip);

  
  gfs_gunlock_m(sdp, 2, lock);


  if (!error)
    d_delete(dentry);

  EXIT(GFS_RMDIR, DEBUG_INODE1);


  return(error);
}


/******************************************************************************
*******************************************************************************
**
** gfs_mknod()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int gfs_mknod(struct inode *dir, struct dentry *dentry, int mode, int rdev)
{
  gfs_inode_t *dip = vi2ino(dir);
  gfs_sbd_t *sdp = dip->i_sbd;
  buffer_head_t *bh;
  gfs_dinode_t dinode;
  struct inode *inode = NULL;
  kdev_t dev;

  uint64 inode_num;
  int error;

  ENTER(GFS_MKNOD, DEBUG_INODE1);

  if (S_ISFIFO(mode))
  {
    EXIT(GFS_MKNOD, DEBUG_INODE1);
  
    return(-ENOSYS);
  }


  error = gfs_createi(dip, &dentry->d_name, O_EXCL, mode, &inode_num, 0);
  if (error)
  {
    EXIT(GFS_MKNOD, DEBUG_INODE1);
  
    return(error);
  }


  bh = gfs_dread(sdp, inode_num, gfs_ino2gl(sdp, inode_num));

  gfs_dinode_in(&dinode, BH_DATA(bh));

  switch (mode & S_IFMT)
  {
  case S_IFREG:
    printk(KERN_ERR "GFS:  using mknod to create a regular file?!?\n");
    break;
  case S_IFBLK:
    dinode.di_type = FILE_BLK;
    dev = to_kdev_t(rdev);
    dinode.di_major = MAJOR(dev);
    dinode.di_minor = MINOR(dev);
    break;
  case S_IFCHR:
    dinode.di_type = FILE_CHR;
    dev = to_kdev_t(rdev);
    dinode.di_major = MAJOR(dev);
    dinode.di_minor = MINOR(dev);
    break;
  default:
    panic("GFS:  weird mode passed into mknod (%d)\n", mode);
    break;
  };

  gfs_dinode_out(&dinode, BH_DATA(bh));

  gfs_dwrite(bh, DW_NOKEEP);

  
  inode = iget(sdp->sd_vsb, inode_num);
  if (!inode)
  {
    gfs_gunlock(sdp, dinode.di_glock, 0);
    gfs_gunlock(sdp, dip->i_di.di_glock, 0);
    EXIT(GFS_MKNOD, DEBUG_INODE1);
   
    return(-EACCES);
  }


  gfs_gunlock(sdp, vi2ino(inode)->i_di.di_glock, 0);
  gfs_gunlock(sdp, dip->i_di.di_glock, 0);


  dir->i_version = ++event;

  d_instantiate(dentry, inode);

  EXIT(GFS_MKNOD, DEBUG_INODE1);
 

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** gfs_is_subdir()
**
** Description:
**   Return TRUE if new_dentry is a subdir of old_dentry
**
**   Do the same thing as Linux's is_subdir(), but do drevalidates to make
**   sure the dcache is consistent.
**
** Input(s):
**   new_dentry - 
**   old _dentry -
**
** Returns:
**   TRUE if new_dentry is a subdir of old_dentry, FALSE otherwise
**
*******************************************************************************
******************************************************************************/

static int gfs_is_subdir(struct dentry *new_dentry, struct dentry *old_dentry)
{
  int ret = FALSE;
  int error;

  ENTER(GFS_IS_SUBDIR, DEBUG_INODE1);
 

  while (TRUE)
  {
    if (new_dentry != old_dentry)
    {
      if (new_dentry->d_parent == new_dentry)
	break;

      error = gfs_drevalidate(new_dentry, 0);
      if (error <= 0)
      {
	EXIT(GFS_IS_SUBDIR, DEBUG_INODE1);

	if (!error)
	  error = -ENOENT;
	return(error);
      }

      new_dentry = new_dentry->d_parent;
      continue;
    }

    ret = TRUE;
    break;
  }

  EXIT(GFS_IS_SUBDIR, DEBUG_INODE1);
 
  return(ret);
}


/******************************************************************************
*******************************************************************************
**
** gfs_rename()
**
** Description:
**   Rename a file
**
** Input(s):
**   odir - the directory containing the old name of the file
**   odentry - the old dentry of the file
**   ndir - the directory containing the new name of the file
**   ndentry - the new dentry for the file
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_rename(struct inode *odir, struct dentry *odentry,
	       struct inode *ndir, struct dentry *ndentry)
{
  gfs_inode_t *odip = vi2ino(odir), *ndip = vi2ino(ndir);
  gfs_sbd_t *sdp = odip->i_sbd;
  gfs_inode_t *ip = vi2ino(odentry->d_inode), *nip = NULL;
  struct qstr name;
  gfs_lockop_t lock[4];
  int num_lock;

  int error;

  ENTER(GFS_RENAME, DEBUG_INODE1);

  if (ndentry->d_inode)
  {
    nip = vi2ino(ndentry->d_inode);

    if (ip == nip)
    {
      error = 0;
      goto fail;
    }

    if ((ip->i_di.di_type == FILE_DIR || nip->i_di.di_type == FILE_DIR) &&
	ip->i_di.di_type != nip->i_di.di_type)
    {
      if (nip->i_di.di_type == FILE_DIR)
	error = -EISDIR;
      else
	error = -ENOTDIR;
      goto fail;
    }
  }


  /*  Make sure we aren't trying to move a dirctory into it's subdir  */

  if (ip->i_di.di_type == FILE_DIR)
  {
    error = gfs_glock(sdp, GFS_RENAME_LOCK, NULL, 0);
    if (error)
      goto fail;

    error = gfs_is_subdir(ndentry->d_parent, odentry);
    switch (error)
    {
    case TRUE:
      error = -EINVAL;
      goto fail_renlock;
      break;
    case FALSE:
      break;
    default:
      goto fail_renlock;
      break;
    }
  }


  memset(lock, 0, 4 * sizeof(gfs_lockop_t));

  lock[0].lock = odip->i_di.di_glock;
  lock[0].ip = odip;

  lock[1].lock = ndip->i_di.di_glock;
  lock[1].ip = ndip;

  num_lock = 2;

  if (nip)
  {
    lock[num_lock].lock = nip->i_di.di_glock;
    lock[num_lock].ip = nip;
    num_lock++;
  }

  if (ip->i_di.di_type == FILE_DIR && odip != ndip)
  {
    lock[num_lock].lock = ip->i_di.di_glock;
    lock[num_lock].ip = ip;
    num_lock++;
  }


  error = gfs_glock_m(sdp, num_lock, lock);
  if (error)
    goto fail_renlock;
  

  /*  Check out the old directory  */

  error = permission(odir, MAY_WRITE | MAY_EXEC);
  if (error)
    goto fail_glock;

  if (!gfs_drevalidate(odentry, 0))
  {
    error = -ENOENT;
    goto fail_glock;
  }


  /*  Check out the new directory  */

  error = permission(ndir, MAY_WRITE | MAY_EXEC);
  if (error)
    goto fail_glock;

  if (nip && !gfs_drevalidate(ndentry, 0))
  {
    error = -ENOENT;
    goto fail_glock;
  }


  /*  Remove the target file, if it exists  */

  if (nip)
  {
    if (nip->i_di.di_type == FILE_DIR)
      error = rmdiri(ndip, &ndentry->d_name, nip);
    else
      error = unlinki(ndip, &ndentry->d_name, nip);

    if (error)
      goto fail_glock;
  }


  /*  Filesystem consistency be damned.  FULL SPEED AHEAD!!!  */

  if (ip->i_di.di_type == FILE_DIR && odip != ndip)
  {
    gfs_change_nlink(ndip, +1);
    gfs_change_nlink(odip, -1);
    name.len = 2;
    name.name = "..";
    gfs_dir_mvino(ip, &name, ndip->i_num);
  }

  gfs_dir_add(ndip, &ndentry->d_name, ip->i_num);
  gfs_dir_del(odip, &odentry->d_name);


  gfs_gunlock_m(sdp, num_lock, lock);
  if (ip->i_di.di_type == FILE_DIR)
    gfs_gunlock(sdp, GFS_RENAME_LOCK, 0);


  d_move(odentry, ndentry);

  EXIT(GFS_RENAME, DEBUG_INODE1);
 

  return(0);



 fail_glock:
  gfs_gunlock_m(sdp, num_lock, lock);

 fail_renlock:
  if (ip->i_di.di_type == FILE_DIR)
    gfs_gunlock(sdp, GFS_RENAME_LOCK, 0);  

 fail:
  EXIT(GFS_RENAME, DEBUG_INODE1);
 
  return(error);
}


/******************************************************************************
*******************************************************************************
**
** gfs_readlink()
**
** Description:
**   Read the value of a symlink
**
** Input(s):
**   dentry - the symlink
**   buf - the buffer to read the symlink data into
**   size - the size of the buffer
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_readlink(struct dentry *dentry, char *buf, int size)
{
  gfs_inode_t *ip = vi2ino(dentry->d_inode);
  gfs_sbd_t *sdp = ip->i_sbd;

  int error;

  ENTER(GFS_READLINK, DEBUG_INODE1);

  error = gfs_glock(sdp, ip->i_di.di_glock, ip, GL_SHARED);
  if (error)
  {
    EXIT(GFS_READLINK, DEBUG_INODE1);
  
    return(error);
  }

  if (size > ip->i_di.di_bsize)
    size = ip->i_di.di_bsize;

  if (copy_to_user(buf, BH_DATA(ip->i_didata) + sizeof(gfs_dinode_t), size))
    error = -EFAULT;

  gfs_gunlock(sdp, ip->i_di.di_glock, 0);
  
  EXIT(GFS_READLINK, DEBUG_INODE1);
 

  if (error)
    return(error);
  else
    return(size);
}


/******************************************************************************
*******************************************************************************
**
** gfs_irevalidate()
**
** Description:
**   Revalidate the inode.  i.e.  grab the Glock and sync the incore dinode
**     with the ondisk inode
**
** Input(s):
**   dentry - then dentry to revalidate
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_irevalidate(struct dentry *dentry)
{
  gfs_inode_t *ip = vi2ino(dentry->d_inode);
  gfs_sbd_t *sdp = ip->i_sbd;

  int error;

  ENTER(GFS_IREVALIDATE, DEBUG_INODE1);

  IFDEBUG(DEBUG_INODE2,
	  printk(DEBUG_HOW "  %lu, %s = %lu\n",
		 dentry->d_parent->d_inode->i_ino,
		 dentry->d_name.name,
		 dentry->d_inode->i_ino););


  if ((error = gfs_glock(sdp, ip->i_di.di_glock, ip, GL_SHARED)))
  {
    EXIT(GFS_IREVALIDATE, DEBUG_INODE1);
       return(error);
  }

  gfs_gunlock(sdp, ip->i_di.di_glock, 0);

  EXIT(GFS_IREVALIDATE, DEBUG_INODE1);

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** gfs_follow_link()
**
** Description:
**   Follow a symlink
**
** Input(s):
**   link - the link to follow
**   follow - some type of flag
**
** Output(s):
**   dentry - the new dentry to be filled in
**
** Returns:
**   The new dentry or NULL, maybe
**
*******************************************************************************
******************************************************************************/

struct dentry *gfs_follow_link(struct dentry *dentry, struct dentry *base, unsigned int follow)
{
  ENTER(GFS_FOLLOW_LINK, DEBUG_INODE1);
 
  gfs_irevalidate(dentry);

  base = lookup_dentry(BH_DATA(vi2ino(dentry->d_inode)->i_didata) + sizeof(gfs_dinode_t),
		       base, follow);

  EXIT(GFS_FOLLOW_LINK, DEBUG_INODE1);
 
  return(base);
}


/******************************************************************************
*******************************************************************************
**
** gfs_bmap()
**
** Description:
**   Translate a block in a file into a block on the disk
**
** Input(s):
**   inode - the inode 
**   block - the block in the file
**
** Returns:
**   The block on the disk, or zero on error (or hole)
**
*******************************************************************************
******************************************************************************/

int gfs_bmap(struct inode *inode, int block)
{
  gfs_inode_t *ip = vi2ino(inode);
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_direct_t *direct;

  int error;
  int disk_block;
  int shared;

  ENTER(GFS_BMAP, DEBUG_INODE1);
 

  if (block < 0)
  {
    printk(KERN_ERR "GFS:  gfs_bmap():  block < 0\n");;
    EXIT(GFS_BMAP, DEBUG_INODE1);
  
    return(0);
  }


  /*  The file could not have shrunk while we had it open.  */

  shared = (ip->i_di.di_height) ? GL_SHARED : 0;

  error = gfs_glock(sdp, ip->i_di.di_glock, ip, shared);
  if (error)
  {
    EXIT(GFS_BMAP, DEBUG_INODE1);
   
    return(0);
  }


  if (ip->i_di.di_height == 0)
  {
    ASSERT(!shared);

    error = gfs_unstuff_dinode(ip);
    if (error)
    {
      gfs_gunlock(sdp, ip->i_di.di_glock, 0);
      EXIT(GFS_BMAP, DEBUG_INODE1);
     
      return(0);
    }
  }


  error = gfs_bmapi_read(ip, (uint64)block * sdp->sd_sb.sb_bsize, 1, &direct);
  if (error)
  {
    gfs_gunlock(sdp, ip->i_di.di_glock, 0);
    EXIT(GFS_BMAP, DEBUG_INODE1);
    
    return(0);    
  }


  gfs_gunlock(sdp, ip->i_di.di_glock, 0);


  if (!direct)
  {
    EXIT(GFS_BMAP, DEBUG_INODE1);
   
    return(0);    
  }

  ASSERT(direct->dr_num == 1);


  if (direct->dr_no & 0xFFFFFFFF80000000ll)
  {
    printk(KERN_ERR "GFS:  gfs_bmap():  block maps out of \"int\" range\n");
    gfree(direct, direct[0].dr_num * sizeof(gfs_blklist_t));
    EXIT(GFS_BMAP, DEBUG_INODE1);
   
    return(0);    
  }

  disk_block = direct->dr_no;

  gfree(direct, direct[0].dr_num * sizeof(gfs_blklist_t));

  EXIT(GFS_BMAP, DEBUG_INODE1);
 

  return(disk_block);
}





struct inode_operations gfs_file_iops =
{
  &gfs_file_fops,
  NULL,  /*  create  */
  NULL,  /*  lookup  */
  NULL,  /*  link  */
  NULL,  /*  unlink  */
  NULL,  /*  symlink  */
  NULL,  /*  mkdir  */
  NULL,  /*  rmdir  */
  NULL,  /*  mknod  */
  NULL,  /*  rename  */
  NULL,  /*  readlink  */
  NULL,  /*  follow_link  */
  generic_readpage,
  NULL,  /*  writepage  */
  gfs_bmap,
  NULL,  /*  truncate  */
  NULL,  /*  permission  */
  NULL,  /*  smap  */
  NULL,  /*  updatepage  */
  gfs_irevalidate
};


struct inode_operations gfs_dir_iops =
{
  &gfs_dir_fops,
  gfs_create,
  gfs_lookup,
  gfs_link,
  gfs_unlink,
  gfs_symlink,
  gfs_mkdir,
  gfs_rmdir,
  gfs_mknod,
  gfs_rename,
  NULL,  /*  readlink  */
  NULL,  /*  follow_link  */
  NULL,  /*  readpage  */
  NULL,  /*  writepage  */
  NULL,  /*  bmap  */
  NULL,  /*  truncate  */
  NULL,  /*  permission  */
  NULL,  /*  smap  */
  NULL,  /*  updatepage  */
  gfs_irevalidate
};


struct inode_operations gfs_symlink_iops =
{
  NULL,  /*  struct file_operations  */
  NULL,  /*  create  */
  NULL,  /*  lookup  */
  NULL,  /*  link  */
  NULL,  /*  unlink  */
  NULL,  /*  symlink  */
  NULL,  /*  mkdir  */
  NULL,  /*  rmdir  */
  NULL,  /*  mknod  */
  NULL,  /*  rename  */
  gfs_readlink,
  gfs_follow_link,
  NULL,  /*  readpage  */
  NULL,  /*  writepage  */
  NULL,  /*  bmap  */
  NULL,  /*  truncate  */
  NULL,  /*  permission  */
  NULL,  /*  smap  */
  NULL,  /*  updatepage  */
  gfs_irevalidate
};





