/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name: bits.c
**
**    Bitmaps:
**
**         These routines are used by the block-list routines (blklist.c)
**    to keep track of block allocation.  Each block is represented by two
**    bits.  One bit indicates whether or not the block is used.  (1=used,
**    0=free)  The other bit indicates whether or not the block contains a
**    dinode or not.  (1=dinode, 0=data block)
**         So, each byte represents GFS_NBBY (i.e. 4) blocks.  
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "bits.h"


/* Function Macros */

#define SETBIT(c, b)	((c) |=  (1 << (b)))
#define CLRBIT(c, b)	((c) &= ~(1 << (b)))
#define TSTBIT(c, b)	((c) &   (1 << (b)))



/******************************************************************************
*******************************************************************************
**
** void gfs_set_bits()
**
** Description:
**   This routine sets a length of bits.
**
** Input(s):
**   buffer - pointer to the bitmap array
**   buflen - length of the bitmap array
**   no - the number of the first block to allocate
**   len - number of blocks to allocate
**   dinode - is nonzero if these bits represent dinodes
**
** Returns:
**   Nothing
**
*******************************************************************************
******************************************************************************/

void gfs_set_bits(char *buffer, uint buflen, uint64 no, uint len, int dinode)
{
  int j;
  uint64 bit, byte = 0;


  IFDEBUG(DEBUG_BITS1,
	  printk(DEBUG_HOW "Entering gfs_set_bits()\n"););


  for (j = 0; j < len; )
  {
    byte = (no + j) / GFS_NBBY;
    bit = ((no + j) % GFS_NBBY) * (8 / GFS_NBBY);
    ASSERT(byte < buflen);

    if (bit == 0 && ((len - j) >= GFS_NBBY))
    {
      ASSERT(buffer[byte] == 0);
      buffer[byte] = (dinode) ? 0xFF : 0x55;
      j += GFS_NBBY;

      continue;
    }

    ASSERT(!TSTBIT(buffer[byte], bit));
    SETBIT(buffer[byte], bit);

    if (dinode)
      SETBIT(buffer[byte], bit + 1);
    else 
      CLRBIT(buffer[byte], bit + 1);

    j++;
  }


  IFDEBUG(DEBUG_BITS1,
	  printk(DEBUG_HOW "Exiting gfs_set_bits()\n"););
}


/******************************************************************************
*******************************************************************************
**
** void gfs_clr_bits()
**
** Description:
**   This routine clears bits in the bitmap.  For each block freed it checks
**   to see if the "dinode" bit is set.  It returns the number of dinodes
**   freed.
**
** Input(s):
**   buffer - pointer to the bitmap array
**   buflen - length of the bitmap array
**   no - the number of the first block to free
**   len - number of blocks to free
**
** Returns:
**   -EXXXX on error, or the number of dinodes freed on success.
**
*******************************************************************************
******************************************************************************/

int gfs_clr_bits(char *buffer, uint buflen, uint64 no, uint len)
{
  int j;
  uint64 byte, bit;
  int dinodes = 0;


  IFDEBUG(DEBUG_BITS1,
	  printk(DEBUG_HOW "Entering gfs_clr_bits()\n"););


  for (j = 0; j < len; )
  {
    byte = (no + j) / GFS_NBBY;
    bit = ((no + j) % GFS_NBBY) * (8 / GFS_NBBY);
    ASSERT(byte < buflen);

    if (bit == 0 && ((len - j) >= GFS_NBBY))
    {
      ASSERT((buffer[byte] & 0x55) == 0x55);
      buffer[byte] = 0x00;
      j += GFS_NBBY;

      continue;
    }

    ASSERT(TSTBIT(buffer[byte], bit));
    CLRBIT(buffer[byte], bit);

    if (TSTBIT(buffer[byte], bit + 1))
      dinodes++;
    CLRBIT(buffer[byte], bit + 1);

    j++;
  }


  IFDEBUG(DEBUG_BITS1,
	  printk(DEBUG_HOW "Exiting gfs_clr_bits()\n"););


  return(dinodes);
}


/******************************************************************************
*******************************************************************************
**
** uint64 gfs_bitfit()
**
** Description:
**   This routine searches for a single free block and returns its number.
**
** Input(s):
**   buffer - pointer to the bitmap array
**   buflen - length of the bitmap array
**   goal_blk - the blk to start looking from
**
** Returns:
**   Position of free block, BFITNOENT if there are no free bits.
**
*******************************************************************************
******************************************************************************/

uint64 gfs_bitfit(char *buffer, uint buflen, uint64 goal_blk)
{
  int i = 0, j;
  uint64 goal;

  goal = ((goal_blk / GFS_NBBY) > buflen) ? buflen : (goal_blk / GFS_NBBY);

  IFDEBUG(DEBUG_BITS1,
	  printk(DEBUG_HOW "Entering gfs_bitfit()\n"););

  for (j = goal; j < buflen; j++)
  {
    /* if the byte is completely filled continue */

    if ((buffer[j] & 0x55) == 0x55)
      continue;

    /* this byte has a free bit in it, find it. */
    
    for (i = 0; i < GFS_NBBY; i++)
      if (!TSTBIT(buffer[j], i * (8 / GFS_NBBY)))
	break;
    
    if (i < GFS_NBBY)
      break;
  }

  if (j == buflen){
    for (j = 0; j < goal; j++){
      /* if the byte is completely filled continue */
      if ((buffer[j] & 0x55) == 0x55)
	continue;
      /* this byte has a free bit in it, find it. */
      for (i = 0; i < GFS_NBBY; i++)
	if (!TSTBIT(buffer[j], i * (8 / GFS_NBBY)))
	  break;
      if (i < GFS_NBBY)
	break;
    }
    IFDEBUG(DEBUG_BITS1,
	    printk(DEBUG_HOW "Exiting gfs_bitfit()\n"););
    if (j ==  goal){
      return(BFITNOENT);
    } else {
      return(j * GFS_NBBY + i);
    }
  }

  IFDEBUG(DEBUG_BITS1,
	  printk(DEBUG_HOW "Exiting gfs_bitfit()\n"););

  return(j * GFS_NBBY + i);
}


/******************************************************************************
*******************************************************************************
**
** uint64 gfs_8bitfit()
**
** Description:
**   This routine searches for a bytes worth or free blocks. 
**
** Input(s):
**   buffer - pointer to the bitmap array
**   buflen - length of the bitmap array
**   goal_blk - the blk to start looking from
**
** Returns:
**   Position of the first free block, BFITNOENT if there are no free bytes.
**
*******************************************************************************
******************************************************************************/

static uint64 gfs_8bitfit(char *buffer, uint buflen, uint64 goal_blk)
{
  int j;
  uint64 goal;

  goal = ((goal_blk / GFS_NBBY) > buflen) ? buflen : (goal_blk / GFS_NBBY);

  IFDEBUG(DEBUG_BITS1,
	  printk(DEBUG_HOW "Entering gfs_8bitfit()\n"););

  for (j = goal; j < buflen; j++)
    if ((buffer[j] & 0x55) == 0x00)
      break;

  if (j == buflen){
    for (j = 0; j < goal; j++)
      if ((buffer[j] & 0x55) == 0x00)
	break;
  IFDEBUG(DEBUG_BITS1,
	  printk(DEBUG_HOW "Exiting gfs_8bitfit()\n"););
    if (j == goal){
      return(BFITNOENT);
    } else {
      return(j * GFS_NBBY);
    }	
  }

  IFDEBUG(DEBUG_BITS1,
	  printk(DEBUG_HOW "Exiting gfs_8bitfit()\n"););

  return(j * GFS_NBBY);

}


/******************************************************************************
*******************************************************************************
**
** uint64 gfs_64bitfit()
**
** Description:
**   This routine searches for a uint64s worth or free blocks. 
**
** Input(s):
**   buffer - pointer to the bitmap array
**   buflen - length of the bitmap array
**   goal_blk - the blk to start looking from
**
** Returns:
**   Position of the first free block, BFITNOENT if there are no free uint64s.
**
*******************************************************************************
******************************************************************************/

static uint64 gfs_64bitfit(char *buffer, uint buflen, uint64 goal_blk)
{
  int j;
  uint64 goal;

  goal = ((goal_blk / GFS_NBBY) > buflen) ? (buflen/sizeof(uint64))
					     : (goal_blk / GFS_NBBY)/sizeof(uint64);


  IFDEBUG(DEBUG_BITS1,
	  printk(DEBUG_HOW "Entering gfs_64bitfit()\n"););

  for (j = goal; j < buflen / sizeof(uint64); j++)
    if ((((uint64 *)buffer)[j] & 0x5555555555555555ll) == 0x0000000000000000ll)
      break;

  if (j >= buflen / sizeof(uint64)){  
    for (j = 0; j < goal; j++)
      if ((((uint64 *)buffer)[j] & 0x5555555555555555ll) == 0x0000000000000000ll)
	break;
  IFDEBUG(DEBUG_BITS1,
	  printk(DEBUG_HOW "Exiting gfs_64bitfit()\n"););
    if (j >= goal){
      return (BFITNOENT);
    } else {
      return (j * GFS_NBBY * sizeof(uint64));
    }
  }

  IFDEBUG(DEBUG_BITS1,
	  printk(DEBUG_HOW "Exiting gfs_64bitfit()\n"););

  return (j * GFS_NBBY * sizeof(uint64));
}


/******************************************************************************
*******************************************************************************
**
** gfs_blklist_t *gfs_firstfit()
**
** Description:
**   This routine allocates up to "len" blocks.  The routine will return a
**   lesser number if there isn't enough free space.
**
** Input(s):
**   buffer - pointer to the bitmap array
**   buflen - length of the bitmap array
**   goal_ptr - a ptr to an entry in the dinode structure that keeps track
**              of the last blk allocated.
**
** Returns:
**   A pointer to the list of blocks or a NULL if there were no free blocks.
**
*******************************************************************************
******************************************************************************/

gfs_blklist_t *gfs_firstfit(char *buffer, uint buflen,
			    uint len, uint64 *goal_blk)
{
  gfs_blklist_t *bhead = NULL, *blast = NULL;
  uint64 no;


  IFDEBUG(DEBUG_BITS1,
	  printk(DEBUG_HOW "Entering gfs_firstfit()\n"););


  /* Perform as many 64 bit fits as possible */

  while (len >= GFS_NBBY * sizeof(uint64))
  {
    if ((no = gfs_64bitfit(buffer, buflen, *goal_blk)) == BFITNOENT)
      break;
    
    if (bhead)
    {
      /*  Merge?  */

      if (blast->blk_no + blast->blk_len == no)
      {
	blast->blk_len += GFS_NBBY * sizeof(uint64);
      }
      else
      {
	blast->blk_next = (gfs_blklist_t *)gmalloc(sizeof(gfs_blklist_t));
	if (!blast->blk_next)
	  break;

	blast = blast->blk_next;
	blast->blk_no = no;
	blast->blk_len = GFS_NBBY * sizeof(uint64);
	blast->blk_next = NULL;
      }
    }
    else
    {
      bhead = (gfs_blklist_t *)gmalloc(sizeof(gfs_blklist_t));
      if (!bhead)
	break;

      bhead->blk_no = no;
      bhead->blk_len = GFS_NBBY * sizeof(uint64);
      bhead->blk_next = NULL;

      blast = bhead;
    }

    gfs_set_bits(buffer, buflen, no, GFS_NBBY * sizeof(uint64), 0);
    *goal_blk = no;
    len -= GFS_NBBY * sizeof(uint64);
  }

  /* Perform as many 8 bit fits as possible */

  while (len >= GFS_NBBY)
  {
    if ((no = gfs_8bitfit(buffer, buflen, *goal_blk)) == BFITNOENT)
      break;
    
    if (bhead)
    {
      /*  Merge?  */

      if (blast->blk_no + blast->blk_len == no)
      {
	blast->blk_len += GFS_NBBY;
      }
      else
      {
	blast->blk_next = (gfs_blklist_t *)gmalloc(sizeof(gfs_blklist_t));
	if (!blast->blk_next)
	  break;

	blast = blast->blk_next;
	blast->blk_no = no;
	blast->blk_len = GFS_NBBY;
	blast->blk_next = NULL;
      }
    }
    else
    {
      bhead = (gfs_blklist_t *)gmalloc(sizeof(gfs_blklist_t));
      if (!bhead)
	break;

      bhead->blk_no = no;
      bhead->blk_len = GFS_NBBY;
      bhead->blk_next = NULL;

      blast = bhead;
    }

    gfs_set_bits(buffer, buflen, no, GFS_NBBY, 0);
    *goal_blk = no;
    len -= GFS_NBBY;
  }
  
  /* Perform as many bit fits as possible */

  while (len)
  {
    if ((no = gfs_bitfit(buffer, buflen, *goal_blk)) == BFITNOENT)
      break;
    
    if (bhead)
    {
      /*  Merge?  */

      if (blast->blk_no + blast->blk_len == no)
      {
	blast->blk_len++;
      }
      else
      {
	blast->blk_next = (gfs_blklist_t *)gmalloc(sizeof(gfs_blklist_t));
	if (!blast->blk_next)
	  break;

	blast = blast->blk_next;
	blast->blk_no = no;
	blast->blk_len = 1;
	blast->blk_next = NULL;
      }
    }
    else
    {
      bhead = (gfs_blklist_t *)gmalloc(sizeof(gfs_blklist_t));
      if (!bhead)
	break;

      bhead->blk_no = no;
      bhead->blk_len = 1;
      bhead->blk_next = NULL;

      blast = bhead;
    }

    gfs_set_bits(buffer, buflen, no, 1, 0);
    *goal_blk = no;
    len--;
  }

  IFDEBUG(DEBUG_BITS1,
	  printk(DEBUG_HOW "Exiting gfs_firstfit()\n"););
  
  return(bhead);
}

