/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**    Portions Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name: blklist.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "bits.h"
#include "blklist.h"
#include "dio.h"
#include "file.h"
#include "glock.h"
#include "fs.h"


/******************************************************************************
*******************************************************************************
**
** gfs_blklist_t *gfs_blkalloc()
**
** Description:
**   This routine allocates "len" free blocks.  It either returns "len" blocks
**   or returns -ENOSPC.  There is no middle ground.
**   
**   NOTE that this routine tries to be nice and skip over resource groups it 
**   can't get Glocks to.  This means that a mostly full filesystem will
**   sometimes look completely full if it's heavily loaded.
**
** Input(s):
**   ip - inode that we are allocating blocks for
**   len - number of blocks to allocate
**   is_meta - whether or not this is meta data (1) or data (0)
**
** Output(s):
**   blks - the list of blocks allocated
**            (This list must be freed by the caller.)
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_blkalloc(gfs_inode_t *ip, uint32 len, gfs_blklist_t **blks,
		 int is_meta)
{
  gfs_rgrpd_t *rgd;
  gfs_blklist_t *bhead = NULL, *blast = NULL, *btmp;
  gfs_sbd_t *sdp = ip->i_sbd;
  
  uint32 set = 0, rgset;
  int i, error = 0;
  int buf;

  uint32 *rgrp_data;
  uint64 *last_data;

  IFDEBUG(DEBUG_BLKLIST1,
	  printk(DEBUG_HOW "Entering gfs_blkalloc()\n"););


  /*  Start looking in the same resource group the dinode is in.  */

  if(is_meta){
    rgrp_data = &(ip->i_di.di_rgrp_meta);
    last_data = &(ip->i_di.di_last_meta);
  } else {
    rgrp_data = &(ip->i_di.di_rgrp_data);
    last_data = &(ip->i_di.di_last_data);
  }
  IFDEBUG(DEBUG_BLKLIST3,
         printk(DEBUG_HOW "\trgrp_data = %u\n\tlast_data = %"FMT_64"u\n",
		*rgrp_data, *last_data););
  ASSERT(*rgrp_data < sdp->sd_sb.sb_rgrps);

  for (i = 0; i < sdp->sd_sb.sb_rgrps; i++)
  {
    rgd = &sdp->sd_rgd[*rgrp_data];

    if ((error = gfs_glock(sdp, rgd->rd_ri.ri_glock, NULL, 0)))
    {
      *rgrp_data = (*rgrp_data + 1) % sdp->sd_sb.sb_rgrps;
      continue;
    }

    /*  Read/Reread the RG block  */

    if (gfs_rgrp_read(sdp, rgd))
    {
      gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);
      *rgrp_data = (*rgrp_data + 1) % sdp->sd_sb.sb_rgrps;	
      continue;
    }

    /*  Check to see if the RG is full  */

    if (rgd->rd_rg.rg_free == 0)
    {
      gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);
      *rgrp_data = (*rgrp_data + 1) % sdp->sd_sb.sb_rgrps;
      continue;
    }

    rgset = 0;

    for (buf = 0; buf < rgd->rd_ri.ri_length; buf++)
    {
      /*  Allocate blocks  */

      btmp = gfs_firstfit(rgd->rd_bits[buf].bi_bitmap,
			  rgd->rd_bits[buf].bi_len,
			  len - (set + rgset),
			  last_data);

      if (bhead == NULL)
	bhead = btmp;
      else
	blast->blk_next = btmp;

      for ( ; btmp; blast = btmp, btmp = btmp->blk_next)
      {
	rgset += btmp->blk_len;
	btmp->blk_no += rgd->rd_ri.ri_data1 + rgd->rd_bits[buf].bi_start * GFS_NBBY;
      }

      if (len - (set + rgset) == 0)
	break;
    }
      
    /*  We allocated some blocks, now write back
	the resource group header  */

    if (rgset)
    {
      set += rgset;

      rgd->rd_rg.rg_free -= rgset;
      rgd->rd_rg.rg_time = CURRENT_TIME;
      
      if (gfs_rgrp_write(sdp, rgd))
      {
	gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);
	gfs_blklist_free(bhead);
	*rgrp_data = (*rgrp_data + 1) % sdp->sd_sb.sb_rgrps;	
	continue;
      }
      gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);
    }
    else
      gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);

      
    if (set == len)
    {
      /*  We've allocated enough blocks.  */

      IFDEBUG(DEBUG_BLKLIST3,
	      for (btmp = bhead; btmp; btmp = btmp->blk_next)
		printk(DEBUG_HOW "  gfs_blklist_t = <%.16lX>, no = %"FMT_64"u, len = %u, next = <%.16lX>\n",
		       (unsigned long)btmp, btmp->blk_no, btmp->blk_len, (unsigned long)btmp->blk_next););

      *blks = bhead;

      IFDEBUG(DEBUG_BLKLIST1,
	      printk(DEBUG_HOW "Exiting gfs_blkalloc()\n"););

      return(0);
    }      

    *rgrp_data = (*rgrp_data + 1) % sdp->sd_sb.sb_rgrps;
  }

  /*  If we get to this point, it means we didn't get enough blocks
      to satisfy the request.  Free what we've allocated so far and
      exit.  */

  if (error == 0)
  {
    printk(KERN_ERR "GFS:  out of free data blocks\n");
    error = -ENOSPC;
  }
  
  if (gfs_blkfree(sdp, &bhead))
    GARBAGE_BLOCKS();
  gfs_blklist_free(bhead);

  *blks = NULL;

  IFDEBUG(DEBUG_BLKLIST1,
	  printk(DEBUG_HOW "Exiting gfs_blkalloc()\n"););

  return(error);
}


/******************************************************************************
*******************************************************************************
**
** gfs_blklist_t *gfs_blkdalloc()
**
** Description:
**   This routine tries to allocate one block as a dinode.
**
**   See the note from gfs_blkalloc()
**
** Input(s):
**   dip - the directory that will contain this new dinode
**   rgrp_pref - which resource group is preferred (F's == no pref)
**   blk_pref - which blk in the resource group is preferred
**   is_dir - Is this blk going to be used for a directory
**
** Output(s):
**   blks - a gfs_blklist_t structure containing the blocknumber of the
**            new dinode  (Must be freed by the caller)
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_blkdalloc(gfs_inode_t *dip, gfs_blklist_t **blks, uint32 rgrp_pref,
		  uint64 blk_pref, int is_dir)
{
  gfs_rgrpd_t *rgd;
  gfs_sbd_t *sdp = dip->i_sbd;
  
  uint64 no = BFITNOENT;
  int i, error = 0;
  int buf;

  uint32 rgrp_meta;
  uint64 last_meta;

  IFDEBUG(DEBUG_BLKLIST1,
	  printk(DEBUG_HOW "Entering gfs_blkdalloc()\n"););


  /*  Start looking in the parent directories resource group  */

  if(!is_dir){
  IFDEBUG(DEBUG_BLKLIST3,
         printk(DEBUG_HOW "\n\nDinode is not a dir...\n"););
    if((rgrp_pref == 0xFFFFFFFF) && (blk_pref == 0xFFFFFFFFFFFFFFFFll)){
      rgrp_meta = dip->i_di.di_rgrp_meta;
      last_meta = dip->i_di.di_last_meta;
      /* doesn't matter if [rgrp last]_data are garbage for a dir */
    } else if((rgrp_pref == 0xFFFFFFFF) || (blk_pref == 0xFFFFFFFFFFFFFFFFll)){
      if(rgrp_pref == 0xFFFFFFFF){
       rgrp_meta = dip->i_di.di_rgrp_meta;
       last_meta = blk_pref;
      } else {
       rgrp_meta = rgrp_pref;
       last_meta = dip->i_di.di_last_meta;
      }
    } else {
      rgrp_meta = rgrp_pref;
      last_meta = blk_pref;
    }
  } else {
  IFDEBUG(DEBUG_BLKLIST3,
         printk(DEBUG_HOW "\n\nDinode is a dir...\n"););
    if((rgrp_pref == 0xFFFFFFFF) && (blk_pref == 0xFFFFFFFFFFFFFFFFll)){
      rgrp_meta = gfs_randrange(0, sdp->sd_sb.sb_rgrps);
      last_meta = gfs_randrange(0, sdp->sd_rgd[rgrp_meta].rd_ri.ri_data);
    } else if((rgrp_pref == 0xFFFFFFFF) || (blk_pref == 0xFFFFFFFFFFFFFFFFll)){
      if(rgrp_pref == 0xFFFFFFFF){
       rgrp_meta = gfs_randrange(0, sdp->sd_sb.sb_rgrps);
       last_meta = blk_pref;
      } else {
       rgrp_meta = rgrp_pref;
       last_meta = gfs_randrange(0, sdp->sd_rgd[rgrp_meta].rd_ri.ri_data);
      }
    } else {
      rgrp_meta = rgrp_pref;
      last_meta = blk_pref;
    }
  }
 
  IFDEBUG(DEBUG_BLKLIST3,
	  printk(DEBUG_HOW "\trgrp_meta = %u\n\tlast_meta = %u\n",
		 rgrp_meta, (u_int)last_meta););

  ASSERT(rgrp_meta < sdp->sd_sb.sb_rgrps);
 
  for (i = 0; i < sdp->sd_sb.sb_rgrps; i++)
  {
    rgd = &sdp->sd_rgd[rgrp_meta];

    if ((error = gfs_glock(sdp, rgd->rd_ri.ri_glock, NULL, 0)))
    {
      rgrp_meta = (rgrp_meta + 1) % sdp->sd_sb.sb_rgrps;
      continue;
    }

    /*  Read/Reread the RG block  */

    if (gfs_rgrp_read(sdp, rgd))
    {
      gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);
      rgrp_meta = (rgrp_meta + 1) % sdp->sd_sb.sb_rgrps;	
      continue;
    }

    /*  Check to see if the RG is full  */

    if (rgd->rd_rg.rg_free == 0)
    {
      gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);
      rgrp_meta = (rgrp_meta + 1) % sdp->sd_sb.sb_rgrps;
      continue;
    }
      
    for (buf = 0; buf < rgd->rd_ri.ri_length; buf++)
    {
      /*  Look for a block  */

      no = gfs_bitfit(rgd->rd_bits[buf].bi_bitmap,
		      rgd->rd_bits[buf].bi_len, last_meta);
      if (no != BFITNOENT)
	break;
    }

    if (buf == rgd->rd_ri.ri_length)
    {
      gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);
      rgrp_meta = (rgrp_meta + 1) % sdp->sd_sb.sb_rgrps;
      continue;
    }

    /*  Set the block and setup the gfs_blklist_t structure  */

    *blks = (gfs_blklist_t *)gmalloc(sizeof(gfs_blklist_t));
    if (!*blks)
    {
      gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);
      error = -ENOMEM;
      rgrp_meta = (rgrp_meta + 1) % sdp->sd_sb.sb_rgrps;
      continue;
    }

    gfs_set_bits(rgd->rd_bits[buf].bi_bitmap, rgd->rd_bits[buf].bi_len,
		 no, 1, TRUE);

    (*blks)->blk_no = no + rgd->rd_ri.ri_data1 + rgd->rd_bits[buf].bi_start * GFS_NBBY;
    (*blks)->blk_len = 1;
    (*blks)->blk_next = NULL;

    /*  Save back the resource group header  */
    rgd->rd_rg.rg_free--;
    rgd->rd_rg.rg_dinodes++;
    rgd->rd_rg.rg_time = CURRENT_TIME;
      
    if (gfs_rgrp_write(sdp, rgd))
    {
      gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);
      gfs_blklist_free(*blks);
      rgrp_meta = (rgrp_meta + 1) % sdp->sd_sb.sb_rgrps;	
      continue;
    }
    gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);    

    IFDEBUG(DEBUG_BLKLIST1,
	    printk(DEBUG_HOW "Exiting gfs_blkdalloc()\n"););

    return(0);
  }

  /*  Couldn't get one lousy block...  */

  if (error == 0)
  {
    printk(KERN_ERR "GFS:  out of free dinode blocks\n");
    error = -ENOSPC;
  }
  
  *blks = NULL;

  IFDEBUG(DEBUG_BLKLIST1,
	  printk(DEBUG_HOW "Exiting gfs_blkdalloc()\n"););

  return(error);
}


/******************************************************************************
*******************************************************************************
**
** int sort_by_rgrp()
**
** Description:
**   Sort a list of blocks by resource group so that freeing them is more
**   efficient.
**
** Input(s):
**   sdp - the filesystem that these blocks belong to
**   blks - a unsorted list of blocks
**
** Returns:
**   An array of gfs_blklist_t lists.  Each element corresponds to the 
**   the blocks in that resource group.  
**
**   Note that the array must be freed by the caller routine.
**
**   Returns NULL if there is no memory to allocate the array.
**
*******************************************************************************
******************************************************************************/

static gfs_blklist_t **sort_by_rgrp(gfs_sbd_t *sdp, gfs_blklist_t *blks)
{
  gfs_blklist_t **ba, *btmp;
  uint32 x;


  IFDEBUG(DEBUG_BLKLIST1,
	  printk(DEBUG_HOW "Entering sort_by_rgrp()\n"););

  
  /*  Get memory for the array  */

  ba = (gfs_blklist_t **)gmalloc(sdp->sd_sb.sb_rgrps * sizeof(gfs_blklist_t *));
  if (!ba)
  {
    printk(KERN_ERR "GFS:  sort_by_rgrp() out of memory\n");

    IFDEBUG(DEBUG_BLKLIST1,
	    printk(DEBUG_HOW "Exiting sort_by_rgrp()\n"););

    return(ba);
  }
  memset(ba, 0, sdp->sd_sb.sb_rgrps * sizeof(gfs_blklist_t *));


  while (blks)
  {
    btmp = blks->blk_next;
    
    /*  Figure out what resource group these blocks are in.  We don't have
	to worry about the "blk_len" structure member because no single 
	gfs_blklist_t structure can hold blocks from two different
	resource groups.  */

    for (x = 0; x < sdp->sd_sb.sb_rgrps; x++)
      if (blks->blk_no < sdp->sd_rgd[x].rd_ri.ri_data1 + sdp->sd_rgd[x].rd_ri.ri_data)
	break;

    ASSERT(x < sdp->sd_sb.sb_rgrps);

    blks->blk_next = ba[x];
    ba[x] = blks;    

    blks = btmp;
  }


  IFDEBUG(DEBUG_BLKLIST1,
	  printk(DEBUG_HOW "Exiting sort_by_rgrp()\n"););

  return(ba);
}


/******************************************************************************
*******************************************************************************
**
** int gfs_blksfree()
**
** Description:
**   This routine deallocates a list of blocks.
**
**   If the routine isn't able to free blocks for some reason (most likely
**   deadlock or a failed initiator), it returns the list of blocks that
**   weren't freed.
**
** Input(s):
**   sdp - the filesystem these blocks live on
**   blks - the list of blocks to free
**
** Output(s):
**   blks - a list of blocks that couldn't be freed.
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_blkfree(gfs_sbd_t *sdp, gfs_blklist_t **blks)
{
  gfs_blklist_t **ba, *btmp;
  gfs_blklist_t *bhead = NULL, *blast = NULL;
  gfs_rgrpd_t *rgd;
  gfs_bitmap_t *bi;

  int dinodes;
  uint32 rgrp, buf;
  int error;
  int len;


  IFDEBUG(DEBUG_BLKLIST1,
	  printk(DEBUG_HOW "Entering gfs_blkfree()\n"););


  if (!blks)
  {
    IFDEBUG(DEBUG_BLKLIST1,
	    printk(DEBUG_HOW "Exiting gfs_blkfree()\n"););
    return(0);
  }

  ba = sort_by_rgrp(sdp, *blks);
  if (!ba)
  {
    IFDEBUG(DEBUG_BLKLIST1,
	    printk(DEBUG_HOW "Exiting gfs_blkfree()\n"););
    return(-ENOMEM);
  }


  for (rgrp = 0; rgrp < sdp->sd_sb.sb_rgrps; rgrp++)
  {
    if (!ba[rgrp])
      continue;

      IFDEBUG(DEBUG_BLKLIST3,
	      printk(DEBUG_HOW "  rgrp = %u\n", rgrp);
	      for (btmp = ba[rgrp]; btmp; btmp = btmp->blk_next)
		printk(DEBUG_HOW "    gfs_blklist_t = <%.16lX>, no = %"FMT_64"u, len = %u, next = <%.16lX>\n",
		       (unsigned long)btmp, btmp->blk_no, btmp->blk_len, (unsigned long)btmp->blk_next););

    rgd = &sdp->sd_rgd[rgrp];

    if ((error = gfs_glock(sdp, rgd->rd_ri.ri_glock, NULL, 0)))
      continue;

    /*  Read/Reread the RG block  */

    if (gfs_rgrp_read(sdp, rgd))
    {
      gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);
      continue;
    }

    for (btmp = ba[rgrp]; btmp; btmp = btmp->blk_next)
    {
      btmp->blk_no -= rgd->rd_ri.ri_data1;

      for (buf = 0; buf < rgd->rd_ri.ri_length; buf++)
      {
	bi = &rgd->rd_bits[buf];

	if (btmp->blk_no < (bi->bi_start + bi->bi_len) * GFS_NBBY)
	{
	  len = (btmp->blk_no + btmp->blk_len) - (bi->bi_start + bi->bi_len) * GFS_NBBY;

	  if (len <= 0)
	  {
	    dinodes = gfs_clr_bits(bi->bi_bitmap, bi->bi_len,
				  btmp->blk_no - bi->bi_start * GFS_NBBY, btmp->blk_len);
	    rgd->rd_rg.rg_free += btmp->blk_len;
	    rgd->rd_rg.rg_dinodes -= dinodes;

	    break;
	  }
	  else
	  {
	    dinodes = gfs_clr_bits(bi->bi_bitmap, bi->bi_len,
				   btmp->blk_no - bi->bi_start * GFS_NBBY, btmp->blk_len - len);
	    rgd->rd_rg.rg_free += btmp->blk_len - len;
	    rgd->rd_rg.rg_dinodes -= dinodes;

	    btmp->blk_no = (bi->bi_start + bi->bi_len) * GFS_NBBY;
	    btmp->blk_len = len;
	  }
	}
      }

      ASSERT(buf != rgd->rd_ri.ri_length);
    }

    /*  Write back rgrp  */

    rgd->rd_rg.rg_time = CURRENT_TIME;

    if (gfs_rgrp_write(sdp, rgd))
    {
      gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);
      continue;
    }
    gfs_gunlock(sdp, rgd->rd_ri.ri_glock, 0);
    gfs_blklist_free(ba[rgrp]);
    ba[rgrp] = NULL;
  }


  /*  Package up any blklist_t that we weren't able to free  */

  for (rgrp = 0; rgrp < sdp->sd_sb.sb_rgrps; rgrp++)
    if (ba[rgrp])
    {
      if (bhead)
	blast->blk_next = btmp = ba[rgrp];
      else
	bhead = btmp = ba[rgrp];

      for ( ; btmp; blast = btmp, btmp = btmp->blk_next) ;
    }


  gfree(ba, sdp->sd_sb.sb_rgrps * sizeof(gfs_blklist_t *));

  *blks = bhead;

  IFDEBUG(DEBUG_BLKLIST1,
	  printk(DEBUG_HOW "Exiting gfs_blkfree()\n"););

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** int gfs_blklist_free()
**
** Description:
**   Free all system memory associated with a block list.  Don't do anything
**   to the disk.
**
** Input(s):
**   blks - the list to be freed
**
*******************************************************************************
******************************************************************************/

void gfs_blklist_free(gfs_blklist_t *blks)
{
  gfs_blklist_t *btmp;


  IFDEBUG(DEBUG_BLKLIST1,
	  printk(DEBUG_HOW "Entering gfs_blklist_free()\n"););


  while (blks)
  {
    btmp = blks->blk_next;
    gfree(blks, sizeof(gfs_blklist_t));
    blks = btmp;
  }


  IFDEBUG(DEBUG_BLKLIST1,
	  printk(DEBUG_HOW "Exiting gfs_blklist_free()\n"););
}

