/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@lcse.umn.edu>
** 
**    Module name: fs.c
**
**    Description:
**      Miscellaneous filesystem routines
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "bits.h"
#include "dio.h"
#include "dir.h"
#include "glock.h"
#include "fs.h"
#include "inode.h"


/******************************************************************************
*******************************************************************************
**
** gfs_ino2gl()
**
** Description:
**   Converts an inode number into its corresponding glock number.
**
** Input(s):
**   sdp - the filesystem
**   ino - the inode number
**
** Returns:
**   Glock number
**
*******************************************************************************
******************************************************************************/

uint64 gfs_ino2gl(gfs_sbd_t *sdp, uint64 ino)
{
  uint rgrp;
  gfs_rindex_t *ri = NULL;
  uint64 gl;

  ENTER(GFS_INO2GL, DEBUG_FS1);

  ASSERT(ino < sdp->sd_sb.sb_fssize);

  if (ino == sdp->sd_sb.sb_rootdi)
    {
      EXIT(GFS_INO2GL, DEBUG_FS1);
      return(GFS_ROOT_LOCK);
    }


  for (rgrp = 0; rgrp < sdp->sd_sb.sb_rgrps; rgrp++)
    {
      ri = &sdp->sd_rgd[rgrp].rd_ri;
      if (ino < ri->ri_data1 + ri->ri_data)
	break;
    }

  gl = (ino % ri->ri_glocks) + ri->ri_glock1;

  EXIT(GFS_INO2GL, DEBUG_FS1);
  return(gl);
}


/******************************************************************************
*******************************************************************************
**
** vnode_t *gfs_ihashlookup()
**
** Description:
**   This routine is the inode hash table lookup routine.
**
** Input(s):
**   sdp - the filesystem
**   ino - inode number
**
** Returns:
**   A pointer to the inode with number "ino", or NULL on failure
**
*******************************************************************************
******************************************************************************/

gfs_inode_t *gfs_ihashlookup(gfs_sbd_t *sdp, uint64 ino)
{
  uint32 hashoff;
  gfs_inode_t *ip = NULL;
  gfs_list_head_t *tmp;

  ENTER(GFS_IHASHLOOKUP, DEBUG_FS1);
 
  hashoff = ino % GFS_IHASH_SIZE;
  
  GFS_SEMA_DOWN(&sdp->sd_ihash_lock[hashoff]);

  for (tmp = sdp->sd_ihash[hashoff].next; tmp != &sdp->sd_ihash[hashoff]; tmp = tmp->next)
  {
    ip = gfs_list_entry(tmp, gfs_inode_t, i_list);
    if (ip->i_num == ino)
      break;
    else
      ip = NULL;
  }

  GFS_SEMA_UP(&sdp->sd_ihash_lock[hashoff]);

  EXIT(GFS_IHASHLOOKUP, DEBUG_FS1);
  return(ip);
}


/******************************************************************************
*******************************************************************************
**
** int gfs_ihashadd()
**
** Description:
**   This routine adds an entry to the inode hash table.
**
** Input(s):
**   ip - the inode to add to the table
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_ihashadd(gfs_inode_t *ip)
{
  uint32 hashoff;
  gfs_sbd_t *sdp = ip->i_sbd;

  ENTER(GFS_IHASHADD, DEBUG_FS1);
 
  hashoff = ip->i_num % GFS_IHASH_SIZE;

  GFS_SEMA_DOWN(&sdp->sd_ihash_lock[hashoff]);

  gfs_list_add(&ip->i_list, &sdp->sd_ihash[hashoff]);

  GFS_SEMA_UP(&sdp->sd_ihash_lock[hashoff]);

  EXIT(GFS_IHASHADD, DEBUG_FS1);

  return (0);
}


/******************************************************************************
*******************************************************************************
**
** int gfs_ihashdel()
**
** Description:
**   This routine deletes an entry from the inode hash table.
**
** Input(s):
**   ip - the inode to delete
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_ihashdel(gfs_inode_t *ip)
{
  gfs_sbd_t *sdp;
  uint32 hashoff;
  gfs_inode_t *tip = NULL;
  gfs_list_head_t *tmp;

  ENTER(GFS_IHASHDEL, DEBUG_FS1);

  if (ip == NULL)
    {
     
      EXIT(GFS_IHASHDEL, DEBUG_FS1);
      return(-1);
    }

  sdp = ip->i_sbd;
  hashoff = ip->i_num % GFS_IHASH_SIZE;

  GFS_SEMA_DOWN(&sdp->sd_ihash_lock[hashoff]);

  for (tmp = sdp->sd_ihash[hashoff].next; tmp != &sdp->sd_ihash[hashoff]; tmp = tmp->next)
    {
      tip = gfs_list_entry(tmp, gfs_inode_t, i_list);
      if (tip == ip)
	break;
    }
  
  if (tip == ip)
    gfs_list_del(&ip->i_list);
  
  GFS_SEMA_UP(&sdp->sd_ihash_lock[hashoff]);
     
  EXIT(GFS_IHASHDEL, DEBUG_FS1);
  return(-1);
}


/******************************************************************************
*******************************************************************************
**
** static gfs_blklist_t *indirfree()
**
** Description:
**   This routine builds a list of blocks to be freed when this file is
**   removed.  It's called repeatedly by freefile()
**
** Input(s):
**   ip - the file being erased
**   indirect - the number of a indirect block
**   height - the height of the indirect block in the metatdata tree
**   bytes - max bytes to truncate to
** Returns:
**   A list of gfs_blklist_t structures or NULL on failure
**
*******************************************************************************
******************************************************************************/

/* I could look to try to reduce indirection, but I didn't. If somebody wants to      */
/* write a ten terabyte file and then truncate it to eight bytes, that's their fault. */
/* If they don't truncate it to zero, they'll probably write to it again anyway       */

static gfs_blklist_t *indirfree(gfs_inode_t *ip, uint64 indirect, uint height, uint64 bytes, uint64 *blocks_freedp)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_blklist_t *bhead = NULL, *blast = NULL, *btmp;
  buffer_head_t *bh;

  uint64 *bnp, bn, blocks_keep, left_over, bytes_pptr;
  uint64 blocks_freed = 0, tmp;
  int i;

  ENTER(INDIRFREE, DEBUG_FS1);
 
  if (height == 0)  /*  Stuffed Dinode  */
    {
      if (blocks_freedp)
	*blocks_freedp = 0;

      EXIT(INDIRFREE, DEBUG_FS1);
      return(NULL);
    }

  bh = gfs_dread(sdp, indirect, ip->i_di.di_glock);
 
  /*bh = gfs_dread(sdp, indirect, ip->i_di.di_glock);*/ 

  if (!bh)
    {

      if (blocks_freedp)
	*blocks_freedp = blocks_freed;

      GARBAGE_BLOCKS();
      EXIT(INDIRFREE, DEBUG_FS1);
      return(NULL);
    }

  bytes_pptr = sdp->sd_sb.sb_bsize;
  
  for (i=1; i < height; i++)
    bytes_pptr *= sdp->sd_inptrs;

  /* eventually we'll want to change all this to bit operations */

  blocks_keep = bytes / bytes_pptr;
  left_over = bytes & (bytes_pptr -1);

  if (height == 1)  /*  Direct Blocks  */
    {
      for (i = blocks_keep; i < sdp->sd_inptrs; i++)
	{
	  bnp = &((uint64 *)(BH_DATA(bh)))[i];
	  bn = be64_to_cpu(*bnp);

	  if (bn == 0)
	    continue;

	  if (i == blocks_keep && left_over) {
	    buffer_head_t *bh2;
	   
	    ASSERT(left_over <= sdp->sd_sb.sb_bsize);
    	    bh2 = gfs_dread(sdp, bn, ip->i_di.di_glock);

   	    memset(BH_DATA(bh2) + left_over, 0, sdp->sd_sb.sb_bsize - left_over);

	    gfs_dwrite(bh2, DW_NOKEEP);
	    continue;
	  }

	  if (bhead && bhead->blk_no + bhead->blk_len == bn)
	    bhead->blk_len++;  /*  Merge  */
	  else
	    {
	      btmp = (gfs_blklist_t *)gmalloc(sizeof(gfs_blklist_t));
	     
	      if (!btmp)
		{
		  GARBAGE_BLOCKS();
		  continue;
		}
	
	      btmp->blk_no = bn;
	      btmp->blk_len = 1;
	      btmp->blk_next = bhead;
	      bhead = btmp;
	    }

	  *bnp = 0;
	  blocks_freed++;
	}
    }
  else  /*  Indirect Blocks  */
    {
      for (i = blocks_keep; i < sdp->sd_inptrs; i++)
	{
	  bnp = &((uint64 *)(BH_DATA(bh)))[i];
	  bn = be64_to_cpu(*bnp);

	  if (bn == 0)
	    continue;

	  /* there shouldn't be too many recursive calls */
	  /* so the stack should be fine                 */

	  if (i == blocks_keep)
	    btmp = indirfree(ip, bn, height - 1, left_over, &tmp);
	  else
	    btmp = indirfree(ip, bn, height - 1, 0, &tmp);

	  blocks_freed += tmp;

	  if (bhead)
	    blast->blk_next = btmp;
	  else {
	    bhead = btmp;
	    for ( ; btmp; blast = btmp, btmp = btmp->blk_next) ;
	  }

	  if (i == blocks_keep && left_over)
	    continue;

	  btmp = (gfs_blklist_t *)gmalloc(sizeof(gfs_blklist_t));
       	 
	  if (!btmp)
	    {
	      GARBAGE_BLOCKS();
	      continue;
	    }

	  btmp->blk_no = bn;
	  btmp->blk_len = 1;
	  btmp->blk_next = bhead;
	  bhead = btmp;

	  *bnp = 0;
	  blocks_freed++;


	  for ( ; btmp; blast = btmp, btmp = btmp->blk_next) ;
	}
    }

  gfs_dwrite(bh, DW_NOKEEP);

  if (blocks_freedp)
    *blocks_freedp = blocks_freed;

  EXIT(INDIRFREE, DEBUG_FS1);
  return(bhead);
}


/******************************************************************************
*******************************************************************************
**
** static int freefile()
**
** Description:
**   This routine creates a "to free" list of the file's data and/or metadata.
**
** Input(s):
**   ip - the file being erased
**
** Returns:
**   A list of gfs_blklist_t structures or NULL on failure
**
*******************************************************************************
******************************************************************************/

static gfs_blklist_t *freefile(gfs_inode_t *ip, uint64 bytes, uint64 *blocks_freedp)
{
  gfs_sbd_t *sdp = ip->i_sbd;
  gfs_blklist_t *bhead = NULL, *blast = NULL, *btmp;

  uint64 *bnp, bn, bytes_pptr, left_over; 
  uint64 tmp, blocks_freed = 0;
  uint32 blocks_keep;
  int i;

  ENTER(FREEFILE, DEBUG_FS1);



  if (ip->i_di.di_height == 0)  /*  Stuffed Dinode  */
    {
      if (blocks_freedp)
	*blocks_freedp = blocks_freed;

      EXIT(FREEFILE, DEBUG_FS1);
      return(NULL);
    }

  bytes_pptr = sdp->sd_sb.sb_bsize;
  
  for (i=1; i < ip->i_di.di_height; i++)
    bytes_pptr *= sdp->sd_inptrs;

  /* eventually we'll want to change all this to bit operations */

  blocks_keep = bytes / bytes_pptr;
  left_over = bytes & (bytes_pptr-1);

  if (ip->i_di.di_height == 1)  /*  Direct Blocks  */
    {
      for (i = blocks_keep; i < sdp->sd_diptrs; i++)
	{
	  bnp = &((uint64 *)(BH_DATA(ip->i_didata) + sizeof(gfs_dinode_t)))[i];
	  bn = be64_to_cpu(*bnp);

	  if (bn == 0)
	    continue;

	  if (i == blocks_keep && left_over) {
	    buffer_head_t *bh;
	
	    ASSERT(left_over <= sdp->sd_sb.sb_bsize);
     	    bh = gfs_dread(sdp, bn, ip->i_di.di_glock);

	    memset(BH_DATA(bh) + left_over, 0, sdp->sd_sb.sb_bsize - left_over);

	    gfs_dwrite(bh, DW_NOKEEP);
	    continue;
	  }

      	  if (bhead && bhead->blk_no + bhead->blk_len == bn)
	    bhead->blk_len++;  /*  Merge  */
	  else
	    {
	      btmp = (gfs_blklist_t *)gmalloc(sizeof(gfs_blklist_t));
	     
	      if (!btmp)
		{
		  GARBAGE_BLOCKS();
		  continue;
		}
	
	      btmp->blk_no = bn;
	      btmp->blk_len = 1;
	      btmp->blk_next = bhead;
	      bhead = btmp;
	    }

	  *bnp = 0;
	  blocks_freed++;
	
	}
      
    }
  else  /*  Indirect Blocks  */
    {
      for (i = blocks_keep; i < sdp->sd_diptrs; i++)
	{
	  bnp = &((uint64 *)(BH_DATA(ip->i_didata) + sizeof(gfs_dinode_t)))[i];
	  bn = be64_to_cpu(*bnp);

	  if (bn == 0)
	    continue;

	  if (i == blocks_keep)
	    btmp = indirfree(ip, bn, ip->i_di.di_height - 1, left_over, &tmp);
	  else
	    btmp = indirfree(ip, bn, ip->i_di.di_height - 1, 0, &tmp);

	  blocks_freed += tmp;

	  IFDEBUG(DEBUG_FS3,
		  gfs_blklist_t *q;
		  for (q = btmp; q; q = q->blk_next)
		  {
		    printk(DEBUG_HOW "  block structure <0x%.16lX>\n", (unsigned long)q);
		    printk(DEBUG_HOW "    blk_no = %"FMT_64"u\n", q->blk_no);
		    printk(DEBUG_HOW "    blk_len = %u\n", q->blk_len);
		    printk(DEBUG_HOW "    blk_next = <0x%.16lX>\n", (unsigned long)q->blk_next);
		  });

	  if (bhead)
	    blast->blk_next = btmp;
	  else {
	    bhead = btmp;
	    for ( ; btmp; blast = btmp, btmp = btmp->blk_next) ;
	  }

	  if (i == blocks_keep && left_over)
	    continue;

	  btmp = (gfs_blklist_t *)gmalloc(sizeof(gfs_blklist_t));
	 
	  if (!btmp)
	    {
	      GARBAGE_BLOCKS();
	      continue;
	    }

	  btmp->blk_no = bn;
	  btmp->blk_len = 1;
	  btmp->blk_next = bhead;
	  bhead = btmp;

	  *bnp = 0;
	  blocks_freed++;
	  
	  /* if we actually start getting large files we might want */
	  /* if (blast)                                             */
	  /*    for (btmp = blast; btmp; blast = btmp, btmp = btmp->blk_next) ; */

	  for ( ; btmp; blast = btmp, btmp = btmp->blk_next) ;
	}
    }

  if (blocks_freedp)
    *blocks_freedp = blocks_freed;

  EXIT(FREEFILE, DEBUG_FS1);
  return(bhead);
}


/******************************************************************************
*******************************************************************************
**
** gfs_difree()
**
** Description:
**   This routine returns a list of that need to be freed to erase this
**     dinode.
**
** Input(s):
**   ip - inode to be deleted.
**
** Returns:
**   A list of gfs_blklist_t structures or NULL on failure
**
*******************************************************************************
******************************************************************************/

gfs_blklist_t *gfs_difree(gfs_inode_t *ip)
{
  gfs_blklist_t *bhead = NULL, *blast = NULL, *btmp;

  ENTER(GFS_DIFREE, DEBUG_FS1);
 
  /*  Create free list for ExHash leaf blocks, if this is a dirctory  */

  if (ip->i_di.di_type == FILE_DIR)
    {
      bhead = gfs_dir_free(ip);

      for (btmp = bhead; btmp; blast = btmp, btmp = btmp->blk_next) ;
    }

  /*  Create free list for file data and metadata blocks  */
  
  btmp = freefile(ip, 0, NULL);

  if (bhead)
    blast->blk_next = btmp;
  else
    bhead = btmp;

  /*  Create free list for the dinode  */

  btmp = (gfs_blklist_t *)gmalloc(sizeof(gfs_blklist_t));
  if (btmp)
    {
      btmp->blk_no = ip->i_num;
      btmp->blk_len = 1;
      btmp->blk_next = bhead;
      bhead = btmp;
    }
  else
    GARBAGE_BLOCKS();

  EXIT(GFS_DIFREE, DEBUG_FS1);
  return(bhead);
}


/******************************************************************************
*******************************************************************************
**
** uint32 gfs_random(void)
**
** Description:
**   Generate a semi-crappy 32-bit pseudo-random number without using
**     floating point.
**
**   The PRNG is from "Numerical Recipes in C", page 284.
**
*******************************************************************************
******************************************************************************/

uint32 gfs_random(void)
{
  static uint32 rand;

  ENTER(GFS_RANDOM, DEBUG_FS1);
 
  rand = (uint32)1664525 * rand + (uint32)1013904223;

  IFDEBUG(DEBUG_FS2,
	  printk(DEBUG_HOW "  %u\n", rand););

  EXIT(GFS_RANDOM, DEBUG_FS1);
  return((uint32)rand);
}


/******************************************************************************
*******************************************************************************
**
** uint32 gfs_randrange(uint32 low, uint32 high)
**
** Description:
**   Generate a semi-crappy 32-bit pseudo-random number, without using
**     floating point, that lies in a given range.
**
**   The PRNG is from "Numerical Recipes in C", page 284.
**
*******************************************************************************
******************************************************************************/

uint32 gfs_randrange(uint32 low, uint32 high)
{
  static uint32 rand;
  uint32 scale;

  ENTER(GFS_RANDRANGE, DEBUG_FS1);
 
  rand = gfs_random() >> 1;
  scale = 0x80000000 / (high - low);
  rand = (rand/scale) + low;
  IFDEBUG(DEBUG_FS2,
	  printk(DEBUG_HOW "  rand = %u, scale = %u\n", rand, scale););

  IFDEBUG(DEBUG_FS2,
	  printk(DEBUG_HOW "  low = %u, high = %u, rand = %u\n", low, high, rand););

  EXIT(GFS_RANDRANGE, DEBUG_FS1);
  return rand;
}


/******************************************************************************
*******************************************************************************
**
** int gfs_change_nlink()
**
** Description:
**   Change a dinode's nlink field by diff.
**     Assumes the lock for dvp is already held.
**
** Input(s):
**   ip - Pointer to the inode.
**   diff - Difference in nlink
**
** Returns:
**   0 on success, -EXXXX on failure.
**
*******************************************************************************
******************************************************************************/

int gfs_change_nlink(gfs_inode_t *ip, int diff)
{
  uint32 udiff;

  ENTER(GFS_CHANGE_NLINK, DEBUG_FS1);

  if (diff < 0)
    {
      udiff = (uint32)(-diff);

      if (ip->i_di.di_nlink >= udiff)
	ip->i_di.di_nlink -= udiff;
      else
	{
	  printk(KERN_ERR "GFS:  negative nlink on dinode %"FMT_64"u\n", ip->i_num);
	  ip->i_di.di_nlink = 0;
	}
    }
  else
    {
      udiff = (uint32)diff;

      if (ip->i_di.di_nlink + udiff < ip->i_di.di_nlink)
	return(-EMLINK);

      ip->i_di.di_nlink += udiff;
    }

  ip->i_di.di_ctime = CURRENT_TIME;
  gfs_inode_attr_in(ip);
  gfs_dinode_out(&ip->i_di, BH_DATA(ip->i_didata));
  gfs_dwrite(ip->i_didata, 0);

  EXIT(GFS_CHANGE_NLINK, DEBUG_FS1);
  return(0);
}


/******************************************************************************
*******************************************************************************
**
** gfs_truncate()
**
** Description:
**   Truncate a file to a certain size
**
** Input(s):
**   ip - the file being truncated
**   length - the new size of the file
**
** Output(s):
**   blks - the list of blocks to free
**
** Returns:
**   0 on success, -EXXXX on failure
**
*******************************************************************************
******************************************************************************/

int gfs_truncate(gfs_inode_t *ip, uint64 length, gfs_blklist_t **blks)
{
  uint64 blocks_freed;

  ENTER(GFS_TRUNCATE, DEBUG_FS1);
 
  *blks = freefile(ip, length, &blocks_freed);

  if (!length)
    ip->i_di.di_height = 0;

  /* gfs_inode_attr_out will take care of this */
  /* ip->i_di.di_bsize = length; */

  ip->i_di.di_blocks -= blocks_freed;

  /* gfs_notify_change will write this out to the bh and */
  /* write the bh to disk                                */

  gfs_inode_attr_in(ip);


  EXIT(GFS_TRUNCATE, DEBUG_FS1);
  return(0);
}


