/******************************************************************************
*******************************************************************************
**
**    Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: gfs_locking.h
**
**    Description:
**	This file defines the interface to the locking modules
**
*******************************************************************************
******************************************************************************/

/*

The GFS Locking Interface


NOTE:  This interface isn't set in stone.  It can (and probably will) change.


     Locking modules are installed using by calling register_lock_proto
function with a valid gfs_lock_prot_t structure as a parameter.  If the
registry is successful, zero is returned.  If there is an error, a negative
number is returned.

The locking module is unregister using the unregister_lock_proto function.

The fields of the gfs_lock_prot_t structure are:

char proto_name[GFS_LOCKNAME_LEN];

       A unique name for the locking protocol.  The module is identified by
  this name.


void *(*mount)(char *table_name, char *host_data,
               gfs_callback_t *cb, void *fsdata);

       Called once by GFS as a file system is mounted.  Mount should make
  the locking module ready to accept lock requests.  Only one mount request
  will be active on a machine at a time, so the locking module doesn't need
  to worry about that.  There can be multiple GFS file systems using the
  same locking protocol (but different lock spaces), so mount can be called
  multiple times.

  The parameters are:

    table_name:
         This is a string that uniquely identifies the lock space that GFS
    wants mounted.  The exact format of the table name is dependent on the
    specific locking module.  For example, the table name for a Dlock module
    might be the name of the Pool mapping the locks.  For a IP based lock
    protocol, the table name might be the IP address and port of the lock
    server.  The table name will be the same for all machines using the same
    lock space.

    host_data:
         This is a string that describes the machine the lock module is running
    on.  It's value is dependent on the specific locking module being used.
    For example, this is used to tell the dlock-0.9.4 module which IP address
    to bind to.

    cb:
         This is a pointer to the GFS call-back function.  The call-back 
    function is discussed more below.

    fsdata:
         This a pointer to private GFS data that should be returned when the
    call-back is called.  (It identifies to GFS which file system the call-back
    is about.)  This pointer is an opaque value to the locking module.

       Mount returns a pointer to a piece of data private to the locking
  module.  This pointer is returned as a parameter to the lock, unlock, and
  unmount calls.  It should identify the lock space that GFS is requesting
  an operation on to the locking module.  The data type pointed to by the
  pointer is opaque to GFS.


void (*unmount)(void *moddata);

       This is called once by GFS when a file system is unmounted.  It does
  the opposite of mount.  Only one unmount call is active on a machine at a
  time.  All locks should have already been unlocked (and released) by the
  time unmount is called.

  The "moddata" parameter is the pointer returned by mount.


int (*lock)(void *moddata, uint64 lock, int op, int flags, void **lockdata);

       This function performs a lock operation.  Lock number "lock" is locked
  in lock space "moddata".

       The lock should remain locked until the unlock operation is called on
  the lock.  It is guaranteed that there aren't two requests for the same
  lock at the same time, but multiple simultaneous requests for different
  are possible.  Some restrictions apply.  Not valid in Tennessee.

  The parameters are:

    moddata:
         The pointer returned from mount.

    lock:
         The lock number

    op:
         The type of lock operation.  The valid values are LCK_OP_ACQUIRE, 
    LCK_OP_TEST, AND LCK_OP_TRY.

      LCK_OP_ACQUIRE:
           Acquire the lock.  If the lock is not immediately available,
      continue to try until the lock is available.  If the lock call returns
      without the lock, it's an error.

      LCK_OP_TRY:
           Try to acquire the lock.  If the lock is not immediately available,
      give up (Don't set LCKOUT_HELD).

      LCK_OP_TEST:
           Return status (described below) that shows what would have happened
      if a LCK_OP_TRY was issued.  Don't actually acquire the lock.

    flags:
         Flags that modify the lock operation.  Each flag is a bit in the
    flags variable.

      LCK_SHARED:
           The lock is allowed to be acquired in the shared state.
      (i.e.  shared is like reader in reader/writer locks.)  If the locking
      protocol used doesn't support shared/exclusive locks, the lock may be
      returned in the exclusive state (as long as the return value is set
      correctly (see below)).  If this bit isn't set, an exclusive lock must
      be acquired.

      LCK_COMMUTE:
           This flag is set if the lock is already held.  It allows a lock held
      in the exclusive state to be converted to the shared state (and visa
      versa.)  In the most straight forward case, this just translates into an
      unlock on the lock and then a lock as if the LCK_COMMUTE flag wasn't set.
      If the locking protocol is smart enough, there can be a direct
      transition from one state to the other.

      LCK_COMMUTE_MOD:
           This flag is the same as LCK_COMMUTE, except that it indicates that
      the data protected by the lock was modified.  The fact that the data was
      modified should be propagated to all the other machines.  In some locking
      protocols, it is possible that is flag is handled in exactly the same way
      as LCK_COMMUTE (see below).

    lockdata:
           This parameter provides a place for the locking module to store
      private data associated with a lock.  If the locking module assigns
      a pointer to *lockdata, that same same pointer will be returned for
      all subsequent calls to lock, unlock, or release (for that lock number). 
      This allows the locking module to store data such as version numbers or
      other data that aids in cache coherence, etc.

         The value returned from lock is an integer.  If there is an error, 
    a negative value corresponding the the errno value of the error is
    returned.  If there is no error, a bit-field of the following bits is
    returned.

      LCKOUT_HELD:
           This flag indicates that the lock operation was successful and the
      lock was acquired.  If the LCK_OP_ACQUIRE operation was used, this flag
      must be set (or a negative value returned).

      LCKOUT_SHARED:
           This flag indicates the lock is held in a shared state.  If the 
      flag is not set, the lock is held in an exclusive state.

      LCKOUT_CACHEABLE:
           This flag is set if no other machine has modified the data
      protected by this lock since this machine last held the lock.

      LCKOUT_EXPIRED_S:
           This lock was held in a shared state by a client who let the lock
      expire.  Recovery of the data protected by the lock may be required.

      LCKOUT_EXPIRED_E:
           This lock was held in an exclusive state by a client who let the
      lock expire.  Recovery of the data protected by the lock may be required.

      LCKOUT_NEED_S:
           This flag is set if the locking module wants GFS to demote the
      just-acquired exclusive lock as soon as possible.

      LCKOUT_NEED_E:
           This flag is set if the locking module wants GFS to release the
      just-acquired lock as soon as possible.

      
int (*unlock)(void *moddata, uint64 lock, int flags, void **lockdata);

       This operations performs the opposite of the lock operation.  A
  previously acquire lock is released.  Again, no two unlocks for the same
  lock will happen at the same time, but unlocks for different locks may
  occur simultaneously.  A lock and unlock on the same lock will never
  happen at the same time.

  The parameters are:

    moddata:
         The pointer returned from mount.

    lock:
         The lock number

    flags:
         Flags that modify the lock operation.  Each flag is a bit in the
    flags variable.

      ULCK_MODIFIED:
           The data protected by the lock was modified.  Other clients who
      next acquire the lock shouldn't have their LCKOUT_CACHEABLE flag set.

      ULCK_RESETEXPIRED:
           This flag means GFS wishes to clear the expired flag on a lock it's
      holding (and unlock in the process).

    lockdata:
         This parameter has the same meaning as the lockdata parameter in the
    lock operation.  See above.


void (*release)(void *moddata, void **lockdata);

       This function is called when a previous lock or unlock action on a lock
  set the lockdata parameter and GFS is not expecting activity on a lock
  for a while.  GFS calls this operation to ask the locking module to free
  the data stored in lockdata.

  If the locking layer never stores data in the lockdata parameter, this
  operation is never called.

  The parameters are:

    moddata:
         The pointer returned from mount.

    lockdata:
         The pointer to the pointer to the data to be freed.   


The Call-Back
   
     This function is called by the lock layer when it wishes GFS to release
(or demote) a lock.  GFS then makes a "best effort" attempt to call the
unlock action for that lock.  If the lock is currently in use, the call may
be delayed (or not happen at all).  Retries may be necessary.

The parameters for the call-back are:

  fsdata:
     The fsdata pointer passed in to the mount operation.

  lock:
     The lock number that the call-back is about.

  flags:
     Each flag is a bit that describes something about the lock.

     CB_NEED_S:
          The locking layer requests that GFS demote the lock to a shared
     state.

     CB_NEED_E:
          The locking layer requests that GFS release the lock.

     CB_EXPIRED_S:
          The lock has expired from a shared state.  (The lock need not be
     held by this client.)

     CB_EXPIRED_E:
          The lock has expired from an exclusive state.


Partial lock implementations

     It's possible (though not recommended) to use lock protocols that don't 
have all the above features:

     A module that doesn't support shared/exclusive locks can ignore the
LCK_SHARED flag if it never returns the LCKOUT_SHARED, LCKOUT_NEED_S,
LCKOUT_EXPIRED_S, CB_NEED_S, or CB_EXPIRED_S flags.

     A module can ignore the ULCK_MODIFIED flag and treat the LCK_COMMUTE_MOD
flag as a LCK_COMMUTE flag, if the LCKOUT_CACHEABLE flag is never returned.

     A module that always returns the LCKOUT_NEED_E flag doesn't ever need to
call the call-back.  (Maybe for expired locks, but not with the CB_NEED_X
flags.)

     A module that never sets the lockdata parameter doesn't need to implement
the release operation.

*/


#ifndef __GFS_LOCKING_DOT_H__
#define __GFS_LOCKING_DOT_H__

#include "global.h"
#include "gfs_ondisk.h"



/*  Flags going into locking layer  */

#define LCK_OP_ACQUIRE      (9)
#define LCK_OP_TEST         (10)
#define LCK_OP_TRY          (11)


#define LCK_SHARED          (0x00000001)
#define LCK_COMMUTE         (0x00000002)
#define LCK_COMMUTE_MOD     (0x00000004)


#define LCKOUT_HELD         (0x00000001)
#define LCKOUT_SHARED       (0x00000002)
#define LCKOUT_CACHEABLE    (0x00000004)
#define LCKOUT_EXPIRED_S    (0x00000008)
#define LCKOUT_EXPIRED_E    (0x00000010)
#define LCKOUT_NEED_S       (0x00000020)
#define LCKOUT_NEED_E       (0x00000040)


#define ULCK_MODIFIED       (0x00000001)
#define ULCK_RESETEXPIRED   (0x00000002)


#define CB_NEED_S           LCKOUT_NEED_S
#define CB_NEED_E           LCKOUT_NEED_E
#define CB_EXPIRED_S        LCKOUT_EXPIRED_S
#define CB_EXPIRED_E        LCKOUT_EXPIRED_E



typedef void (gfs_callback_t)(void *fsdata, uint64 lock, int flags);

typedef struct gfs_lock_proto_s
{
  char proto_name[GFS_LOCKNAME_LEN];

  void *(*mount)(char *table_name, char *host_data,
		 gfs_callback_t *cb, void *fsdata);
  void (*unmount)(void *moddata);

  int (*lock)(void *moddata, uint64 lock, int op, int flags, void **lockdata);
  int (*unlock)(void *moddata, uint64 lock, int flags, void **lockdata);
  void (*release)(void *moddata, void **lockdata);

  struct gfs_lock_proto_s *next;
} gfs_lock_proto_t;



int register_lock_proto(gfs_lock_proto_t *prot);
void unregister_lock_proto(gfs_lock_proto_t *prot);


#endif /*  __GFS_LOCKING_DOT_H__  */
