/******************************************************************************
*******************************************************************************
**
**    Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: main.c
**
*******************************************************************************
******************************************************************************/

#include "gfs.h"
#include "gfs_debug.h"
#include "gfs_locking.h"
#include "gfs_mount.h"
#include "dio.h"
#include "glock.h"
#include "locking.h"





/*  Global Variables  */

gfs_lock_proto_t *gfs_lock_proto = NULL;
semaphore_t gfs_lock_sema;  /*  Initialized by the OS specific code  */





/******************************************************************************
*******************************************************************************
**
** register_lock_proto()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int register_lock_proto(gfs_lock_proto_t *prot)
{
  gfs_lock_proto_t *lp;


  IFDEBUG(DEBUG_LOCKMODULE,
	  printk(DEBUG_HOW "Entering register_lock_proto(%s)\n",
		 prot->proto_name););

  if (!prot)
  {
    IFDEBUG(DEBUG_LOCKMODULE,
	    printk(DEBUG_HOW "Exiting register_lock_proto(bad)\n"););
    return(-EINVAL);
  }

  if (prot->next)
  {
    IFDEBUG(DEBUG_LOCKMODULE,
	    printk(DEBUG_HOW "Exiting register_lock_proto(bad)\n"););
    return(-EBUSY);
  }
  

  GFS_SEMA_DOWN(&gfs_lock_sema);

  for (lp = gfs_lock_proto; lp; lp = lp->next)
    if (!strcmp(lp->proto_name, prot->proto_name))
    {
      GFS_SEMA_UP(&gfs_lock_sema);
      IFDEBUG(DEBUG_LOCKMODULE,
	      printk(DEBUG_HOW "Exiting register_lock_proto(bad)\n"););
      return(-EEXIST);
    }


  prot->next = gfs_lock_proto;
  gfs_lock_proto = prot;

  GFS_SEMA_UP(&gfs_lock_sema);


  IFDEBUG(DEBUG_LOCKMODULE,
	  printk(DEBUG_HOW "Exiting register_lock_proto(good)\n"););

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** unregister_lock_proto()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void unregister_lock_proto(gfs_lock_proto_t *prot)
{
  gfs_lock_proto_t *lp, *last = NULL;
  int done = FALSE;


  IFDEBUG(DEBUG_LOCKMODULE,
	  printk(DEBUG_HOW "Entering unregister_lock_proto(%s)\n", prot->proto_name););

  GFS_SEMA_DOWN(&gfs_lock_sema);

  if (gfs_lock_proto == prot)
  {
    gfs_lock_proto = gfs_lock_proto->next;
    done = TRUE;
  }
  else
    for (lp = gfs_lock_proto; lp; last = lp, lp = lp->next)
      if (lp == prot)
      {
	last->next = lp->next;
	done = TRUE;
      }

  GFS_SEMA_UP(&gfs_lock_sema);


  if (!done)
    printk(KERN_ERR "GFS:  can't unregister lock protocol %s\n", prot->proto_name);

  IFDEBUG(DEBUG_LOCKMODULE,
	  printk(DEBUG_HOW "Exiting unregister_lock_proto()\n"););
}


/******************************************************************************
*******************************************************************************
**
** gfs_mount_lockproto()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int gfs_mount_lockproto(gfs_sbd_t *sdp, gfs_args_t *args)
{
  char proto[GFS_LOCKNAME_LEN + 1], table[GFS_LOCKNAME_LEN + 1];
  char host[256];

  gfs_lock_proto_t *tmp;
  buffer_head_t *bh;
  gfs_sb_t sb;


  IFDEBUG(DEBUG_LOCKMODULE,
	  printk(DEBUG_HOW "Entering gfs_mount_lockproto()\n"););


  /*  Copy in values from "args"  */

  strcpy(proto, args->lockproto);
  strcpy(table, args->locktable);
  strcpy(host, args->hostdata);


  /*  Try to autodetect  */

  if (!proto[0] || !table[0])
  {
    bh = gfs_sread(sdp, GFS_SB_ADDR, GFS_BASIC_BLOCK);
    if (bh)
    {
      gfs_sb_in(&sb, BH_DATA(bh));
      gfs_sforget(bh);
    }

    if (!proto[0])
      strcpy(proto, sb.sb_lockproto);

    if (!table[0])
      strcpy(table, sb.sb_locktable);
  }


  GFS_SEMA_DOWN(&gfs_lock_sema);

  for (tmp = gfs_lock_proto; tmp; tmp = tmp->next)
    if (!strcmp(tmp->proto_name, proto))
    {
      sdp->sd_lockstruct = tmp;
      break;
    }

  if (!tmp)
  {
    GFS_SEMA_UP(&gfs_lock_sema);
    printk(KERN_ERR "GFS:  locking protocol %s not found\n", proto);
    IFDEBUG(DEBUG_LOCKMODULE,
	    printk(DEBUG_HOW "Exiting gfs_mount_lockproto()\n"););
    return(-1);
  }


  sdp->sd_lockdata = sdp->sd_lockstruct->mount(table, host, gfs_glock_cb, sdp);

  if (!sdp->sd_lockdata)
  {
    GFS_SEMA_UP(&gfs_lock_sema);
    printk(KERN_ERR "GFS:  can't mount table %s\n", table);
    IFDEBUG(DEBUG_LOCKMODULE,
	    printk(DEBUG_HOW "Exiting gfs_mount_lockproto()\n"););
    return(-1);
  }

  GFS_SEMA_UP(&gfs_lock_sema);


  IFDEBUG(DEBUG_LOCKMODULE,
	  printk(DEBUG_HOW "Exiting gfs_mount_lockproto()\n"););

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** gfs_unmount_lockproto()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void gfs_unmount_lockproto(gfs_sbd_t *sdp)
{
  IFDEBUG(DEBUG_LOCKMODULE,
	  printk(DEBUG_HOW "Entering gfs_unmount_lockproto()\n"););

  GFS_SEMA_DOWN(&gfs_lock_sema);

  sdp->sd_lockstruct->unmount(sdp->sd_lockdata);

  GFS_SEMA_UP(&gfs_lock_sema);

  IFDEBUG(DEBUG_LOCKMODULE,
	  printk(DEBUG_HOW "Exiting gfs_unmount_lockproto()\n"););
}


