/******************************************************************************
*******************************************************************************
**
**    Copyright 1999 Sistina Software, Inc.  <kpreslan@sistina.com>
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      Kenneth W. Preslan <kpreslan@sistina.com>
** 
**    Module name: udp.c
**
*******************************************************************************
******************************************************************************/

#include <linux/kernel.h>
#include <asm/uaccess.h>
#include <linux/smp_lock.h>

#include "dlock094.h"



typedef struct message_s
{
  uint32 lock;
  uint32 flags;
} message_t;



/******************************************************************************
*******************************************************************************
**
** message_in()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static void message_in(message_t *mess, unsigned char *buff)
{
  message_t *m = (message_t *)buff;

  mess->lock = be32_to_cpup(&m->lock);
  mess->flags = be32_to_cpup(&m->flags);
}


/******************************************************************************
*******************************************************************************
**
** message_out()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

static void message_out(message_t *mess, unsigned char *buff)
{
  message_t *m = (message_t *)buff;

  m->lock = cpu_to_be32p(&mess->lock);
  m->flags = cpu_to_be32p(&mess->flags);
}


/******************************************************************************
*******************************************************************************
**
** open_socket()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int open_socket(dlock094_t *md)
{
  struct sockaddr_in myaddr;

  int error;



  error = sock_create(AF_INET, SOCK_DGRAM, 0, &md->recv_sock);
  if (error < 0)
  {
    printk(KERN_ERR "dlock094:  can't open socket\n");
    return(error);
  }


  myaddr.sin_family = AF_INET;
  myaddr.sin_addr.s_addr = md->clientid;
  myaddr.sin_port = md->port;


  error = md->recv_sock->ops->bind(md->recv_sock, (struct sockaddr *)&myaddr, sizeof(struct sockaddr_in));
  if (error < 0)
  {
    sock_release(md->recv_sock);
    printk(KERN_ERR "dlock094:  can't bind to address\n");
    return(error);
  }


  error = sock_create(AF_INET, SOCK_DGRAM, 0, &md->send_sock);
  if (error < 0)
  {
    sock_release(md->recv_sock);
    printk(KERN_ERR "dlock094:  can't open socket2\n");
    return(error);
  }


  md->send_lock = MUTEX;


  return(0);
}


/******************************************************************************
*******************************************************************************
**
** close_socket()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void close_socket(dlock094_t *md)
{
  sock_release(md->send_sock);
  sock_release(md->recv_sock);
}


/******************************************************************************
*******************************************************************************
**
** send_mess()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

void send_mess(dlock094_t *md, uint32 to, uint32 lock, uint32 flags)
{
  message_t mess;

  struct sockaddr_in youraddr;
  struct msghdr msg;
  struct iovec iov;

  unsigned long proc_flags;
  mm_segment_t	oldfs;
  sigset_t	oldset;

  char buff[sizeof(message_t)];
  int error;


  mess.lock = lock;
  mess.flags = flags;

  message_out(&mess, buff);


  youraddr.sin_family = AF_INET;
  youraddr.sin_addr.s_addr = to;
  youraddr.sin_port = md->port;

  iov.iov_base = buff;
  iov.iov_len = sizeof(message_t);

  msg.msg_name = (struct sockaddr *)&youraddr;
  msg.msg_namelen = sizeof(struct sockaddr_in);
  msg.msg_iov = &iov;
  msg.msg_iovlen = 1;
  msg.msg_control = NULL;
  msg.msg_controllen = 0;
  msg.msg_flags = 0;


  down(&md->send_lock);


  oldfs = get_fs();
  set_fs(get_ds());
	
  spin_lock_irqsave(&current->sigmask_lock, proc_flags);
  oldset = current->blocked;
  sigfillset(&current->blocked);
  recalc_sigpending(current);
  spin_unlock_irqrestore(&current->sigmask_lock, proc_flags);
	

  do 
  {
    error = sock_sendmsg(md->send_sock, &msg, iov.iov_len);
  } while (error == -ECONNREFUSED);

  
  spin_lock_irqsave(&current->sigmask_lock, proc_flags);
  current->blocked = oldset;
  recalc_sigpending(current);
  spin_unlock_irqrestore(&current->sigmask_lock, proc_flags);

  set_fs(oldfs);


  up(&md->send_lock);
}


/******************************************************************************
*******************************************************************************
**
** recvd()
**
** Description:
**
** Input(s):
**
** Output(s):
**
** Returns:
**
*******************************************************************************
******************************************************************************/

int recvd(void *data)
{
  dlock094_t *md = (dlock094_t *)data;
  message_t mess;

  struct msghdr msg;
  struct iovec iov;

  unsigned long flags;
  mm_segment_t	oldfs;

  struct sockaddr_in youraddr;

  unsigned char buff[sizeof(message_t)];
  int n;



  ASSERT(md);


  /*  Set up the thread  */

  lock_kernel();

  exit_mm(current);

  current->session = 1;
  current->pgrp = 1;

  spin_lock_irqsave(&current->sigmask_lock, flags);
  sigfillset(&current->blocked);
  recalc_sigpending(current);
  spin_unlock_irqrestore(&current->sigmask_lock, flags);

  current->fs->umask = 0;
  
  sprintf(current->comm, "dlock_recvd");

  unlock_kernel();


  /*  Wake up the process who started us  */

  up(&md->recvd_lock);


  while (1)
  {
    iov.iov_base = buff;
    iov.iov_len = sizeof(message_t);

    msg.msg_name = (struct sockaddr *)&youraddr;
    msg.msg_namelen = sizeof(struct sockaddr_in);
    msg.msg_iov = &iov;
    msg.msg_iovlen = 1;
    msg.msg_control = NULL;
    msg.msg_controllen = 0;
    msg.msg_flags = 0;


    oldfs = get_fs();
    set_fs(get_ds());
	
    n = sock_recvmsg(md->recv_sock, &msg, iov.iov_len, 0);

    set_fs(oldfs);


    if (md->recvd_stop)
      break;


    if (n == sizeof(message_t))
    {
      message_in(&mess, buff);

      mess.flags &= CB_NEED_S | CB_NEED_E;

      md->cb(md->fsdata, mess.lock, mess.flags);
    }
    else
      printk(KERN_ERR "dlock094:  strange message (%d)\n", n);
  }


  up(&md->recvd_lock);

  return(0);
}




