/******************************************************************************
*
*    Copyright 1999 Regents of the University of Minnesota
*
*    This is free software released under the GNU General Public License.
*    There is no warranty for this software.  See the file COPYING for
*    details.
*
*    See the file CONTRIBUTORS for a list of contributors.
*
*    This file is maintained by:
*      Mike C Tilstra <tilstra@lcse.umn.edu>
* 
*    Module name: dlock_ip.o
*
* These are just some wrapped up functions for working the comm layer.
*/

#include "dlip.h"
#include "dlrpc.h"
#include "dlip_cmd_wrapper.h"


/******************************************************************************
 * temp solution.  keeps so that each client can only send one request at a
 * time.
 */
#ifdef ONE_AT_A_TIME
#define OAT_DOWN down(&oneatime)
#define OAT_UP up(&oneatime)
static struct semaphore oneatime = MUTEX;
#else
#define OAT_DOWN
#define OAT_UP
#endif

/******************************************************************************
 * do_dlock()
 * Builds up the transport structures, and does the rpc_call to the server.
******************************************************************************/
int do_dlock(dlock_ip_t *md, lock_t *data,
		u8 action, uint64 lock, 
		dlock_return_t *rtn)
{
	CDBs cdb;
	RDFs rdf;
	int error;
#ifdef BREAK_BLOCKER
	sigset_t oldset;
	unsigned long proc_flags;
#endif

	memset(&cdb, 0, sizeof(CDBs));
	memset(&rdf, 0, sizeof(RDFs));

	cdb.type = 0;
	cdb.cdb_u.dlock.opcode = 0xA0;
	cdb.cdb_u.dlock.action = action;
	cdb.cdb_u.dlock.lock_number = lock;
	cdb.cdb_u.dlock.input_version_number = data->vn;
	cdb.cdb_u.dlock.cid = md->mycid;

	rdf.type = 0;
	rdf.rdf_u.dlock.dsr_cnt = md->maxclnts;
	rdf.rdf_u.dlock.cids = data->cids;
#ifdef BREAK_BLOCKER
	spin_lock_irqsave(&current->sigmask_lock, proc_flags);
	oldset = current->blocked;
	sigfillset(&current->blocked);
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->sigmask_lock, proc_flags);
#endif
	OAT_DOWN;
#ifdef RPC_PER_LOCK
	error = rpc_call(data->clnt, DLIP_DODLOCK, &cdb, &rdf, 0);
#else
	error = rpc_call(md->clnt, DLIP_DODLOCK, &cdb, &rdf, 0);
#endif
	OAT_UP;

#ifdef BREAK_BLOCKER
	spin_lock_irqsave(&current->sigmask_lock, proc_flags);
	current->blocked = oldset;
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->sigmask_lock, proc_flags);
#endif	
	if( error !=0 || rdf.type != 0)
	printk("dlock_ip: DODLOCK: error = %d; rdf.type = %d "
			 "on line %d in file %s"
			 "\n", error, rdf.type,__LINE__,__FILE__);
	if(error) return error;
	if(rdf.type != 0) return -EIO;

	/* The clnt list should have been happily pulled into the data->cids.*/
	rtn->res = rdf.rdf_u.dlock.result;
	rtn->vn  = rdf.rdf_u.dlock.version_number;
	rtn->exp = rdf.rdf_u.dlock.expired;
	rtn->state   = rdf.rdf_u.dlock.state;
	data->cidcnt = rdf.rdf_u.dlock.num_holders;

	return 0;
}

/******************************************************************************
 * do_refresh()
 * refreshes all of the locks held by cid.  This will be called from a
 * timer of some kind.
******************************************************************************/
int do_refreshAll(struct rpc_clnt *clnt, u32 cid)
{
	CDBs cdb;
	RDFs rdf;
	int error;
#ifdef BREAK_BLOCKER
	sigset_t oldset;
	unsigned long proc_flags;
#endif

	memset(&cdb, 0, sizeof(CDBs));
	memset(&rdf, 0, sizeof(RDFs));

	cdb.type = 0;
	cdb.cdb_u.dlock.opcode = 0xA0;
	cdb.cdb_u.dlock.action = 0x04;
	cdb.cdb_u.dlock.lock_number = ~0;
	cdb.cdb_u.dlock.cid = cid;

#ifdef BREAK_BLOCKER
	spin_lock_irqsave(&current->sigmask_lock, proc_flags);
	oldset = current->blocked;
	sigfillset(&current->blocked);
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->sigmask_lock, proc_flags);
#endif
	OAT_DOWN;
	error = rpc_call(clnt, DLIP_DODLOCK, &cdb, &rdf, 0);
	OAT_UP;

#ifdef BREAK_BLOCKER
	spin_lock_irqsave(&current->sigmask_lock, proc_flags);
	current->blocked = oldset;
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->sigmask_lock, proc_flags);
#endif	
	if( error !=0 || rdf.type != 0)
	printk("dlock_ip: DODLOCK: error = %d; rdf.type = %d "
			 "on line %d in file %s"
			 "\n", error, rdf.type,__LINE__,__FILE__);
	if(error)
		return error;
	if(rdf.type != 0)
		return -EIO;

	return 0;
}

/******************************************************************************
 * do_modepage()
 * For getting the max number of dlocks & max number of cids on each lock 
******************************************************************************/
int do_modepage(struct rpc_clnt *clnt, u8 *maxclnt,
		u32 *numlock, u32 *lock_timeout)
{
	CDBs cdb;
	RDFs rdf;
	int error;
#ifdef BREAK_BLOCKER
	sigset_t oldset;
	unsigned long proc_flags;
#endif

	memset(&cdb, 0, sizeof(CDBs));
	memset(&rdf, 0, sizeof(RDFs));

	cdb.type = 1;
	cdb.cdb_u.mode.tag = 0;

#ifdef BREAK_BLOCKER
	spin_lock_irqsave(&current->sigmask_lock, proc_flags);
	oldset = current->blocked;
	sigfillset(&current->blocked);
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->sigmask_lock, proc_flags);
#endif

	OAT_DOWN;
	error = rpc_call(clnt, DLIP_DODLOCK, &cdb, &rdf, 0);
	OAT_UP;

#ifdef BREAK_BLOCKER
	spin_lock_irqsave(&current->sigmask_lock, proc_flags);
	current->blocked = oldset;
	recalc_sigpending(current);
	spin_unlock_irqrestore(&current->sigmask_lock, proc_flags);
#endif
	
	if( error !=0 || rdf.type != 2)
	printk("dlock_ip: DODLOCK: error = %d; rdf.type = %d "
			 "on line %d in file %s"
			 "\n", error, rdf.type,__LINE__,__FILE__);
	if(error)
		return error;
	if(rdf.type != 2)
		return -EIO;
	*numlock = rdf.rdf_u.mpag.num_locks;
	*maxclnt = rdf.rdf_u.mpag.max_clients;
	*lock_timeout = rdf.rdf_u.mpag.lock_time_ms;

	return 0;
}

