/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1998 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@lcse.umn.edu>
** 
**    Module name: lock.c
**
**    Description:
**	Device lock routines.
**
*******************************************************************************
******************************************************************************/

#ifdef IRIX
#include <sys/ddi.h>
#include <sys/types.h>
#include <sys/kmem.h>
#include <sys/systm.h>
#include <sys/scsi.h>
#include <sys/ksynch.h>
#include <sys/conf.h>
#include <sys/debug.h>
#include <sys/sysmacros.h>
#endif

#ifdef linux
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/blk.h>

#include <asm/system.h>
#include <asm/io.h>

#include "scsi.h"
#include "sd.h"
#endif

#include "global.h"
#include "pool.h"
#include "pool_head.h"
#include "pool_debug.h"
#include "lock.h"

#ifdef IRIX
#include <sys/scsiInitiator_api.h>
#include <sys/scsiInitiator.h>
#endif

/* Preprocessor Defines */


/* Function Macros */


/* Type Definitions */


/* Global Variables */

#ifdef linux
extern Scsi_Disk *rscsi_disks;
#endif

/* Function Prototypes */

#ifdef IRIX
static void pool_notify_dksc(struct scsi_request *srp);
static void pool_notify_prisa(void *arg);
static lock_ret_t device_lock_dksc(u_char *cmd, dev_t dev);
static lock_ret_t device_lock_prisa(u_char *cmd, dlockdev_t *dlp);
#endif
#if 0
static lock_ret_t device_lock_void(pool_t *pp, dlock_params_t *params);
#endif
#ifdef linux
static lock_ret_t device_lock_sd(u_char *cmd, kdev_t dev);
static void pool_notify_sd(Scsi_Cmnd *SCpnt);
#endif


/******************************************************************************
*******************************************************************************
** NAME
**	device_lock
**
** DESCRIPTION 
**	Put together the actual scsi request and forward it to the correct
**	lower level handler.
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

/* XXX cache alignment removal ? */

lock_ret_t device_lock(uint32 action, uint32 pln, uint32 clock, dlockdev_t *dlp)
{
  u_char 	*cmd;
  lock_ret_t	ret;
  kdev_t	dev = dlp->dl_dev;
  
    /* Create the command */

  IFDEBUG(POOL_DEBUG_LOCK2,
	  printk("Entered device_lock: lock=%d, action=0x%x\n", 
		 pln, action););

  ASSERT(POOL_CMD == 2);

#ifdef IRIX
  cmd = kmem_zalloc(LOCK_CDBLEN, KM_SLEEP | KM_CACHEALIGN);
#endif
#ifdef linux
  cmd = kmalloc(LOCK_CDBLEN, GFP_KERNEL);
#endif

#if (POOL_CMD == 2)
  cmd[0] = LOCK_OPCODE;
  
  if (action == POOL_LOCK)
    cmd[1] = 0x1;
  else if (action == POOL_UNLOCK)
    cmd[1] = 0x2;
  else if (action == POOL_UNLOCK_INCR)
    cmd[1] = 0x3;
  else if (action == POOL_RESET)
    cmd[1] = 0x4;
  else if (action == POOL_ACTIVITY_ON)
    cmd[1] = 0x5;
  else if (action == POOL_ACTIVITY_OFF)
    cmd[1] = 0x6;
  
  cmd[2] = (pln >> 24) & 0xff;
  cmd[3] = (pln >> 16) & 0xff;
  cmd[4] = (pln >>  8) & 0xff;
  cmd[5] = (pln)       & 0xff;

  cmd[6] = (clock >> 24) & 0xff;
  cmd[7] = (clock >> 16) & 0xff;
  cmd[8] = (clock >>  8) & 0xff;
  cmd[9] = (clock)       & 0xff;  

  cmd[10] = LOCK_RETSIZE;
  cmd[11] = 0x00;
#endif

#ifdef IRIX
  if (emajor(dev) != POOL_PRISA) 
  {
    /* dksc or equivalent */
    
    ret = device_lock_dksc(cmd, dev);
    kmem_free(cmd, LOCK_CDBLEN);
    return ret;
  }
  else
  {
    ret = device_lock_prisa(cmd, dlp);
    kmem_free(cmd, LOCK_CDBLEN);
    return ret;
  }
#endif
#ifdef linux
  ret = device_lock_sd(cmd, dev);
  kfree(cmd);
  return ret;
#endif
}



/******************************************************************************
*******************************************************************************
**
** static lock_ret_t device_lock_void()
**
** Description:
**   This routine implements the device lock command using the SCSI reserve 
**   and release commands. No actual device locking is performed; however, the
**   latency of such a command is simulated by intserting a short delay.
**
** Input(s):
**  *pp     - 
**  *params - 
**
** Output(s):
**
**
** Returns:
**
**
*******************************************************************************
******************************************************************************/

#if 0
static lock_ret_t device_lock_void(pool_t *pp, dlock_params_t *params)
{
  lock_ret_t	ret;
  int32		action = params->dp_action;

  IFDEBUG(POOL_DEBUG_LOCK2,
	  printf("device_lock: lock=%d, action=%d\n", 
		 params->dp_phyln, action););
  
  /* Update the return structure */

  ret.r_status = 0;
  ret.r_result = 1;
  ret.r_locktab.l_activity = 0;
  
  switch (action)
    {
    case POOL_LOCK: 
      ret.r_locktab.l_state = 1;
      ret.r_locktab.l_clock = pp->p_locktab[params->dp_ctlarg->l_id].l_clock;
      break;
    case POOL_UNLOCK:
      ret.r_locktab.l_state = 0;
      ret.r_locktab.l_clock = pp->p_locktab[params->dp_ctlarg->l_id].l_clock;
      break;  
    case POOL_UNLOCK_INCR:
      ret.r_locktab.l_state = 0;
      ret.r_locktab.l_clock = pp->p_locktab[params->dp_ctlarg->l_id].l_clock + 1;
      break;
    case POOL_RESET:
      ret.r_locktab.l_state = 0;
      ret.r_locktab.l_clock = pp->p_locktab[params->dp_ctlarg->l_id].l_clock + 1;
      break;
    default:
      printf("device_lock: bad action\n");
      ret.r_status = 1;
      return(ret);
    }

  /* Simulate the latency of an actual device lock operation */

  delay(drv_usectohz(FAKE_DLOCK_DELAY));
	
  return(ret);
}

#endif 

/******************************************************************************
*******************************************************************************
**
** static lock_ret_t device_lock_dksc()
**
** Description:
**   This routine sends a SCSI device lock command.
**
** Input(s):
**
**
** Output(s):
**
**
** Returns:
**
**
*******************************************************************************
******************************************************************************/

/* ARGSUSED */
#ifdef IRIX
static lock_ret_t device_lock_dksc(u_char *cmd, dev_t dev)
{
  scsi_request_t	*req;
  lock_ret_t		 ret;
  unsigned char		 byte4, byte3, byte2, byte1, byte0;
  int			 s;
  sema_t		*reqdone;
#ifndef CELL_IRIX
  u_char		 ctlr;
  u_char		 targ;
  u_char		 lun;
  int			 driver;
#else
  scsi_lun_info_t	*linfo;
  scsi_targ_info_t	*tinfo;
  scsi_ctlr_info_t	*cinfo;
  vertex_hdl_t	         lvhdl;
#endif

#ifndef CELL_IRIX
  ctlr = dksc_ctlr(dev);
  targ = dksc_unit(dev);
  lun = dksc_lun(dev);
#else

  /* XXX relying on passemble passing in the dev_t for the partition! */

  lvhdl = SDI_LUN_VHDL(SPI_DISK_INFO(scsi_part_info_get(dev_to_vhdl(dev))));
  linfo = scsi_lun_info_get(lvhdl);
  tinfo = linfo->sli_targ_info;
  cinfo = tinfo->sti_ctlr_info;
#endif

#ifndef CELL_IRIX
  IFDEBUG(POOL_DEBUG_LOCK2,
	  printf("ctlr(%d)drive(%d)lun(%d)\n", ctlr, targ, lun););
#endif

  /* Initialize the connection between the driver and the host adapater */

#ifndef CELL_IRIX
  driver = scsi_driver_table[ctlr];

  s = (*scsi_alloc[driver])(ctlr, targ, lun, SCSIALLOC_QDEPTH, NULL);
#else
  s = SCI_ALLOC(cinfo)(lvhdl,SCSIALLOC_QDEPTH, NULL);
#endif

  if (s != SCSIALLOCOK)
    {
      printf("dlockopen: not SCSIALLOCOK\n");
      ret.r_status = -1;
      ret.r_result = 0;
      ret.r_locktab.l_state = 0;
      ret.r_locktab.l_activity = 0;
      ret.r_locktab.l_clock = 0;

      return(ret);
    }

  /* Allocate the request structure. Hey, it's inefficient but it's easy */

  reqdone = (sema_t *)kmem_alloc(sizeof(sema_t), KM_SLEEP | KM_CACHEALIGN);
  initnsema(reqdone, 0, "POOL_SCS");

  req = (scsi_request_t *)
    kmem_zalloc(sizeof(scsi_request_t), KM_SLEEP | KM_CACHEALIGN);

#ifndef CELL_IRIX
  req->sr_ctlr = ctlr;
  req->sr_target = targ;
  req->sr_lun = lun;
#else 
  req->sr_ctlr = cinfo->sci_adap;
  req->sr_target = tinfo->sti_targ;
  req->sr_lun = linfo->sli_lun;
  req->sr_lun_vhdl = lvhdl;
#endif
#if HOQ_LOCK
  req->sr_tag = SC_TAG_HEAD;
#else
  if (cmd[1] != POOL_LOCK)
    req->sr_tag = SC_TAG_HEAD;
  else 
    req->sr_tag = SC_TAG_SIMPLE;
#endif

  req->sr_command = cmd;
  req->sr_cmdlen = LOCK_CDBLEN;
  req->sr_flags = SRF_DIR_IN|SRF_AEN_ACK;
  req->sr_timeout = 60 * HZ;

#if LOCK_RETSIZE != 0

  req->sr_buffer = (u_char *)
    kmem_alloc(LOCK_RETSIZE, KM_SLEEP|KM_CACHEALIGN);

#endif

  req->sr_buflen = LOCK_RETSIZE;
  req->sr_sense = (u_char *)kmem_zalloc(64, KM_SLEEP|KM_CACHEALIGN);
  req->sr_senselen = 64; 
  
  req->sr_notify = pool_notify_dksc; 
  req->sr_bp = NULL; 

  /* XXX we don't touch sr_dev_vhdl.  Don't even know what its for */

  req->sr_dev = (void *)reqdone; 
  
  /* Invalidate the cache - R4000 */

  dki_dcache_inval(req->sr_buffer, req->sr_buflen);

  /* Send the command */

#ifdef CELL_IRIX
  SCI_COMMAND(cinfo)(req);
#else
  (*scsi_command[driver])(req);
#endif

  /* Wait for notification */

  psema(reqdone, PRIBIO);


  IFDEBUG(POOL_DEBUG_LOCK2,
	  printf("device_lock: sr_status = %d\n", req->sr_status);
	  printf("device_lock: sr_scsi_status = %d\n", req->sr_scsi_status););
  
  IFDEBUG(POOL_DEBUG_LOCK3,
	  printf("device_lock: sr_ha_flags = %d\n", req->sr_ha_flags);
	  printf("device_lock: sr_resid = %d\n", req->sr_resid);
	  printf("device_lock: sr_sensegotten = %d\n", req->sr_sensegotten););
  
  IFDEBUG(POOL_DEBUG_LOCK3,
	  {
	    int i;
	    
	    printf("Command:\n");
	    for (i = 0; i < req->sr_cmdlen; i++)
	      {
		printf("%x  ", req->sr_command[i]);
	      }
	    printf("\n\n");
	    
	    printf("Return:\n");
	    for (i = 0; i < (req->sr_buflen - req->sr_resid); i++)
	      {
		printf("%x  ", req->sr_buffer[i]);
	      }
	    printf("\n\n");
	  }
	  );

  /* Fill out the returned data structure */

#if (POOL_CMD == 2)
  ret.r_status = req->sr_status;
  
  byte4 = req->sr_buffer[0];
  byte3 = req->sr_buffer[1];
  byte2 = req->sr_buffer[2];
  byte1 = req->sr_buffer[3];
  byte0 = req->sr_buffer[4];
  
  ret.r_result = byte4 >> 7;
  ret.r_locktab.l_state = (byte4 >> 6) & 0x1;
  ret.r_locktab.l_activity = (byte4 >> 5) & 0x1;
  ret.r_locktab.l_clock = byte3 << 24 | byte2 << 16 | byte1 << 8 | byte0;;
#endif

  /* Clean house */

#ifdef CELL_IRIX
  SCI_FREE(cinfo)(lvhdl, NULL);
#else
  (*scsi_free[driver])(ctlr, targ, lun, NULL);
#endif
  
  freesema(reqdone);
  kmem_free(reqdone, sizeof(sema_t));

#if LOCK_RETSIZE != 0
    kmem_free(req->sr_buffer, LOCK_RETSIZE);
#endif

  kmem_free(req->sr_sense, req->sr_senselen);
  kmem_free(req, sizeof(scsi_request_t));
  
  return(ret);
}
#endif

/******************************************************************************
*******************************************************************************
**
** void pool_notify_dksc()
**
** Description:
**   This routine serves as an interrupt handler for the requests.
**
** Input(s):
**  *srp - 
**
** Output(s):
**
**
** Returns:
**
**
*******************************************************************************
******************************************************************************/
#ifdef IRIX
static void pool_notify_dksc(struct scsi_request *srp) 
{
  sema_t	*reqdone;

  IFDEBUG(POOL_DEBUG_LOCK2,
	  printf("Entered pool_notify()\n"););
  
  reqdone = (sema_t *)(srp->sr_dev);
  vsema(reqdone);

  IFDEBUG(POOL_DEBUG_LOCK2,
	  printf("Exiting pool_notify()\n"););

  return;
}
#endif

/******************************************************************************
*******************************************************************************
**
** static lock_ret_t device_lock_prisa()
**
** Description:
**   This routine sends a SCSI device lock command.
**
** Input(s):
**  *pp     - 
**  *params - 
**
** Output(s):
**
**
** Returns:
**
**
*******************************************************************************
******************************************************************************/

/* ARGSUSED */
#ifdef IRIX
static lock_ret_t device_lock_prisa(u_char *cmd, dlockdev_t *dlp)
{
#ifndef HOQ_LOCK
  int32			 action = cmd[1];
#endif
  struct siRequest	*req;
  lock_ret_t		 ret;
  unsigned char		 byte4, byte3, byte2, byte1, byte0;
  sema_t		*reqdone;


  ASSERT(dlp->dl_data != NULL);

  /* Allocate the request structure. Hey, it's inefficient but it's easy */

  if ((req = psiGetRequest((struct siEndpoint *)dlp->dl_data)) == 0)
    {
      printf("device_lock: Error doing siGetRequest\n");
      ret.r_status = -1;
      ret.r_result = 0;
      ret.r_locktab.l_state = 0;
      ret.r_locktab.l_activity = 0;
      ret.r_locktab.l_clock = 0;
      
      return(ret);
    }

  reqdone = (sema_t *)kmem_alloc(sizeof(sema_t), KM_SLEEP);
  initnsema(reqdone, 0, "POOL_PRI");
  
  /* Fil the request structure */

  req->lun = 0;
#if HOQ_LOCK
  req->task_attribute = siHeadOfQueue;
#else 
  if (action != POOL_LOCK)
    req->task_attribute = siHeadOfQueue;
  else 
    req->task_attribute = siSimple;
#endif
  req->transtype = siRead;
  req->timeout = 500;
  
  req->data = (caddr_t)kmem_zalloc(LOCK_RETSIZE, KM_SLEEP|KM_CACHEALIGN);
  req->data_buf = 0;
  req->data_bytes = LOCK_RETSIZE;
  req->offset = 0;
  req->line = 0;
  req->stride = 0;
  
  req->sense = (u_char *)kmem_zalloc(64, KM_SLEEP);
  req->sense_bytes = 64;
  
  req->notify = pool_notify_prisa;
  req->notify_arg = (void *)reqdone;
  
  /* Create the command - guarantee endianess */

#if (POOL_CMD == 2)
  req->cdb.w1 = cmd[0] << 24 | cmd[1] << 16 | cmd[2] << 8 | cmd[3];
  req->cdb.w2 = cmd[4] << 24 | cmd[5] << 16 | cmd[6] << 8 | cmd[7];
  req->cdb.w3 = cmd[8] << 24 | cmd[9] << 16 | cmd[10] << 8 | cmd[11];
  req->cdb.w4 = 0x00000000;
#endif

  /* Invalidate the cache - R4000 */

  dki_dcache_inval(req->data, (int)req->data_bytes);

  /* Send the command */

  psiExecuteCommand(req);
  
  /* Wait for notification */

  psema(reqdone, PRIBIO);
  
  IFDEBUG(POOL_DEBUG_LOCK2,
	  {
	    int i;
		  
	    printf("Command:\n");
	    printf("%x ", req->cdb.w1);
	    printf("%x ", req->cdb.w2);
	    printf("%x ", req->cdb.w3);
	    printf("%x ", req->cdb.w4);
	    printf("\n\n");
		  
	    printf("Return:\n");
	    for (i = 0; i < LOCK_RETSIZE; i++)
	      {
		printf("%x  ", req->data[i]);
	      }
	    printf("\n\n");
	  }
	  );

  /* Fill out the returned data structure */

#if (POOL_CMD == 2)
  ret.r_status = req->result; 
  
  byte4 = req->data[0];
  byte3 = req->data[1];
  byte2 = req->data[2];
  byte1 = req->data[3];
  byte0 = req->data[4];
  
  ret.r_result = byte4 >> 7;
  ret.r_locktab.l_state = (byte4 >> 6) & 0x1;
  ret.r_locktab.l_activity = (byte4 >> 5) & 0x1;
  ret.r_locktab.l_clock = byte3 << 24 | byte2 << 16 | byte1 << 8 | byte0;
#endif

  /* Clean house */

  freesema(reqdone);
  kmem_free(reqdone, sizeof(sema_t));
  kmem_free(req->data, req->data_bytes);
  kmem_free(req->sense, req->sense_bytes);
  psiFreeRequest((struct siEndpoint*)dlp->dl_data, req); 
  
  return(ret);
}
#endif


/******************************************************************************
*******************************************************************************
**
** void pool_notify_prisa()
**
** Description:
**   This routine serves as an interrupt handler for the requests.
**
** Input(s):
**  *srp - 
**
** Output(s):
**
**
** Returns:
**
**
*******************************************************************************
******************************************************************************/

#ifdef IRIX
static void pool_notify_prisa(void *arg)
{
  sema_t *reqdone;

  IFDEBUG(POOL_DEBUG_LOCK2,
	  printf("Entered pool_notify()\n"););
  
  reqdone = (sema_t *)arg;  
  vsema(reqdone);

  IFDEBUG(POOL_DEBUG_LOCK2,
	  printf("Exiting pool_notify()\n"););

  return;
}
#endif

/******************************************************************************
*******************************************************************************
**
** static lock_ret_t device_lock_sd()
**
** Description:
**   This routine sends a SCSI device lock command.
**
** Input(s):
**
**
** Output(s):
**
**
** Returns:
**
**
*******************************************************************************
******************************************************************************/

/* ARGSUSED */
#ifdef linux
static lock_ret_t device_lock_sd(u_char *cmd, kdev_t dev)
{

  Scsi_Cmnd		* SCpnt;
  Scsi_Device           * device;
  char			* buff;
  int 			wait = 1;
  lock_ret_t		ret;
  struct semaphore  	sem;
  unsigned char	        byte4, byte3, byte2, byte1, byte0;
  unsigned long flags;

  IFDEBUG(POOL_DEBUG_LOCK3,
	  printk("Entered device_lock_sd\n"););

  /* hack - rscsi_disks exported from sd */
  device = rscsi_disks[MINOR(dev) >> 4].device;

  if (!(buff = (u_char *)scsi_malloc(512)))
  {
      ret.r_status = -1;
      ret.r_result = 0;
      ret.r_locktab.l_state = 0;
      ret.r_locktab.l_activity = 0;
      ret.r_locktab.l_clock = 0;
      return(ret);
  }

  spin_lock_irqsave(&io_request_lock, flags);

  /* no request structure (CURRENT) affects scsi functions which 
   * usually deal with these request structs for ordinary requests */

  if (!(SCpnt=scsi_allocate_device(NULL, device, wait))) 
    {
      spin_unlock_irqrestore(&io_request_lock, flags);
      scsi_free(buff, 512);
      printk("dlockopen: not SCSIALLOCOK\n");
      ret.r_status = -1;
      ret.r_result = 0;
      ret.r_locktab.l_state = 0;
      ret.r_locktab.l_activity = 0;
      ret.r_locktab.l_clock = 0;
      return(ret);
    }

  SCpnt->tag = SC_TAG_HEAD;  /* overwritten by HBA driver? */

  sem = MUTEX_LOCKED;
  SCpnt->request.sem = &sem;
  SCpnt->request.rq_status = RQ_SCSI_BUSY;

  scsi_do_cmd( SCpnt,           /* Scsi_Cmnd pointer */
               cmd,             /* CDB */
               buff,            /* buffer for returned lock data */
	       LOCK_RETSIZE,    /* buff length */
	       pool_notify_sd, 	/* scsi cmd callback function */
	       DL_TIMEOUT,      /* timeout value */
	       DL_MAXRETRY);	/* number of retried attempted */

  spin_unlock_irqrestore(&io_request_lock, flags);

  down(&sem);
  SCpnt->request.sem = NULL; 

  IFDEBUG(POOL_DEBUG_LOCK3,
	  printk("device_lock: SCpnt->result = %d\n", SCpnt->result););
  
  IFDEBUG(POOL_DEBUG_LOCK3,
	  {
	    int i;
	    
	    printk("Command:\n");
	    for (i = 0; i < SCpnt->cmd_len; i++)
	      {
		printk("%x  ", SCpnt->cmnd[i]);
	      }
	    printk("\n");
	    
	    printk("Return:\n");
	    for (i = 0; i < SCpnt->bufflen; i++)
	      {
		printk("%x  ", buff[i]);
	      }
	    printk("\n");
	  }
	  );

	  
  /* fill in ret struct from buff data */
#if (POOL_CMD == 2)
  ret.r_status = status_byte(SCpnt->result);  /* status of scsi cmd */
  
  byte4 = buff[0];
  byte3 = buff[1];
  byte2 = buff[2];
  byte1 = buff[3];
  byte0 = buff[4];
  
  ret.r_result = byte4 >> 7;
  ret.r_locktab.l_state = (byte4 >> 6) & 0x1;
  ret.r_locktab.l_activity = (byte4 >> 5) & 0x1;
  ret.r_locktab.l_clock = byte3 << 24 | byte2 << 16 | byte1 << 8 | byte0;;
#endif

  if (status_byte(SCpnt->result) != 0) {
    unsigned char *buf = SCpnt->sense_buffer;
    printk("device_lock_sd: DLOCK command failed, status: %d\n", status_byte(SCpnt->result));
    printk("device_lock_sd: sense: 0: 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n"
	   "device_lock_sd: sense: 1: 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x 0x%02x\n",
	   buf[0], buf[1], buf[2], buf[3], buf[4], buf[5], buf[6], buf[7],
	   buf[8], buf[9], buf[10], buf[11], buf[12], buf[13], buf[14], buf[15]);
  }

  scsi_free(buff, 512);

  wake_up(&SCpnt->device->device_wait);
  scsi_release_command(SCpnt);

  IFDEBUG(POOL_DEBUG_LOCK3,
	  printk("Exiting device_lock_sd\n"););

  return(ret);

}
#endif


/******************************************************************************
*******************************************************************************
**
** void pool_notify_sd()
**
** Description:
**   This routine serves as an interrupt handler for the requests.
**
** Input(s):
**  
**
** Output(s):
**
**
** Returns:
**
**
*******************************************************************************
******************************************************************************/

#ifdef linux
static void pool_notify_sd(Scsi_Cmnd *SCpnt)
{

  IFDEBUG(POOL_DEBUG_LOCK3,
	  printk("Entered pool_notify_sd\n"););

  SCpnt->request.rq_status = RQ_SCSI_DONE;

  if (SCpnt->request.sem != NULL)
    up(SCpnt->request.sem);

  IFDEBUG(POOL_DEBUG_LOCK3,
	  printk("Exiting pool_notify_sd\n"););
  
  return;

}
#endif


