/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1998 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@lcse.umn.edu>
** 
**    Module name: pool.c
**
**    Description:
**	Network storage pool device driver routines.
**
*******************************************************************************
******************************************************************************/

#ifdef linux
#include <linux/types.h>
#include <linux/fs.h>
#include <linux/blk.h>
#include <linux/malloc.h>
#include <linux/vmalloc.h>
#include <asm/uaccess.h>
#endif

#ifdef IRIX
#include <sys/types.h>
#include <sys/buf.h>
#include <sys/cmn_err.h>
#include <sys/debug.h>
#include <sys/ddi.h>
#include <sys/conf.h>
#include <sys/cred.h>
#include <sys/errno.h>
#include <sys/file.h>
#include <sys/kmem.h>
#include <sys/ksynch.h>
#include <sys/ktime.h>
#include <sys/mips_addrspace.h>
#include <sys/open.h>
#include <sys/param.h>
#include <sys/scsi.h>
#include <sys/sema.h>
#include <sys/sysmacros.h>
#include <sys/systm.h>
#include <sys/scsiInitiator.h>
#include <sys/scsiInitiator_api.h>
#include <sys/xlv_ioctl.h>
#endif

#include "global.h"
#include "pool_debug.h"
#include "pool.h"
#include "pool_head.h"
#include "lock.h"

#ifdef IRIX
/* Prisa interface stuff, ugly hack... */
int  (*psiOpen)(scsiPortID, u_int, struct siEndpoint **);
void (*psiClose)(struct siEndpoint *);
struct siRequest *(*psiGetRequest)(struct siEndpoint *);
void (*psiFreeRequest)(struct siEndpoint *, struct siRequest *);
void (*psiExecuteCommand)(struct siRequest *);
#endif

/* Variables for pool devices */

pool_t		pools[POOL_MAX_DEVICES];	/* lookup table */


#ifdef IRIX
extern volatile void   *iotimer_addr;      /* physical address of timer */
extern uint64	 	iotimer_max;
extern uint64           iotimer_rate_ps;   /* picoseconds per tick */
extern uint         	iotimer_size;
#endif


/* These are defined in /var/sysgen/master.d/pool */
extern const uint32_t pool_reset_timeout_us;
extern const uint32_t pool_cnf_delay_us;
extern const uint32_t pool_major_number;

/* Function Prototypes */

static int gen_pool(pool_add_t *add, pool_t *pool);


/******************************************************************************
*******************************************************************************
**
** int pool_add(pool_add_t *add, cred_t *crp)
**
** Description:
**	This routine reads the pool label information from user space and
**	assigns a pool minor number to a newly created pool.  The arguement is
**	a pointer to an pool_add_t structure which has been copied to kernel
**	space.  All of the pointers in this structure reference addresses in
**	user space.  If the pool label is already loaded this routine returns
**	the minor number already assigned to the pool. It is returned in the
**	pa_pool_dev field of the pool_add_t structure which is copied back out
**	to user space by poolioctl upon return from this function.  If the
**	label is not already loaded, pool_add finds an empty pool_t structure
**	in the pools array and allocates it the new pool.  It then reads in the
**	pool information and returns the minor number assigned to this new
**	pool.  See the pool_t definition in pool_head.h for a description of
**	the locking protocol for the pool structure.
**
*******************************************************************************
******************************************************************************/

int
pool_add(pool_add_t *add)
{
  pool_t     		*new_pp = NULL;
  int         		 i;
  int         		 ret;
  int			 pl;


  IFDEBUG(POOL_DEBUG1,
	  printk("Entered pool_add()\n");
	  printk("  Adding label: (%d,%d) 0x%x\n", GMAJOR(add->pa_label_dev),
		 GMINOR(add->pa_label_dev), add->pa_label_dev););

  /* Check if the pool label device exists. */
  
  for (i=1; i < POOL_MAX_DEVICES; i++) /* start at 1 to skip comm device */ 
  {
    POOL_SLOCK(pl,&pools[i].p_mutex, pldisk);
    if (pools[i].p_flags & PF_FREE) 
    {
      POOL_SUNLOCK(&pools[i].p_mutex, pl);
      continue;
    }
    if (add->pa_label_dev == pools[i].p_label_dev)
    {
      if (pools[i].p_flags & PF_VALID) 
      {
	POOL_SUNLOCK(&pools[i].p_mutex, pl);
	add->pa_pool_dev = makedevice(POOL_MAJOR,i);
	return 0;
      }
      else
      {
	POOL_SUNLOCK(&pools[i].p_mutex, pl);
	add->pa_pool_dev = makedevice(POOL_MAJOR,i);
	return KRET(EBUSY);
      }
    }
    else
      {
	POOL_SUNLOCK(&pools[i].p_mutex, pl);
      }
  }

  /* the device is not already configured so let's assign a minor number to 
   * it */

  for (i=1; i < POOL_MAX_DEVICES; i++) 
  {
    POOL_SLOCK(pl,&pools[i].p_mutex, pldisk);
    if (pools[i].p_flags & PF_FREE && pools[i].p_flags & PF_VALID) 
    {
      pools[i].p_flags = PF_INIT; /* clear valid and free flags */
      ASSERT(!POOL_BUSY(&pools[i]));
      /* clear out all the old stats */
      new_pp = &pools[i];
      memset(&new_pp->p_stats, 0, sizeof(pool_stat_t));
      memset(new_pp->p_dstats, 0, sizeof(pool_dstat_t) * POOL_DSTATS_MAX);
      memset(&new_pp->p_dlstats, 0, sizeof(pool_dlstat_t));

      new_pp->p_label_dev = add->pa_label_dev;
      POOL_SUNLOCK(&pools[i].p_mutex, pl);
      break;
    } 
    else 
    {
      POOL_SUNLOCK(&pools[i].p_mutex, pl);
    }
  }
  
  if (i == POOL_MAX_DEVICES)
    {
      return KRET(ENOMEM);
    }

  add->pa_pool_dev = makedevice(POOL_MAJOR, i);

  IFDEBUG(POOL_DEBUG_ADD,
	  printk("pool_add: generating pool\n"););
  
  ret = gen_pool(add, new_pp);

  if (ret != 0) {
    POOL_SLOCK(pl,&new_pp->p_mutex, pldisk);
    new_pp->p_flags |= PF_VALID | PF_FREE;
    POOL_SUNLOCK(&new_pp->p_mutex, pl);
    return ret;
  }

#ifdef linux
  ret = open_devices(new_pp);

  if (ret != 0) {
    POOL_SLOCK(pl, &new_pp->p_mutex, pldisk);
    new_pp->p_flags |= PF_VALID | PF_FREE;
    POOL_SUNLOCK(&new_pp->p_mutex, pl);
    return ret;
  }
#endif

  IFDEBUG(POOL_DEBUG_ADD,
	  print_pool(new_pp););

  POOL_SLOCK(pl,&new_pp->p_mutex, pldisk);
  new_pp->p_flags |= PF_VALID | PF_LOWEROPEN;
  POOL_SUNLOCK(&new_pp->p_mutex, pl);

  return (0);
}


/*****************************************************************************
 *****************************************************************************
 **
 ** int gen_pool(pool_add_t *add, pool_t *pool)
 **
 ** Description:
 **   This routine takes a pointer to a pool_add_t structure and a pool_t
 **   structure and copies the data from the pointers to user space in the add
 **   struct to the pool struct. This includes the pool label, subpool
 **   structures, device structures, and dlockdev structures.  A minimum of
 **   error checking is performed.
 **
 ** Input(s): 
 **   pointer to add structure.  This structure must be in kernel
 **   space, but ** the pointers in the add structure may be in user or kernel
 **   space
 **
 ** Output(s):
 **   a filled in pool_t structure
 **
 ** Returns:
 **   EFAULT, EINVAL, or 0 upon success
 **
 *****************************************************************************
 *****************************************************************************/

/* ARGSUSED */
static int 
gen_pool(pool_add_t *add, pool_t *pool)
{
  pool_labels_t	*labels	= &pool->p_label;
  subpool_t	*spoolp = NULL;
  int		j;
  int		errno	= 0;  
  int		size;

  IFDEBUG(POOL_DEBUG_ADD,
	  printk("Entered gen_pool()\n"););

  /* copy in the label */
    
  if (copyin((void *)add->pa_label, (void *)labels, sizeof(pool_labels_t)))
  {
    IFDEBUG(POOL_DEBUG_ADD,
	    printk("gen_pool: fault reading 0x%p (pool label)\n", 
		   add->pa_label););
    return KRET(EFAULT);
  }

  /* Perform a few simple tests.  Most error checking should be done by 
   * passemble.  Here we just check the magic number and the pool version.
   * The on disk pool label must match down to the subrevison.  Thus, versions
   * 0.3.0 and 0.4.0 don't match and are assumed to have different on disk
   * structures.   On the other hand opening a 0.4.2 label with a 0.4.1 or
   * 0.4.5 kernel is not an error */

  if (labels->pl_magic != POOL_MAGIC)
  {
    printk("gen_pool: Error - bad magic number 0x%"FMT_64"x\n",	
	   labels->pl_magic);

    return KRET(EINVAL);
  }
  if ((labels->pl_version >> 16) != (POOL_VERSION >> 16) || 
      ((labels->pl_version >> 8) & 0xFF) != ((POOL_VERSION >> 8) & 0xFF)) 
  {
    printk("gen_pool: version mismatch in pool: %s:\n", labels->pl_devlabel);
    printk("gen_pool: \ton disk: v%d.%d.%d\n", labels->pl_version >> 16,
	   (labels->pl_version >> 8) & 0xFF, labels->pl_version & 0xFF);
    printk("gen_pool: \tin core: v%d.%d.%d\n", POOL_VERSION >> 16,
	   (POOL_VERSION >> 8) & 0xFF, POOL_VERSION & 0xFF);
    return KRET(EINVAL);
  }
  
  /* ok, we assume the label is valid, copy in the rest */
    
  /*
   *  copy data from user structure to dlockdev_t
   */

  size = labels->pl_dldevices*sizeof(dlockdev_t);
  if (size > 127*1024)
    pool->p_dlockdevs = (dlockdev_t *)vmalloc(size);
  else
    pool->p_dlockdevs = (dlockdev_t *)kmem_alloc(size, KM_SLEEP);

  if (copyin(add->pa_dlockdevs, pool->p_dlockdevs, size)) 
  {
    IFDEBUG(POOL_DEBUG_ADD,
	    printk("pool_add: fault reading 0x%p (pl_dldevices)\n", 
		   add->pa_dlockdevs););
    if (size > 127*1024)
      vmem_free(pool->p_dlockdevs, size);
    else
      kmem_free(pool->p_dlockdevs, size);
      
    pool->p_dlockdevs = NULL;
    return KRET(EFAULT);
  }

  /* make sure we're properly linked with prisa if we're opening a prisa pool 
   */

#ifdef IRIX
  for (j=0; j < labels->pl_dldevices; j++)
    {
      kdev_t	dev = pool->p_dlockdevs[j].dl_dev;
      if (pool_fc_open == NULL && GMAJOR(dev) == POOL_PRISA)
	{
	  kmem_free(pool->p_dlockdevs, size);
	  pool->p_dlockdevs = NULL;
	  return KRET(ENOPKG);
	}
    }
#endif
  
  /*
   *  copy data from user add structure to subpool_t
   */

  size = labels->pl_subpools*sizeof(subpool_t);
  pool->p_subpoollist = (subpool_t *) kmem_alloc(size, KM_SLEEP);

  if (copyin(add->pa_spools, pool->p_subpoollist, size))
  {	
    IFDEBUG(POOL_DEBUG_ADD,
	    printk("pool_add: fault reading 0x%p (subpoollist)\n",
		   add->pa_spools););
    kmem_free(pool->p_dlockdevs, labels->pl_dldevices*sizeof(dlockdev_t));
    kmem_free(pool->p_subpoollist, size);
    pool->p_dlockdevs = NULL;
    pool->p_subpoollist = NULL;
    return KRET(EFAULT);
  }
  
  /*
   *  fill in the devices that constitute the subpools 
   */

  for (j=0; j<labels->pl_subpools; j++)
  {
    device_t	*user_copy;
    
    spoolp = &(pool->p_subpoollist[j]);

    /* generate a IRIX64 address if needbe. */
#if defined(IRIX) && _MIPS_SIM == _ABI64
    user_copy = (device_t *)((app64_ptr_t)spoolp->sp_udevlist.u_user);
#else
    user_copy = spoolp->sp_devlist;
#endif

    size = spoolp->sp_devices*sizeof(device_t);
    spoolp->sp_devlist = (device_t *)kmem_alloc(size, KM_SLEEP);

    if (copyin(user_copy, spoolp->sp_devlist, size))
    {
      IFDEBUG(POOL_DEBUG_ADD,
	    printk("pool_add: mem fault reading subpool %d devlist\n", j););
      errno = EFAULT;
      break;
    }

    IFDEBUG(POOL_DEBUG_ADD,
	    int i;
	    printk("Added subpool: %d  ", j);
	    for (i=0;i<spoolp->sp_devices; i++) {
	      printk("Device %d: (%d,%d)\n", j, 
		     GMAJOR(spoolp->sp_devlist[i].d_dev), 
		     GMINOR(spoolp->sp_devlist[i].d_dev));
	    });
  }

  /* 
   * back out the zalloc's if there was an error
   */

  if (errno)
  {
    for (; j >= 0; j--)
    {
      spoolp = &(pool->p_subpoollist[j]);
      kmem_free(spoolp->sp_devlist, spoolp->sp_devices*sizeof(device_t));
    }
    kmem_free(pool->p_dlockdevs, sizeof(dlockdev_t) * labels->pl_dldevices);
    kmem_free(pool->p_subpoollist, labels->pl_subpools*sizeof(subpool_t));
    pool->p_dlockdevs = NULL;
    pool->p_subpoollist = NULL;
    return errno;
  }    

  return (0);
}


/******************************************************************************
*******************************************************************************
**
** int pool_remove(uint32_t dev)
**
** Description:
**	This routine checks that a pool is not in use and then removes it from
**	the list of managed pools.
**
*******************************************************************************
******************************************************************************/

int
pool_remove(uint32_t dev)
{
  int 		pl;
  int		rc;
  pool_t	*pp;

  IFDEBUG(POOL_DEBUG1,
	  printk("Entered pool_remove\n"););

  pp = &pools[dev];

  POOL_SLOCK(pl,&pp->p_mutex, pldisk);
  if (pp->p_flags & PF_FREE) {

    POOL_SUNLOCK(&pp->p_mutex, pl);
    return KRET(EINVAL);
  }
  
  if (!(pp->p_flags & PF_VALID) || POOL_BUSY(pp)) {

    POOL_SUNLOCK(&pp->p_mutex, pl);
    return KRET(EBUSY);
  }

#ifdef linux
  fsync_dev(MKDEV(POOL_MAJOR,dev));

  if (pp->p_flags & PF_LOWEROPEN) {
    close_devices(pp);
  }
#endif
    
  rc = free_pool(pp);
  pp->p_flags = PF_VALID | PF_FREE;
  POOL_SUNLOCK(&pp->p_mutex, pl);
  IFDEBUG(POOL_DEBUG_REMOVE,
	  printk("pool (%d,%d) removed\n", POOL_MAJOR, dev););
  return (rc);
}


/******************************************************************************
*******************************************************************************
**
** int pool_list(pool_list_t *list)
**
** Description:
**	This routine runs through the list of in-use pools and returns
**	information to the caller.  It provides a mapping between pool devices
**	and label devices and provides a versioning system by returning the
**	unix time at which the pool was last read in.  It also provides
**	information about whether the device is in use or not.
**
*******************************************************************************
******************************************************************************/

int
pool_list(pool_list_t *list)
{
  int			 i,j,k;
  int			 pl;
  pool_status_t		*stats;
  pool_t		*pp;

  IFDEBUG(POOL_DEBUG1,
	  printk("Entered pool_list\n"););

  if (list->pl_count > POOL_MAX_DEVICES)
    list->pl_count = POOL_MAX_DEVICES;

  stats = (pool_status_t*) 
    kmem_alloc(list->pl_count * sizeof(pool_status_t), KM_SLEEP);
  memset(stats, 0, list->pl_count * sizeof(pool_status_t));

  for (i=1,j=0; i<POOL_MAX_DEVICES; i++) 
    {
      pp = &pools[i];
      POOL_SLOCK(pl,&pp->p_mutex, pldisk);
      if (pp->p_flags & PF_FREE || !(pp->p_flags & PF_VALID)) 
	{
	  POOL_SUNLOCK(&pp->p_mutex, pl);
	  continue;
	}
      stats[j].ps_label_major = (uint32_t)GMAJOR(pp->p_label_dev);
      stats[j].ps_label_minor = (uint32_t)GMINOR(pp->p_label_dev);
      stats[j].ps_pool_major = (uint32_t)POOL_MAJOR;
      stats[j].ps_pool_minor = (uint32_t)i;
      stats[j].ps_pool_id = pp->p_label.pl_pool_id;
      for(k=0; k< POOL_NAME_SIZE; k++)
	stats[j].ps_pool_name[k] = pp->p_label.pl_devlabel[k]; 
      if (POOL_BUSY(pp)) 
	{
	  stats[j].ps_in_use = 1;
	}
      else
	{
	  stats[j].ps_in_use = 0;
	}
      POOL_SUNLOCK(&pp->p_mutex, pl);

      IFDEBUG(POOL_DEBUG_LIST,
	      printk("  label device: (%d,%d)\n", stats[j].ps_label_major,
		     stats[j].ps_label_minor);
	      printk("  pool device: (%d,%d)\n", stats[j].ps_pool_major,
		     stats[j].ps_pool_minor);
	      printk("  pool id: 0x%"FMT_64"x\n", stats[j].ps_pool_id);
	      printk("  device is %sin use\n", stats[j].ps_in_use ? "" : "not ");
	      );

      if (++j >= list->pl_count)
	break;
    }
  list->pl_count = j;

  if (!(j > 0)) 
  {
    kmem_free(stats, list->pl_count * sizeof(pool_status_t));
    return 0;
  }

  if (copyout(stats, list->pl_list, (int)(sizeof(pool_status_t) * j)))
  {
    kmem_free(stats, list->pl_count * sizeof(pool_status_t));
    return KRET(EFAULT);
  }

  kmem_free(stats, list->pl_count * sizeof(pool_status_t));
  return 0;
}


/******************************************************************************
*******************************************************************************
** NAME
**	pool_splist
**
** DESCRIPTION
**	Generate a list of subpools for a given pool and copy their sizes out
**	to user space
**
** RETURN VALUE
**	Standard unix errno's
**
*******************************************************************************
******************************************************************************/

int
pool_splist(pool_t *pp, pool_splist_t *param)
{
  pool_spstatus_t	*list;
  subpool_t		*spool;
  int	i,j;
  int	max;
  device_t 		*devp;
  
  IFDEBUG(POOL_DEBUG1,
	  printk("Entering pool_splist\n"););

  max = pp->p_label.pl_subpools > param->spl_count ? param->spl_count : 
    pp->p_label.pl_subpools;

  list = kmem_alloc(max * sizeof(pool_spstatus_t), KM_SLEEP);
  memset(list, 0, max * sizeof(pool_spstatus_t));
  
  for(i=0; i < max; i++)
    {
      spool = &pp->p_subpoollist[i];
      
      list[i].sps_boffset = spool->sp_start;
      list[i].sps_bcount = spool->sp_len;
      list[i].sps_stripe = spool->sp_striping;
      list[i].sps_numdevs = spool->sp_devices;

      if (spool->sp_devices <= 64){ /* hack so we can get dev list for pinfo */
	for(j=0; j< spool->sp_devices; j++){
	  devp = &(spool->sp_devlist[j]);
	  list[i].sps_minors[j] = MINOR(devp->d_dev);
	}
      }    
    }

  if (copyout(list, param->spl_list, (int)(sizeof(pool_spstatus_t) * max)) != 0)
    {
      kmem_free(list, sizeof(pool_spstatus_t) * max);
      return KRET(EFAULT);
    }

  param->spl_count = max;
  param->spl_devlocks = pp->p_label.pl_dlocks;

  kmem_free(list, sizeof(pool_spstatus_t) * max);
  return (0);
}

/******************************************************************************
*******************************************************************************
** NAME
**	free_pool()
**
** DESCRIPTION:  free memory associated with the pool_t
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

int 
free_pool(pool_t *pool)
{
  pool_labels_t		*label = &pool->p_label;
  subpool_t           	*spoolp;
  int                 	j, size;


  IFDEBUG(POOL_DEBUG2,
	  printk("Entered free_pool\n");
	  printk("  free_labels:  0x%p\n",label);
	  printk("  pl_subpools    = %u\n",label->pl_subpools););

  /*
   *  free data from dlockdev_t
   */
  IFDEBUG(POOL_DEBUG2,
	  printk("Freeing dlockdev 0x%p\n", pool->p_dlockdevs););
  size = label->pl_dldevices*sizeof(dlockdev_t);
  if (size > 127*1024)
    vmem_free(pool->p_dlockdevs, size);
  else
    kmem_free(pool->p_dlockdevs, size);
  pool->p_dlockdevs = NULL;
  
  /*
   *  free data from subpool_t
   */
  for (j=0; j<label->pl_subpools; j++)
  {
    spoolp = &(pool->p_subpoollist[j]);
    IFDEBUG(POOL_DEBUG2,
	    printk("Freeing devices 0x%p\n", spoolp->sp_devlist););

    kmem_free(spoolp->sp_devlist, spoolp->sp_devices*sizeof(device_t));
  }

  IFDEBUG(POOL_DEBUG2,
	  printk("Freeing subpools 0x%p\n", pool->p_subpoollist););

  kmem_free(pool->p_subpoollist, label->pl_subpools * sizeof(subpool_t));
  pool->p_subpoollist = NULL;

  return(0);
}



/******************************************************************************
*******************************************************************************
** NAME
**	print_pool()
**
** DESCRIPTION:  print labels for debugging
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/
#ifdef DEBUG
int 
print_pool(pool_t *pp)
{
  pool_labels_t		*label = &pp->p_label;
  dlockdev_t          	*dlockp;
  subpool_t           	*spoolp;
  device_t            	*devp;
  int                 	i,j;

  printk("\n\nPool Labels:\n\n");

  printk("pl_magic       = 0x%"FMT_64"x\n",label->pl_magic);
  printk("pl_labelsize   = %u\n",label->pl_labelsize);
  printk("pl_version     = v%u.%u.%u\n", label->pl_version >> 16,
	 (label->pl_version >> 8) & 0xFF, label->pl_version & 0xFF);
  printk("pl_devlabel    = %s\n",label->pl_devlabel);
  printk("pl_subpools    = %u\n",label->pl_subpools);

  printk("pl_dlocks      = %u\n",label->pl_dlocks);
  printk("pl_dldevices   = %u\n",label->pl_dldevices);

  printk("pl_devices     = %u\n",label->pl_devices);
  printk("pl_capacity    = %"FMT_64"u\n",label->pl_capacity);
  printk("p_oblk         = %u\n", pp->p_oblk);
#ifdef IRIX
  printk("p_ochr         = %u\n", pp->p_ochr);
  printk("p_olyr         = %u\n", pp->p_oblk);
#endif
  printk("p_flags        = 0x%x\n", pp->p_flags);
  printk("p_label_dev    = (%u,%u)\n", GMAJOR(pp->p_label_dev), 
	 GMINOR(pp->p_label_dev));
  printk("\n");


  for (j=0; j<label->pl_dldevices; j++)
  {
    dlockp = &(pp->p_dlockdevs[j]);

    printk("Dlock device %d:\n", j);
    printk("    dl_dev      = (%u,%u)\n", GMAJOR(dlockp->dl_dev), 
	   GMINOR(dlockp->dl_dev));
    printk("	dl_wwn	    = 0x%"FMT_64"x\n", dlockp->dl_wwn);
    printk("    dl_start    = %u\n",dlockp->dl_start);
    printk("    dl_number   = %u\n",dlockp->dl_number);
    printk("\n");
  }

  printk("\n");

  for (j=0; j<label->pl_subpools; j++)
  {
    spoolp = &(pp->p_subpoollist[j]);

    printk("Subpool %d:\n", j);
    printk("    sp_start    = %"FMT_64"u\n",spoolp->sp_start);
    printk("    sp_len      = %u\n",spoolp->sp_len);
    printk("    sp_striping = %u\n",spoolp->sp_striping);
    printk("    sp_devices  = %u\n",spoolp->sp_devices);
    printk("\n");


    for (i=0; i<spoolp->sp_devices; i++)
    {
      devp = &(spoolp->sp_devlist[i]);
      printk("dev 0x%p\n", devp);

      printk("    Device %d:\n", i);
      printk("        d_dev   = (%d,%d)\n", GMAJOR(devp->d_dev),
	     GMINOR(devp->d_dev));
      printk("        d_start = %"FMT_64"u\n",devp->d_start);
      printk("        d_nblks = %"FMT_64"u\n",devp->d_nblks);
    }
    printk("\n\n");
  }

  return(0);
}
#endif /* DEBUG */

/******************************************************************************
*******************************************************************************
** NAME
**	pool_mapping
**
** DESCRIPTION  psector in basic blocks, pbcount in bytes
**
** RETURN VALUE 
**
*******************************************************************************
******************************************************************************/

int
pool_mapping(pool_t *pp, uint64 psector, uint32 pbcount, 
	     uint64 *dev_sector, uint32 *dev_bcount, device_t **out_dev)
{
  pool_labels_t		*labels = &pp->p_label;
  subpool_t           	*spoolp = NULL;
  device_t            	*devp	= NULL;
  int			i,j;
  
  uint32		current_spool = 0;
  uint32		current_dev;
  
  uint64		blk;
  /*  
  unsigned		s;
  uint64		o;
  */
  uint32		devnum = 0;    /* # devices in selected spool */
  
  IFDEBUG(POOL_DEBUG_MAP,
	  printk("Entered pool_mapping: psec=%"FMT_64"u, pbcnt=%u\n",
		 psector, pbcount););

  /* select sub-pool */

  for (j=current_spool; j<labels->pl_subpools; j++) {
      spoolp = &(pp->p_subpoollist[j]);

      if ((psector >= spoolp->sp_start) &&
	  (psector < spoolp->sp_start+spoolp->sp_len)) {
	break;
      }
  }
  if (j == labels->pl_subpools) {
    printk("block not in any subpools at line %d!\n", __LINE__);
    return KRET(EIO);
  }
  
  current_spool = j;
  devnum = spoolp->sp_devices;

  IFDEBUG(POOL_DEBUG_MAP,
	  printk("  spl=%u, ", current_spool););


  /* Determine device within selected sub-pool */

  if (spoolp->sp_striping) {

    IFDEBUG(POOL_DEBUG_MAP,
	    printk("str=%u, ", spoolp->sp_striping););

    if ((spoolp->sp_striping<<BBSHIFT) % (1<<BBSHIFT) != 0) {
      printk("stripe width must be a mutliple of the basic block");
      return KRET(EIO);
    }
    current_dev = (uint32)(((psector - spoolp->sp_start) 
			    / spoolp->sp_striping) % devnum);
  } else {
    for (i=0; i<devnum; i++) {
      devp = &(spoolp->sp_devlist[i]);
      /*
      if ((psector - spoolp->sp_start - devp->d_start) < devp->d_nblks) {
	break;
	}*/
      if (((psector - spoolp->sp_start) >= devp->d_start) &&
	  ((psector - spoolp->sp_start) < (devp->d_start + devp->d_nblks)))
	break;
    }
    if (i == devnum) {
      printk("\nblock requested past end of subpool at line %d!\n", __LINE__);
      IFDEBUG(POOL_DEBUG_MAP,
	      printk("psector: %"FMT_64"d, sp_start: %"FMT_64"d, d_start: "
		     "%"FMT_64"d, d_nblks: %"FMT_64"d\n", psector, 
		     spoolp->sp_start, devp->d_start, devp->d_nblks););
      return KRET(EIO);
    }

    current_dev = i;
  }
  IFDEBUG(POOL_DEBUG_MAP,
	  printk("dev=%u, ", current_dev););


  devp = &(spoolp->sp_devlist[current_dev]);


  /* Determine the block number on the device */

  if (spoolp->sp_striping) {
    blk = (((psector - spoolp->sp_start) / 
	    (spoolp->sp_striping*spoolp->sp_devices)) *
	   spoolp->sp_striping) + 
      ((psector - spoolp->sp_start) % spoolp->sp_striping);
  } else {
    blk = psector - spoolp->sp_start - devp->d_start;
  }
  IFDEBUG(POOL_DEBUG_MAP,
	  printk("blk=%"FMT_64"u\n", blk););


  /* Determine the number of bytes on this device */
  /*
  if (spoolp->sp_striping) {

    s = (uint32)((blk % spoolp->sp_striping) << BBSHIFT);

    if ((pbcount + s) >= spoolp->sp_striping << BBSHIFT)
      *dev_bcount = (spoolp->sp_striping << BBSHIFT) - s;
    else
      *dev_bcount = pbcount;
  } else {
    o = blk << BBSHIFT;
    if ((pbcount + o) >= (devp->d_nblks << BBSHIFT))
      *dev_bcount = (unsigned)((devp->d_nblks << BBSHIFT) - o);
    else
      *dev_bcount = pbcount;
  }
  */

  /* account for partition header */

  blk += POOL_SIZE_PHEADER(sizeof(disk_partition_header_t));

  *dev_sector = blk;
  *out_dev = devp;

  return (0);
}

/******************************************************************************
 ******************************************************************************
 **
 ** static int lock_mapping()
 **
 ** Description:
 **   This routine maps logical lock numbers to physical device lock numbers.
 **
 ** Input(s):
 **  *pp     - 
 **   logln  - 
 **  *params - 
 **
 ** Output(s):
 **
 **
 ** Returns:
 **   0 if OK, error condition on error.
 **
 ******************************************************************************
 *****************************************************************************/

int 
lock_mapping(pool_t *pp, dlock_params_t *params) 
{
  pool_labels_t        	*labels = &pp->p_label;
  dlockdev_t		*dlockdevp;
  uint32		logln;
  int			i;

  IFDEBUG(POOL_DEBUG_LOCK1,
	  printk("Entering lock_mapping\n"););

  if ((logln = params->dp_ctlarg->l_id) >= labels->pl_dlocks)
  {
    printk("lock_mapping: lock number too large\n");
    return KRET(EINVAL);
  }

  if (labels->pl_dlstriping == 0)
  {
    for (i=0; i<labels->pl_dldevices; i++)
    {
      dlockdevp = &(pp->p_dlockdevs[i]);

      if (logln < dlockdevp->dl_start+dlockdevp->dl_number)
      {
	params->dp_devnum = i;
	params->dp_dldev = dlockdevp;
	params->dp_phyln = (logln - dlockdevp->dl_start);
	return(0);
      }
    }
  }
  else
  {
    params->dp_devnum = 
      (logln/labels->pl_dlstriping) % (labels->pl_dldevices); 
    params->dp_dldev = &(pp->p_dlockdevs[params->dp_devnum]);
    params->dp_phyln = 
      (logln%labels->pl_dlstriping) + 
	(((logln/labels->pl_dlstriping) / (labels->pl_dldevices)) *  
	 ((labels->pl_dlstriping)));

    return(0);
  }
  printk("lock_mapping: unknown\n");
  return KRET(EINVAL);
}


/******************************************************************************
 ******************************************************************************
 **
 ** int pool_dlock()
 **
 ** Description:
 **   This routine performs device lock operations requested by pool ioctls's.
 **   PL_NDELAY instructs pool_dlock to attempt the lock once and return the
 **   results.  PL_TRY instructs pool_dlock to attempt the lock until one
 **   quarter of pool_reset_timeout_us has expired, then the results are
 **   returned.  PL_RESET instructs pool_dlock to get the lock.  This implies
 **   that if necessary the lock will be reset after pool_reset_timeout_us has
 **   expired and no activity is present.  PL_EXP can be used with any of these
 **   options.  It instructs pool_dlock to exponentially back off if it isn't
 **   getting the lock.  The function returns the lock counter value, a boolean
 **   (timeout) indicating whether the timeout operation was performed on this
 **   lock request, and cachable, which is a boolean indicating whether the
 **   clock has changed.  Since the clock can be changed by another process on
 **   this machine and cachable would still be set, the use of cachable is
 **   depricated.  The timeout operation for PL_NDELAY is a return without the
 **   lock.  The timeout operation for PL_TRY is a return without the lock.
 **   The timeout operation for PL_RESET is a return with a lock that has been
 **   reset.
 **
 **
 ** Input(s):
 **  *pp     - 
 **  *params - 
 **
 ** Output(s):
 **
 **
 ** Returns:
 **   0 if OK, error condition on error.
 **
 ******************************************************************************
 ******************************************************************************/
int 
pool_dlock(pool_t *pp, dlock_params_t *params)
{
  uint32       		act;
  uint32		flg;
  uint32		clk;
  uint32		pln;
  uint32		lln;
  dlockdev_t   	       *dev;

  lock_ret_t		ret;

  uint32		delay    = pool_cnf_delay_us;
  uint32		delayed  = 0;
  uint32		timeout  = 0;
  uint32		activity = 0;
  uint32		success  = 0;
  uint32		cache    = 0;
  int			pl;		/* for SLOCK */
  time_t		ts;

  act = params->dp_action;		/* lock, unlock, ulincr */
  flg = params->dp_ctlarg->l_flags;	/* ndelay, try, reset, exp */
  lln = params->dp_ctlarg->l_id;	/* logical lock num */
  pln = params->dp_phyln;		/* physical lock num */
  dev = params->dp_dldev;		/* lock device dev_t */

  POOL_SLOCK(pl,&pp->p_dlmutex, pldisk);
  clk = pp->p_locktab[lln].l_clock; 	/* clock for lln */
  POOL_SUNLOCK(&pp->p_dlmutex, pl);

  IFDEBUG(POOL_DEBUG_LOCK1,
	  printk("Entered pool_dlock: lock=%d, act=0x%x, flg=0x%x, clk=%d\n", 
		 lln, act, flg, clk););

  if (act == POOL_LOCK) {
    
    ret = device_lock(act, pln, clk, dev);
    if (ret.r_status) return KRET(EIO);
    success = ret.r_result;
    clk = ret.r_locktab.l_clock;

    if (flg & (PL_TRY | PL_RESET)) {
      
      while (!success && !timeout) {
	
	us_delay(delay);
	delayed += delay;
	if (flg & PL_EXP) {
	  delay <<= 1;
	  if (delay > 1000000) delay = 1000000; 
	}

	IFDEBUG(POOL_DEBUG_LOCK1,	
		printk("  lock conflict.  delayed=%d, activ=%d, clk=%d\n",
		       delayed, activity, clk););
	
	ret = device_lock(act, pln, clk, dev);
	if (ret.r_status) return KRET(EIO);
	success = ret.r_result;

	if (activity && (ret.r_locktab.l_clock != clk)) {
	  delayed = 0;		       
	  delay = pool_cnf_delay_us;
	  activity = 0;
	  ret = device_lock(POOL_ACTIVITY_OFF, pln, clk, dev);	  
          if (ret.r_status) return KRET(EIO);
	}

	clk = ret.r_locktab.l_clock;
	activity = ret.r_locktab.l_activity;

	if (!success) {
	  if (delayed >= pool_reset_timeout_us / 4) {
	    if (flg & PL_TRY)
	      timeout = 1;
	    else {
	      if (delayed >= pool_reset_timeout_us)
		timeout = 1;
	      else if (!activity) {
		ret = device_lock(POOL_ACTIVITY_ON, pln, clk, dev);
		if (ret.r_status) return KRET(EIO);
		clk = ret.r_locktab.l_clock;
		activity = 1;
	      }	      
	    }
	  }
	}
      }		/* while */
    
      if (timeout && (flg & PL_RESET)) { 	
	printk("Resetting lock %d. clk(in)=%d, activ=%d\n", 
	       lln, clk, activity);
	ret = device_lock(POOL_RESET, pln, clk, dev);
	if (ret.r_status) return KRET(EIO);
	ret = device_lock(act, pln, clk, dev);
	if (ret.r_status) return KRET(EIO);
	if (activity) {
	  ret = device_lock(POOL_ACTIVITY_OFF, pln, clk, dev);
          if (ret.r_status) return KRET(EIO);
	  clk = ret.r_locktab.l_clock;
	}
      }          
    }
    else if (flg & PL_NDELAY) {

      if (!ret.r_result) timeout = 1;
      
    }
  }
  else {  /* action is POOL_UNLOCK or UNLOCK_INCR */
    ts = timestamp();
    POOL_SLOCK(pl,&pp->p_dlmutex, pldisk);
    pp->p_locktab[lln].l_ultime = ts;
    POOL_SUNLOCK(&pp->p_dlmutex, pl);
    ret = device_lock(act, pln, clk, dev);
    if (ret.r_status) return KRET(EIO);
  }

  ts = timestamp();

  POOL_SLOCK(pl,&pp->p_dlmutex, pldisk);
  if ((ts - pp->p_locktab[lln].l_ultime) > DLOCK_ROLL_TIME)
    cache = 0;
  else
    cache = (ret.r_locktab.l_clock == pp->p_locktab[lln].l_clock);
  POOL_SUNLOCK(&pp->p_dlmutex, pl);

  /* output values */

  params->dp_ctlarg->l_timeout  = timeout;
  params->dp_ctlarg->l_counter  = ret.r_locktab.l_clock;
  params->dp_ctlarg->l_cacheable = cache;

  POOL_SLOCK(pl,&pp->p_dlmutex, pldisk);

  pp->p_locktab[lln].l_state    = ret.r_locktab.l_state;
  pp->p_locktab[lln].l_activity = ret.r_locktab.l_activity;
  pp->p_locktab[lln].l_clock    = ret.r_locktab.l_clock;

  POOL_SUNLOCK(&pp->p_dlmutex, pl);

  IFDEBUG(POOL_DEBUG_LOCK1,
	  printk("Exiting pool_dlock\n");
	  printk("  timeout: %d\n", timeout);
	  printk("  state:   %d\n", ret.r_locktab.l_state);
	  printk("  clock:   %d\n", ret.r_locktab.l_clock);
	  printk("  activ:   %d\n", ret.r_locktab.l_activity);
	  printk("  cache:   %d\n", params->dp_ctlarg->l_cacheable););
  

  return 0;
}



