/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@lcse.umn.edu>
** 
**    Module name: assemble.c
**
**    Description:
**	Network storage pool assembler routines.
**
*******************************************************************************
******************************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <sys/dir.h>
#include <sys/stat.h>
#include <paths.h>
#include <ctype.h>
#include <assert.h>

#include "global.h"
#include "assemble.h"


int		ops_flag;
char		*destroy = NULL;
label_list_t	lalist;



/******************************************************************************
*******************************************************************************
** NAME
**	main
**
** DESCRIPTION
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

int main(int argc, char *argv[])
{
  char			 path[256];
  char			*remove = NULL;
  int			 commfd;
  int			 c, i;
  pool_list_t		 inuse_list;
  pool_add_t		*pool = NULL;
  struct stat	       	 statbuf;
  char			 response[3];

  
  ops_flag = 0;
  lalist.l_cnt = 0;
  inuse_list.pl_count = 0;

  while ((c = getopt(argc, argv, "hr:fve:")) != -1) {
    switch (c) {

    case 'h':
      printUsage(argv[0], EXIT_SUCCESS);
      exit(0);
      break;
    case 'f':
      ops_flag |= AS_FORCE;
      break;
    case 'v':
      ops_flag |= AS_VERBOSE;
      break;
    case 'r':
      ops_flag |= AS_REMOVE;
      remove = optarg;
      if (strcmp(optarg,"all") == 0) {
	remove = NULL;
      }
      break;
    case 'e':
      ops_flag |= AS_DESTROY;
      destroy = optarg;

      fprintf(stderr, "Warning!  About to destroy %s on disk.\n"
	      "Are you certain you want to continue? [yN] ", destroy);
      scanf("%s", response);

      if (tolower(response[0]) != 'y') {
	fprintf(stderr, "Destroy %s canceled.\n", destroy);
	return(-1);
      }
      break;	 
    default:
      printUsage(argv[0], EXIT_FAILURE);
      exit(1);
    }
  }



  /* Set up pool directory and poolcomm device */

  strcpy(path, _PATH_DEV);
  strcat(path, "pool");
  mkdir(path, 0755);

  mknod("/dev/poolcomm", 0600 | S_IFBLK, makedevice(POOL_MAJOR, 0));
  
  if ((commfd = open("/dev/poolcomm", O_RDONLY)) < 0) {
    fprintf(stderr, "Couldn't open comm device: %s.\n", strerror(errno));
    fprintf(stderr, "Be sure pool module is installed.\n");
    exit(1);
  }
  clean_dir(commfd);


  /* We have comm device open.  This is exclusive, so we know all addition
   * and removal of pools is serialized through this program. */
  /* Get status of all pools in the kernel */

  pool_list(&inuse_list, commfd);



  /* Remove one or all pools from kernel. Doesn't require scanning devices */

  if (ops_flag & AS_REMOVE) {
    remove_pools(&inuse_list, commfd, remove);
    clean_dir(commfd);
    exit(0);
  }



  /* Remove all in-use pools */

  if (ops_flag & AS_FORCE)
   remove_pools(&inuse_list, commfd, NULL); 



  /* Check if in-use pool is being erased. */

  if (ops_flag & AS_DESTROY) {
    for (i=0; i < inuse_list.pl_count; i++)
      if (!strcmp(inuse_list.pl_list[i].ps_pool_name, destroy)) {
	printf("Remove %s before erasing.\n", destroy);
	exit(0);
      }
  }



  /* Scan each partition out there and read its label */

  scan_partitions(NULL);       
  scan_partitions("/dev/pool/");

  if (ops_flag & AS_DESTROY) 	/* we erased labels while scanning */
    exit(0);



  /* Create pool_add structs to pass into kernel */

  while (!create_pool(&pool)) {

    if (pool == NULL)
      continue;

    if (!(ops_flag & AS_FORCE) && skip_pool(&pool, &inuse_list)) {
      printf("Pool %s already in use.\n", pool->pa_name);
      free_mem(pool);
      continue;
    }

    if (check_name(&pool, &inuse_list)) {
      printf("The name \"%s\" is already used.\n", pool->pa_name);
      free_mem(pool);
      continue;
    }

    if (ops_flag & AS_VERBOSE)
      printLabels(pool);

    if (ioctl(commfd, POOL_ADD, (void *)pool) < 0) {
      fprintf(stderr, "Adding %s (ioctl) failed: %s.\n", 
	      pool->pa_name, strerror(errno));
      fprintf(stderr, "Check the pool version.\n");
    }    
    else { 	       
       
      char buf_bdsk[1024];

      strcpy(buf_bdsk, _PATH_DEV);
      strcat(buf_bdsk, "pool/");
      strncat(buf_bdsk, pool->pa_name, POOL_NAME_SIZE);

      if (stat(buf_bdsk, &statbuf) != 0){
	if (mknod(buf_bdsk, 0600 | S_IFBLK, (dev_t)pool->pa_pool_dev) < 0)
	  fprintf(stderr, "Unable to make %s: %s.\n",
		  buf_bdsk, strerror(errno));
      }

      printf("Added %s.\n", pool->pa_name);
    }

    free_mem(pool);
  }
  
  clean_dir(commfd);
  
  return (0);
}  


/******************************************************************************
*******************************************************************************
** NAME
**	printUsage
**
** DESCRIPTION
**
*******************************************************************************
******************************************************************************/

void printUsage(char *progName, int status)
{
  printf("Usage: %s [-hvfe] [-r all|<poolname>]\n", progName);
  
  if (status != EXIT_SUCCESS)
    printf("Try `%s -h' for more information.\n", progName);
  
  if (status != EXIT_FAILURE) {
    printf("  -h         Print this help, then exit.\n");
    printf("  -v         Verbose operation.\n");
    printf("  -f         Force reload. Unload and reload all pools.\n");
    printf("  -r         Remove named or all pools from use.\n");
    printf("  -e         Erase named pool on disk.\n");
  }
  exit(status);
}


/******************************************************************************
*******************************************************************************
** NAME
**	scan_partitions
**
** DESCRIPTION
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

int scan_partitions(char *src)
{
  FILE			*fp;
  DIR 			*dir;
  struct direct 	*file;
  char			*path;
  int			fd=0;
  char		 	linebuf[256], size[10], device[10];
  int		  	major, minor, count=0;
  int		    	d_namlen, ent=0, md=0;

  if (src == NULL) {
    if ((fp = fopen("/proc/partitions", "r")) < 0) {
      fprintf(stderr, "/proc/partitions not available\n");
      return (0);
    }

    while (fgets(linebuf, 256, fp) != NULL) {
      
      if (count++ < 2)	/* skip first two lines */
	continue;

      sscanf(linebuf, "%d %d %s %s\n", &major, &minor, size, device);

      path = (char *)malloc(strlen("/dev/") + strlen(device) + 1);      
      strcpy(path, "/dev/");
      strcat(path, device);

      if (!strncmp(device, "md", 2))
	md = 1;

      if (ops_flag & AS_VERBOSE) 
	fprintf(stderr, "scan: %10s  ", path);

      if (md && skip_md_devs(device)) {
	if (ops_flag & AS_VERBOSE)
	  fprintf(stderr, "Skipping partitions used by md.\n");
	goto next;
      }

      if ((fd = open(path, O_RDWR | O_NONBLOCK)) < 0) {
	if (ops_flag & AS_VERBOSE)
	  fprintf(stderr, "not readable: %s\n", strerror(errno));
	goto next;
      }

      if (scan_label(fd, path) < 0) {
	if (ops_flag & AS_VERBOSE)
	  fprintf(stderr, "Error reading label on %s - ignoring.\n", path);
      }    
    
    next:
      free(path);
      close(fd);
    }

    fclose(fp);
    return (0);
  }

  dir = opendir(src);
  rewinddir(dir);

  while ((file = readdir(dir)) != NULL) {

    ent = strncmp(file->d_name, "pool", 4);
    if (ent != 0)
      continue;

    d_namlen = strlen(file->d_name);
            
    path = (char *)malloc(strlen(src) + d_namlen + 1);

    strcpy(path, "/dev/pool/");
    strcat(path, file->d_name);

    if (ops_flag & AS_VERBOSE) 
      fprintf(stderr, "scan: %10s  ", path);

    if ((fd = open(path, O_RDWR | O_NONBLOCK)) < 0) {
      if (ops_flag & AS_VERBOSE)
	fprintf(stderr, "not readable: %s\n", strerror(errno));
      goto next2;
    }

    if (scan_label(fd, path) < 0) {
      if (ops_flag & AS_VERBOSE)
	fprintf(stderr, "Error reading label on %s - ignoring.\n", path);
    }
    
  next2:
    free(path);
    close(fd);
  }

  closedir(dir);
  return (0);
}


/******************************************************************************
*******************************************************************************
** NAME
**	scan_label
**
** DESCRIPTION
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

int scan_label(int fd, char *devName)
{
  char	       	buf[SIZEOF_POOL_LABEL];
  pool_label_t 	*label;
  dev_t		devnum = 0;


  if (read(fd, buf, SIZEOF_POOL_LABEL) < SIZEOF_POOL_LABEL)
    return (-1);

  label = (pool_label_t *)malloc(sizeof(pool_label_t));

  
  pool_label_in(label, buf);

  if (label->pl_magic != POOL_MAGIC) {
    if (ops_flag & AS_VERBOSE)
      fprintf(stderr, "Not a pool partition.\n");
    free(label);
    return (0);
  }


  if (ops_flag & AS_DESTROY){
    if (!strcmp(label->pl_pool_name, destroy)) {
      memset(buf, 0, sizeof(pool_label_t));
      lseek(fd, 0, SEEK_SET);
      write(fd, buf, sizeof(pool_label_t));	  
      printf("Destroying %s label.\n", label->pl_pool_name);
    }	
    free(label);
    return (0);
  }


  if ((label->pl_version >> 8) != (POOL_VERSION >> 8)) {
    fprintf(stderr, "Version mismatch for %s. Label: %d.%d.%d;"
	    " Pool: %d.%d.%d\n", label->pl_pool_name, 
	    label->pl_version >> 16, (label->pl_version >> 8) & 0xFF,
	    label->pl_version & 0xFF, POOL_VERSION >> 16,
	    (POOL_VERSION >> 8) & 0xFF, POOL_VERSION & 0xFF);
    free(label);
    return (0);
  }


  if (label->pl_sp_id > label->pl_subpools - 1) {
    if (ops_flag & AS_VERBOSE)
      fprintf(stderr, "Label error (%s): spid=%d, spools=%d\n", 
	      label->pl_pool_name, label->pl_sp_id, label->pl_subpools);
    free(label);
    return (-1);
  }

  if (label->pl_sp_devid > label->pl_sp_devs - 1) {
    if (ops_flag & AS_VERBOSE)
      fprintf(stderr, "Label error (%s): devid=%d, devs=%d\n", 
	      label->pl_pool_name, label->pl_sp_devid, label->pl_sp_devs);
    free(label);
    return (-1);
  }

  if (label->pl_sp_dlid > label->pl_sp_dldevs - 1) {
    if (ops_flag & AS_VERBOSE)
      fprintf(stderr, "Label error (%s): dlid=%d, dldevs=%d\n", 
	      label->pl_pool_name, label->pl_sp_dlid, label->pl_sp_dldevs);
    free(label);
    return (-1);
  }

  if ((label->pl_blocks == 0) && (label->pl_dlocks == 0)) {
    if (ops_flag & AS_VERBOSE)
      fprintf(stderr, "Label error (%s): blocks=0, dlocks=0.\n",
	      label->pl_pool_name);
    free(label);
    return (-1);
  }


  if (ops_flag & AS_VERBOSE)
    fprintf(stderr, "Found label (%s, id: %"FMT_64"x)\n", 
	    label->pl_pool_name, label->pl_pool_id);


  if ((lalist.l_cnt % 20) == 0) {
    lalist.l_lbs = (pool_label_t *)
      realloc(lalist.l_lbs, sizeof(pool_label_t) * (lalist.l_cnt + 20));
    lalist.l_devs = (dev_t *)
      realloc(lalist.l_devs, sizeof(dev_t) * (lalist.l_cnt + 20));
  }

  if (lalist.l_lbs == NULL || lalist.l_devs == NULL) {
    fprintf(stderr, "realloc failed\n");
    exit(1);
  }

  getdevnum(devName, &devnum);

  lalist.l_lbs[lalist.l_cnt] = *label;
  lalist.l_devs[lalist.l_cnt++] = devnum;

  /*
  if (ops_flag & AS_VERBOSE){
    fprintf(stderr, "Labels found: %d\n", lalist.l_cnt);
  }
  */

  return (0);
}




/******************************************************************************
*******************************************************************************
** NAME
**	create_pool
**
** DESCRIPTION
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

int create_pool(pool_add_t **pool)
{
  pool_label_t			*la;
  pool_add_t			*np;
  subpool_t			*ns;
  dlockdev_t			*nl;
  device_t			*nd;
  int		       	       	i=0,j;
  uint64		      	spblk_off, pblk_off;
  uint64			spdl_off, pdl_off;
  uint64			pid;

  while (i < lalist.l_cnt) {
    if (lalist.l_lbs[i].pl_magic != 0)
      break;
    i++;
  }
  
  if (i == lalist.l_cnt)
    return (1); 	/* we're done */


  np = (pool_add_t *)malloc(sizeof(pool_add_t));
  memset(np, 0, sizeof(pool_add_t)); 
  np->pa_pool_id = lalist.l_lbs[i].pl_pool_id;
  np->pa_version = lalist.l_lbs[i].pl_version;
  np->pa_num_spools = lalist.l_lbs[i].pl_subpools;
  memcpy(np->pa_name, lalist.l_lbs[i].pl_pool_name, POOL_NAME_SIZE);

  np->pa_spools = (subpool_t *)malloc(np->pa_num_spools * sizeof(subpool_t));
  memset(np->pa_spools, 0, np->pa_num_spools * sizeof(subpool_t)); 

  pid = np->pa_pool_id;

  for (j=0; j < lalist.l_cnt; j++) {
    la = &(lalist.l_lbs[j]);

    if ((la->pl_magic != 0) && (la->pl_pool_id == pid)) {      
      ns = &(np->pa_spools[la->pl_sp_id]);

      ns->sp_devs = la->pl_sp_devs;
      ns->sp_dldevs = la->pl_sp_dldevs;
      ns->sp_stripe = la->pl_striping;

      if (ns->sp_devlist == NULL) {
	ns->sp_devlist = (device_t *)malloc(ns->sp_devs * sizeof(device_t));
	memset(ns->sp_devlist, 0, ns->sp_devs * sizeof(device_t)); 
      }
      if (ns->sp_dldevlist == NULL) {
	ns->sp_dldevlist = (dlockdev_t *)malloc(ns->sp_dldevs * 
						sizeof(dlockdev_t));
	memset(ns->sp_dldevlist, 0, ns->sp_dldevs * sizeof(dlockdev_t));
      }
      if (la->pl_blocks) {
	nd = &(ns->sp_devlist[la->pl_sp_devid]);
	nd->d_nblks = la->pl_blocks;
	nd->d_adev = (uint32)lalist.l_devs[j];
      }
      if (la->pl_dlocks) {
	nl = &(ns->sp_dldevlist[la->pl_sp_dlid]);
	nl->dl_nlocks = la->pl_dlocks;
	nl->dl_adev = (uint32)lalist.l_devs[j];
      }
      la->pl_magic = 0;      /* done with this label */
    }
  }


  /* make sure all data and dlock devices are found */
  /* calculate the starting block and dlock values */

  pblk_off = 0;		/* logical blk offsets - scope is entire pool */
  pdl_off = 0;	       	/* same for dlock offsets */
  spblk_off = 0;     	/* logical blk offsets - scope is one subpool */
  spdl_off = 0;		/* same for dlock offsets */

  for (i=0; i < np->pa_num_spools; i++) {
    ns = &(np->pa_spools[i]);

    ns->sp_start = pblk_off;
    ns->sp_dlstart = pdl_off;
    spblk_off = 0;
    spdl_off = 0;

    if (ns->sp_devs == 0)		/* expected a label here */
      goto exit_err;    
    
    for (j=0; j < ns->sp_devs; j++) {
      nd = &(ns->sp_devlist[j]);

      if (nd->d_nblks == 0)		/* expected a label here */
	goto exit_err;

      nd->d_start = spblk_off;
      spblk_off += nd->d_nblks;
    }
    for (j=0; j < ns->sp_dldevs; j++) {
      nl = &(ns->sp_dldevlist[j]);

      if (nl->dl_nlocks == 0)		/* expected a label here */
	goto exit_err;

      nl->dl_start = spdl_off;
      spdl_off += nl->dl_nlocks;
    }
    
    ns->sp_len = spblk_off;
    ns->sp_dlocks = spdl_off;
    pblk_off += spblk_off;
    pdl_off += spdl_off;
  }
  np->pa_capacity = pblk_off;
  np->pa_dlocks = pdl_off;

  *pool = np;
  return (0);

 exit_err:

  if (ops_flag & AS_VERBOSE) {
    fprintf(stderr, "Ignoring fragment of pool %s (id: %"FMT_64"x).\n", 
	    np->pa_name, np->pa_pool_id);
  }

  /* free memory */

  for (i=0; i < np->pa_num_spools; i++) {
    ns = &(np->pa_spools[i]);

    if (ns->sp_devlist != NULL)
      free(ns->sp_devlist);
    if (ns->sp_dldevlist != NULL)
      free(ns->sp_dldevlist);
  }
  free(np->pa_spools);
  free(np);
	
  *pool = (pool_add_t *)NULL;
  return (0);
}




/******************************************************************************
*******************************************************************************
** NAME
**	remove_pools
**
** DESCRIPTION
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

int remove_pools(pool_list_t *list, int commfd, char *name)
{
  pool_status_t			*stats = NULL;
  int 		 	 i;

  /* if name==NULL (all), remove everything in list.pl_list[i] */ 

  /* for (i=0; i < list->pl_count; i++) { */
  for (i = list->pl_count - 1; i >= 0; i--) {
    stats = &(list->pl_list[i]);
    
    if (!name || !strcmp(stats->ps_pool_name, name)){
      
      if (stats->ps_in_use){
	fprintf(stderr, "%s in use.  Unmount before removing.\n", 
		stats->ps_pool_name);
	if (!name)
	  continue;

	exit(1);
      }
      if (ioctl(commfd, POOL_REMOVE, &stats->ps_pool_minor) < 0) {
	fprintf(stderr, "Unable to remove pool %s.\n", 
		stats->ps_pool_name);
	if (!name)
	  continue;

	exit(1);
      }
      printf("Removed %s.\n", stats->ps_pool_name);
      stats->ps_pool_name[0] = '0';

      if (name)
	return 0;
    }    
  }

  if (name){
    fprintf(stderr, "%s is not in use.\n", name);
    exit(1);
  }
  return 0;
}





/******************************************************************************
*******************************************************************************
**
** int printLabels()
**
** Description:
**   This routine prints out the labels for a storage pool device.
**
** Input(s):
**  *labels - A pointer to a storage pool labels structure.
**
** Returns:
**  0 if OK, 1 on error.
**
*******************************************************************************
******************************************************************************/

int printLabels(pool_add_t *pa)
{
  subpool_t	*sp;
  unsigned int 	 i;
  
  printf("Pool information:\n");
  printf("  Name:                     %s\n", pa->pa_name);
  printf("  ID:                       0x%"FMT_64"x\n", pa->pa_pool_id);
  printf("  Version:                  %d.%d.%d\n", 
	 pa->pa_version >> 16, (pa->pa_version >> 8) & 0xFF, 
	 pa->pa_version & 0xFF); 
  printf("  Capacity:                 %"FMT_64"d\n", pa->pa_capacity);
  printf("  Number of Dlocks:         %u\n", pa->pa_dlocks);
  printf("  Number of Subpools:       %u\n", pa->pa_num_spools);

  for (i = 0; i < pa->pa_num_spools; i++) {
    sp = &(pa->pa_spools[i]);
      
    printf("      Sub-pool %d:\n", i);
    printf("        Striping:                %d\n", sp->sp_stripe);
    printf("        Starting block:          %"FMT_64"u\n", sp->sp_start);
    printf("        Number of blocks:        %u\n", sp->sp_len);
    printf("        Data Devices:            %u\n", sp->sp_devs);
    printf("        Dlock Devices:           %u\n", sp->sp_dldevs);
  }
  printf("\n");

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** int clean_dir()
**
** Description:
**   This routine removes all pool device nodes which are not in use.
**
** Input(s):
**  
**   
**
** Returns:
**  0 if OK, 1 on error.
**
*******************************************************************************
******************************************************************************/

int clean_dir(int commfd)
{
  char		 bdir[1024];
  char		 path[1024];
  DIR		*dir;
  struct direct *file;
  int		 i=0;
  struct stat	 statbuf;
  pool_list_t  	 list;
 
  list.pl_count = 0;
  list.pl_list = NULL;

  pool_list(&list, commfd);

  strcpy(bdir, _PATH_DEV);

  strcat(bdir, "pool/");

  dir = opendir(bdir);
  rewinddir(dir);

  while ((file = readdir(dir)) != NULL) {

    strcpy(path, bdir);
    strcat(path, file->d_name);

    if (stat(path, &statbuf) >= 0){
      for (i=0; i < list.pl_count; i++) {
	if (list.pl_list[i].ps_pool_minor == (uint32_t)MINOR(statbuf.st_rdev))
	  break;
      }	
      if (i != list.pl_count)
	continue;
    }	
    unlink(path);
  }
  closedir(dir);

  return (0);
}



/******************************************************************************
*******************************************************************************
** NAME
**	getdevnum
**
** DESCRIPTION
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

int getdevnum(char *path, dev_t *dev)
{
  struct stat	devStats;

  if ((stat(path, &devStats)) < 0)
    return(-1);
  
  else
    *dev = devStats.st_rdev;
  return (0);
}



/******************************************************************************
*******************************************************************************
** NAME
**	pool_list
**
** DESCRIPTION
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

int pool_list(pool_list_t *inuse_list, int commfd)
{
  int 	     		statsize = 2;
  pool_status_t		*pool_stats = NULL;


  do {     /* don't know how many pools there are... */

    statsize <<= 1;
    if (pool_stats)
      free(pool_stats);

    pool_stats = (pool_status_t *)malloc(sizeof(pool_status_t) * statsize);

    if (pool_stats == NULL) {
      fprintf(stderr, "Unable to malloc for stats.\n");
      exit(1);
    }

    inuse_list->pl_count = statsize;
    inuse_list->pl_list = pool_stats;

    if (ioctl(commfd, POOL_LIST, inuse_list) < 0) {
      fprintf(stderr, "pool communications error: %s\n", strerror(errno));
      exit(1);
    }
  } while (statsize <= inuse_list->pl_count);

  return (0);
}


/******************************************************************************
*******************************************************************************
** NAME
**	skip_pool
**
** DESCRIPTION
**
** RETURN VALUE
**	1 if pool is found in list
**	0 if not
**
*******************************************************************************
******************************************************************************/

int skip_pool(pool_add_t **pool, pool_list_t *list)
{
  int i;

  for (i=0; i < list->pl_count; i++) {

    if (list->pl_list[i].ps_pool_id == (*pool)->pa_pool_id)
      return (1);
  }

  return (0);
}


/******************************************************************************
*******************************************************************************
** NAME
**	skip_md_devs
**
** DESCRIPTION
**
** RETURN VALUE		1 if dev is listed in /proc/mdstat
**			0 otherwise
*******************************************************************************
******************************************************************************/

int skip_md_devs(char *dev)
{
  FILE *fp;
  char linebuf[256];
  char *p;	

  if ((fp = fopen("/proc/mdstat", "r")) < 0) {
    fprintf(stderr, "/proc/mdstat not available\n");
    return (0);
  }

  while (fgets(linebuf, 256, fp) != NULL) {
    p = linebuf;
    
    while ((p = strstr(p, "sd")) != NULL) {

      if (!strncmp(dev, p, strlen(dev)))
	return (1);
     
      p++;
    }
  }
  return (0);
}


/******************************************************************************
*******************************************************************************
** NAME
**	check_name
**
** DESCRIPTION  Check if this pool name is currently in use in the kernel 
**	or if it's duplicated in the labels.
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

int check_name(pool_add_t **pool, pool_list_t *list)
{
  int i;

  for (i=0; i < list->pl_count; i++) {

    if (!strcmp(list->pl_list[i].ps_pool_name, (*pool)->pa_name))
      return (1);
  }


  for (i=0; i < lalist.l_cnt; i++) {
    
    if (lalist.l_lbs[i].pl_pool_id != (*pool)->pa_pool_id) {
      if (!strcmp(lalist.l_lbs[i].pl_pool_name, (*pool)->pa_name))
	return (1);
    }
  }

  return (0);
}


/******************************************************************************
*******************************************************************************
** NAME
**	free_mem
**
** DESCRIPTION
**
** RETURN VALUE
**
*******************************************************************************
******************************************************************************/

int free_mem(pool_add_t *pool)
{
  int i;
  subpool_t *ns;

  if (pool != NULL) {

    if (pool->pa_spools != NULL) {

      for (i=0; i < pool->pa_num_spools; i++) {
	ns = &(pool->pa_spools[i]);

	if (ns->sp_devlist != NULL)
	  free(ns->sp_devlist);
	if (ns->sp_dldevlist != NULL)
	  free(ns->sp_dldevlist);
      }
      free(pool->pa_spools);      
    }
    free(pool);
  }

  return (0);
}




