/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@lcse.umn.edu>
** 
**    Module name: gfsck.h
**
**    Description:
**	Filesystem checker.
**
*******************************************************************************
******************************************************************************/

#ifndef __GFSCK_H__
#define __GFSCK_H__

#define GFS_MAGIC_SB               (0x01011670) /* FIXME, from gfs_incore */

#include "GFS.cf"
#include "gfs_ondisk.h"


long long int llseek(int, long long int, int);
#define LLSEEK(x, y, z) llseek((x), (y), (z))


/*  Function Macros  */

#define type_zalloc(ptr, type, count) \
{ \
  if (((ptr) = (type *)malloc(sizeof(type) * (count))) != NULL) \
    memset((char *)(ptr), 0, sizeof(type) * (count)); \
  else \
  { \
    fprintf(stderr, "%s: unable to allocate memory on line %d.\n", prog_name, __LINE__); \
    exit(EXIT_FAILURE); \
  } \
}

#define type_alloc(ptr, type, count) \
{ \
  if (((ptr) = (type *)malloc(sizeof(type) * (count))) == NULL) \
  { \
    fprintf(stderr, "%s: unable to allocate memory on line %d.\n", prog_name, __LINE__); \
    exit(EXIT_FAILURE); \
  } \
}

#ifndef BTOBB
#define BTOBB(x) (((x) + GFS_BASIC_BLOCK - 1) / GFS_BASIC_BLOCK)
#endif  /*  BTOBB  */

#define do_lseek(fd, off) \
{ \
  if (LLSEEK((fd), (off), SEEK_SET) != (off)) \
  { \
    fprintf(stderr, "%s: bad seek: %s\n", prog_name, strerror(errno)); \
    exit(EXIT_FAILURE); \
  }\
}

#define do_write(fd, buff, len) \
{ \
  if (write((fd), (buff), (len)) != (len)) \
  { \
    fprintf(stderr, "%s: bad write: %s\n", prog_name, strerror(errno)); \
    exit(EXIT_FAILURE); \
  } \
}

#define do_read(fd, buff, len) \
{ \
  if (read((fd), (buff), (len)) != (len)) \
  { \
    fprintf(stderr, "%s: bad read: %s\n", prog_name, strerror(errno)); \
    exit(EXIT_FAILURE); \
  } \
}


#define BLACK			0
#define RED   			1

#define Left(i) 		(2*i)
#define Right(i) 		(2*i + 1)

#define CLEAR_DI		0x04
#define MARK_DI			0x03
#define MARK_BLK		0x01


#define	DI_MAGIC		0x0001
#define DI_RGRP			0x0002
#define DI_TYPE			0x0004
#define DI_OPEN			0x0008
#define DI_BLKS			0x0010
#define DI_DEPTH		0x0020
#define DI_SAMED		0x0040
#define DI_NODOT		0x0080
#define DI_BADDOT		0x0100
#define DI_ENTCNT		0x0200
#define DI_LINK			0x0400
#define DI_NOHOME		0x0800
#define DI_CONNECT		0x1000
#define DI_ADDR			0x2000

#define DE_HASH		    	0x0001
#define DE_RECLEN	     	0x0002
#define DE_NODI			0x0004



typedef struct node_s	       	/* Tree node */
{
  struct node_s	*left;
  struct node_s	*right;
  struct node_s	*p;
  uint64       	key;
  char		*data;
  int		color;
} node_t;


typedef struct filesystem_s
{
  int		fd;		/* Opened device file descriptor */
  char 		*fsdev;		/* Block device being checked */
  gfs_sb_t 	*sb;         	/* The Superblock  */
  gfs_rindex_t 	*ri;     	/* Array of resource indexes  */
  gfs_rgrp_t 	*rg;       	/* Array of resource group headers  */
  node_t	*dirs;		/* Tree of directory records */
  char		*bm;		/* Bitmap of allocated blocks */
  char 		*dibm;		/* Bitmap of dinodes only */
  node_t	*flink;		/* Tree of files with more than one link */
  uint64	blkcnt;		/* Number of blks allocated */
  uint64	dicount;	/* Number of dinodes found */
  node_t	*bad_blk;	/* Blocks used by two or more files */
  node_t	*bad_di;	/* Bad dinodes to be fixed */
  node_t	*bad_ent;	/* Bad dir entries to be fixed */
  node_t	*bad_lf;	/* Bad leaf blocks to be fixed */
} filesystem_t;


typedef struct dir_info_s
{
  uint64	addr;		/* dinode number */
  uint32      	di_ents;	/* numbers of entries from dinode */
  uint32	di_depth;	/* depth from dinode */
  uint32	di_nlink;	/* number of links from dinode */
  node_t       	*dblks;       	/* tree of data blks containing hash table */
  uint32	entcnt;		/* number of entries found */
  uint64	diblks;		/* number of blocks found for this dinode */
  uint32	subdirs;	/* count of subdirs */
  uint64	dot;		/* the inode number of '.' entry */
  uint64	dotdot;		/* the inode number of '..' entry */
  uint64	parent;		/* parent directory */
  uint32	same_depth;	/* count of leaves at same depth as dinode */
  int		mark;		/* indicates this dir is connected to root */
  uint32	errcode;	/* error condition for directory */
} dir_info_t;


typedef struct lnk_info_s	/* for flink list */
{
  uint64	addr;		/* file dinode number */
  uint32	nlink;		/* nlink count from dinode */
  uint32	found;		/* number of links found */
} lnk_info_t;


typedef struct blk_info_s	/* for bad_blk list */
{
  uint64	addr;		/* block number */
  uint64	dinode;		/* dinode using this block */
} blk_info_t;


typedef struct di_info_s     	/* for bad_di list */
{
  uint64	addr;
  uint32	code;
  uint64	other;
} di_info_t;


typedef struct ent_info_s	/* for bad_ent list */
{
  uint64	ino;
  uint64	addr;
  dir_info_t	*dir;
  int		islf;
  uint32	code;
} ent_info_t;


node_t *New_Treenode(uint64 key, char *data);

void RB_Insert(node_t **pT, node_t *x);
void Tree_Insert(node_t **pT, node_t *z);
void Inorder_Tree_Walk(node_t *x);
void Preorder_Tree_Walk(node_t *x);
void RB_Delete_Fixup(node_t **pT, node_t *x);
void Free_Tree(node_t *x);
void Free_Node(node_t *x);
void Left_Rotate(node_t **pT, node_t *x);
void Right_Rotate(node_t **pT, node_t *x);
node_t *RB_Delete(node_t **pT, node_t *z);
node_t *Tree_Delete(node_t **pT, node_t *z);
node_t *Tree_Search(node_t *x, uint64 k);
node_t *Tree_Search_GE(node_t *x, uint64 k);
node_t *Tree_Successor(node_t *x);
node_t *Tree_Predecessor(node_t *x);
node_t *Tree_Minimum(node_t *x);
node_t *Tree_Maximum(node_t *x);

filesystem_t *fsck_init(int argc, char *argv[]);
void decode_arguments(filesystem_t *fs, int argc, char *argv[]);
void print_usage(void);
void sync_disks(void);
void check_superblock(filesystem_t *fs);
void check_rindex(filesystem_t *fs);
void check_rgrps(filesystem_t *fs);
void check_dinodes(filesystem_t *fs);
void check_dirdinodes(filesystem_t *fs);
int check_difields(filesystem_t *fs, gfs_dinode_t *dip, uint32 rg, uint64 di);
dir_info_t *check_diblocks(filesystem_t *fs, gfs_dinode_t *dip);
void check_dirs(filesystem_t *fs);
void check_bitmaps(filesystem_t *fs);
void check_dentries(filesystem_t *fs, dir_info_t *dir, gfs_dinode_t *dip);
void get_ents(filesystem_t *fs, dir_info_t *dir, gfs_leaf_t *lf, 
	      char *data, uint64 addr);
uint64 get_leaves(filesystem_t *fs, dir_info_t *dir, uint64 *data, uint32 len);
uint32 dir_hash(char *filename, uint32 len);
void alloc_bitmaps(filesystem_t *fs);
int mark_bm(filesystem_t *fs, uint64 diaddr, uint64 blk, int action);
int map_bm(filesystem_t *fs, uint64 *blk, uint32 *rg, uint64 *off);
dir_info_t *keep_dir_blks(filesystem_t *fs, gfs_dinode_t *dip, uint64 blk);
void keep_dirs(filesystem_t *fs, gfs_dinode_t *dip);
uint64 bitmap_di_search(filesystem_t *fs, uint64 blk);
void check_dibitmap(filesystem_t *fs);
uint64 bitmap_di_next(filesystem_t *fs, uint64 blk);
void set_parent(filesystem_t *fs, uint64 entdi, dir_info_t *dir);
void record_flink(filesystem_t *fs, uint64 di);
void check_flink(filesystem_t *fs);


void record_badblk(filesystem_t *fs, uint64 blknum, uint64 diaddr);
void record_baddi(filesystem_t *fs, uint64 di, uint32 code, uint64 other);
void record_badent(filesystem_t *fs, uint64 ino, uint64 addr, dir_info_t *dir,
		   int islf, uint32 code);
void record_badlf(filesystem_t *fs, dir_info_t *dir, gfs_leaf_t *lf);

void repair_all(filesystem_t *fs);
void repair_bm(filesystem_t *fs, int rg, int byte, int bit, char old, 
	       char new);
int recover_bmdi(filesystem_t *fs, uint64 addr);
void erase_entry(filesystem_t *fs, uint64 addr, dir_info_t *dir, uint64 ino, 
		 int islf);
int repair_difield(filesystem_t *fs, uint64 addr, uint32 code, uint64 other);
void erase_dinode(filesystem_t *fs, uint64 addr);
int in_badblk(filesystem_t *fs, uint64 di);
int query_yn(char *str);
int fake_log(uint32 n);

void print_sb(filesystem_t *fs); 
void print_ri(filesystem_t *fs);
void print_rg(filesystem_t *fs);
void print_di(gfs_dinode_t *dip);
void print_bm(filesystem_t *fs);


#endif /* __GFSCK_H__ */






