/******************************************************************************
*******************************************************************************
**
**    Copyright 1995-1999 Regents of the University of Minnesota
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland  <teigland@lcse.umn.edu>
** 
**    Module name: ptool.c
**
**    Description:
**	Creates, prints, and syncs labels for the storage pool device driver.
**
*******************************************************************************
******************************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <time.h>
#include <sys/ioctl.h>
#include <linux/fs.h>
#include <sys/stat.h>

#include "global.h"
#include "ptool.h"


static int		 debug		= FALSE;
static int		 isQuiet	= FALSE;
static const char	*programName    = NULL;

static int 		 path_buf_count = 0;
static path_buf_t	*path_buf	= NULL;



int main(int argc, char *argv[])
{
  new_pool_t	*pool;
  
  programName = argv[0];

  decodeOptions(argc, argv);

  if ((pool = getPoolParams(argv[optind])) == NULL) {
    fprintf(stderr, "Error occurred while reading parameters.\n");
    exit(1);
  }

  writeLabels(pool);

  if (!isQuiet)
    printPool(pool);
 
  return(0);
}


/******************************************************************************
*******************************************************************************
**
** void decodeOptions()
**
** Description:
**   This routine steps through the command-line arguments, parsing out
**   recognzied options.
**
** NOTE: This routine alters a number of global variables.
**
** Input(s):
**    argc - Number of arguments on the command line.
**  **argv - Pointer to an array of strings containing command-line arguments.
**
*******************************************************************************
******************************************************************************/

void decodeOptions(int argc, char *const argv[])
{
  int	errFlag = 0;
  int	optchar;
  
  if (argc < 2) {
    printUsage(EXIT_FAILURE);
  }

  while((optchar = getopt(argc, argv, OPTION_STRING)) != EOF) {
    switch (optchar) {
    case 'd':
      debug = TRUE;
      break;
    case 'h':
      printUsage(EXIT_SUCCESS);
      break;
    case 'q':
      isQuiet = TRUE;
      break;
    case 'V':
      printVersion();
      exit(0);
      break;
    case '?':
    default:
      errFlag++;
    }
  }

  if (errFlag) {
    printUsage(EXIT_FAILURE);
  }

  return;
}


/******************************************************************************
*******************************************************************************
**
** void printUsage()
**
** Description:
**   This routine prints out the proper command line usage for this program.
**
** Input(s):
**  status - Flag determining what usage information will be printed and what
**           the exit status of the program will be after the information is
**           printed.
**
*******************************************************************************
******************************************************************************/

void printUsage(int status)
{
  printf("Usage: %s [-hqV] <parameter file>\n", 
	 programName);

  if (status != EXIT_SUCCESS)
    printf("Try `%s -h' for more information.\n", programName);

  if (status != EXIT_FAILURE) {
    printf("  file       Build storage pool labels according to file.\n");
    printf("  -h         Print this help, then exit.\n");
    printf("  -q         Quiet operation; program output is supressed.\n");
    printf("  -V         Print program version information, then exit.\n");
  }
  exit(status);
}


/******************************************************************************
*******************************************************************************
**
** void printVersion()
**
** Description:
**   This routine prints out version and compilation information for this
**   program.
**
*******************************************************************************
******************************************************************************/

void printVersion(void)
{
  printf("\
ptool - Storage Pool Configuration Tool\n\
Version: %d.%d.%d\n\
Built: %s\n\
Copyright (c) 1999 University of Minnesota\n",
	 (POOL_VERSION >> 16), (POOL_VERSION >> 8) & 0xff, 
	 POOL_VERSION & 0xff, __DATE__);

  return;
}


/******************************************************************************
*******************************************************************************
**
** int getDeviceID()
**
** Description:
**   This routine returns the device ID (i.e. major and minor numbers) of a
**   block special or character special file.  It also stores the mapping in
**   a global table for latter use.
**
** Input(s):
**  *specFile - A string containing the name of the block special or
**              character special file.
**  *deviceID - A pointer to storage for the device ID.
**
** Output(s):
**  *deviceID - Contains 0 if an error occurred, otherwise contains the 
**              device ID of specFile.
**
** Returns:
**   Upon successful completion a value of 0 is returned. 
**   Otherwise, a value of -1 is returned and errno is set to 
**   indicate the error.
**
*******************************************************************************
******************************************************************************/

int getDeviceID(const char *specFile, dev_t *deviceID)
{
  struct stat	devStats;

  if ((stat(specFile, &devStats)) < 0) {
    deviceID = 0;
    return(-1);
  }
  else if (!S_ISCHR(devStats.st_mode) && !S_ISBLK(devStats.st_mode)) {
    *deviceID = 0;
    return(-1);
  }
  else {
    *deviceID = devStats.st_rdev;
  }

  if (path_buf_count % 20 == 0) {
    path_buf = (path_buf_t*)realloc(path_buf, 
				    (path_buf_count + 20) * sizeof(*path_buf));
    if (path_buf == NULL) {
      fprintf(stderr, "Failed to extend device path name buffer: %s\n",
	      strerror(errno));
      return(-1);
    }
  }

  path_buf[path_buf_count].dev = *deviceID;
  strncpy(path_buf[path_buf_count++].path, specFile, POOL_NAME_SIZE);

  return(0);
}


/******************************************************************************
*******************************************************************************
**
** new_pool_t *getPoolParams()
**
** Description:
**   This routine builds pool information from an input file.
**
** Input(s):
**  *paramFile - Name of the parameter file containing the storage pool build
**               parameters.
**
** Returns:
**  A pointer to a new_pool structure, NULL on error.
**
*******************************************************************************
******************************************************************************/

new_pool_t *getPoolParams(const char *paramFile)
{
  FILE		*fp = NULL;
  char	        lineBuf[LINE_MAX];
  char	        keyword[24];
  char	        arg1[24], arg2[24], arg3[24], arg4[24], arg5[24];
  char	        errString[256];

  uint	      	i, j, lineCnt;
  int	      	s;
  int		spid, pdid;
  int	      	stripe, spdevs;
  uint64	blks;
  uint		dlocks, size;
  dev_t	     	deviceID = 0;

  new_pool_t	*np = NULL;
  new_spool_t	*ns = NULL;
  new_dev_t	*nd = NULL;


  /* Open the parameter file */
  
  if ((fp = fopen(paramFile, "r")) == NULL) {
    fprintf(stderr, "Error opening the parameter file \"%s\"\n", paramFile);
    perror("Reason");
    exit(1);
  }


  np = (new_pool_t *)malloc(sizeof(new_pool_t));
  memset(np, 0, sizeof(new_pool_t));

  /* Read the parameter file */
  
  lineCnt = 0;

  while (fgets(lineBuf, LINE_MAX, fp) != NULL) {

    lineCnt++;

    if (lineBuf[0] == '#')
      continue;
      
    if (lineBuf[0] == '\n')
      continue;
      
    /* Scan for a keyword and up to five arguments */
      
    sscanf(lineBuf, "%s %s %s %s %s %s", keyword, arg1,arg2,arg3,arg4,arg5);

      
    /* Check for various recognized keywords */

    if (!strcmp(keyword,"poolname")) {

      strncpy(np->np_name, arg1, POOL_NAME_SIZE);
      if (strlen(arg1) <= 0) {
	parseError(paramFile, lineCnt, lineBuf, "poolname error");
	exit(1);
      }
    }

    else if (!strcmp(keyword,"subpools")) {

      if ((np->np_num_spools = atoi(arg1)) < 1) {
	parseError(paramFile, lineCnt, lineBuf, "subpools error");
	exit(1);
      }

      size = np->np_num_spools * sizeof(new_spool_t);

      np->np_spools = (new_spool_t *)malloc(size);
	
      memset(np->np_spools, 0, size);
    }

    else if (!strcmp(keyword,"subpool")) {

      spid = atoi(arg1);
      stripe = atoi(arg2);
      spdevs = atoi(arg3);
	  
      if (np->np_num_spools == 0) {
	parseError(paramFile, lineCnt, lineBuf, 
		   "\"subpools\" keyword must come before \"subpool\"");
	exit(1);
      }

      if (spid >= np->np_num_spools) {
	parseError(paramFile, lineCnt, lineBuf, "bad subpool id");
	exit(1);
      }
	  
      if (spdevs < 1) {
	parseError(paramFile, lineCnt, lineBuf, 
		   "subpool must have at least 1 device");
	exit(1);
      }

      ns = &(np->np_spools[spid]);
      ns->ns_stripe = stripe;
      ns->ns_num_devs = spdevs;

      size = ns->ns_num_devs * sizeof(new_dev_t);

      ns->ns_devs = (new_dev_t *)malloc(size);
	  
      memset(ns->ns_devs, 0, size);
    }

    else if (!strcmp(keyword,"pooldevice")) {
	
      spid = atoi(arg1);
      pdid = atoi(arg2);
      		  
      if ((s = getDeviceID(arg3, &deviceID)) < 0) {
	strcpy(errString, strerror(errno));
	parseError(paramFile, lineCnt, lineBuf, errString);
	exit(1);
      }
      
      if (!strcmp(arg4, "LOCKS-ONLY"))
	blks = 0;
      else
	blks = atoll(arg4);

      dlocks = atoi(arg5);

      if (blks != 0)
	ns->ns_blk_devs++;
      if (dlocks != 0)
	ns->ns_dlk_devs++;

      if (spid >= np->np_num_spools) {
	parseError(paramFile, lineCnt, lineBuf, "bad subpool id");
	exit(1);
      }
	
      ns = &(np->np_spools[spid]);

      if (pdid >= ns->ns_num_devs) {
	parseError(paramFile, lineCnt, lineBuf, "bad device id");
	exit(1);
      }

      nd = &(ns->ns_devs[pdid]);
      nd->nd_dev = deviceID;
      nd->nd_blocks = blks;
      nd->nd_dlocks = dlocks;
	
      if (blks == 0)
	nd->nd_dlonly = 1;
      else
	nd->nd_blocks -= 16;	/* 16 blocks offset for label */


      /* make all partions multiples of 512Kb in size for alignment */

      /* nd->nd_blocks &= 0xFFFFFFFFFFFFFC00ULL; */

    }

      
    bzero(lineBuf, 256);
    bzero(keyword, 24);
    bzero(arg1, 24);
    bzero(arg2, 24);
    bzero(arg3, 24);
    bzero(arg4, 24);
    bzero(arg5, 24);
  }
  
  fclose(fp);
  

  /* verify all subpools and pooldevices are defined */

  for (i=0; i < np->np_num_spools; i++) {
    ns = &(np->np_spools[i]);
    if (ns->ns_num_devs == 0) {
      fprintf(stderr, "Error.  subpool %d has no devices.\n", i);
      exit(1);
    }
    for (j=0; j < ns->ns_num_devs; j++) {
      nd = &(ns->ns_devs[j]);
      if ((nd->nd_blocks == 0) && (nd->nd_dlonly == 0)) {
	fprintf(stderr, "Error. subpool %d, device %d.\n", i, j);
	exit(1);
      }
    }
  }

  return(np);
}


/******************************************************************************
*******************************************************************************
**
** void parseError()
**
** Description:
**   This routine prints out an error message to standard error listing the
**   file in which a parsing error was found, the line number in which it was
**   found, and the actual line containing the error.
**
** Input(s):
**  *fileName - String containing the name of the file in which the parse
**              error was found.
**   lineNum  - Line number on which the parse error was found
**  *lineBuf  - String containing the line on which the parse error was found
**  *errMsg   - String containing an error message informing the user what
**              the error was.
**
*******************************************************************************
******************************************************************************/

void parseError(const char *fileName, int lineNum, char *lineBuf, char *errMsg)
{
  fprintf(stderr, "\"%s\", line %d: error: %s\n%s", 
	  fileName, lineNum, errMsg, lineBuf);

  return;
}


/******************************************************************************
*******************************************************************************
**
** static char *get_path()
**
** Description:
**   This routine takes a device structure as an input and returns the path
**   name corresponding to that device if one is found in a global table.
**
** Input(s):
**   dev - Device structure to attempt to match to the global table.
**
** Returns:
**   A pointer to the device name if OK, NULL on error.
**
*******************************************************************************
******************************************************************************/

static char *get_path(dev_t dev)
{
  int i;

  for (i = 0; i < path_buf_count; i++) {
    if (dev == path_buf[i].dev)
      return(path_buf[i].path);    
  }

  return(NULL);
}



/******************************************************************************
*******************************************************************************
**
** int printPool()
**
** Description:
**   This routine prints out the definition of a storage pool device.
**
** Input(s):
**  *np - the new_pool_t structure created from the parameter file.
**
** Returns:
**  0 if OK, 1 on error.
**
*******************************************************************************
******************************************************************************/

int printPool(new_pool_t *np)
{
  uint		i, j;
  new_spool_t	*ns;
  new_dev_t	*nd;
  
  printf("\nPool information:\n");
  printf("  Pool name:		%s\n", np->np_name);
  printf("  Number of subpools:	%u\n", np->np_num_spools);

  for (i=0; i < np->np_num_spools; i++) {
    ns = &(np->np_spools[i]);
    printf("  Subpool %d:\n", i);
    printf("    Number of devices:  %u\n", ns->ns_num_devs);
    printf("    Stripe size:        %u\n", ns->ns_stripe);
    
    for (j=0; j < ns->ns_num_devs; j++) {
      nd = &(ns->ns_devs[j]);
      printf("    Device %d:           %d,%d\n", j,
	     (int)MAJOR(nd->nd_dev), (int)MINOR(nd->nd_dev));
      printf("      Blocks:           %u\n", nd->nd_blocks);
      printf("      Dlocks:           %u\n", nd->nd_dlocks);
      /*printf("      Dlocks only:      %s\n", nd->nd_dlonly ? "YES" : "NO");*/
    }
  }
  printf("\n");
  return (0);
}


/******************************************************************************
*******************************************************************************
**
** int writeLabels()
**
** Description:
**   This routine writes out a storage pool labels structure to a device.
**
** Input(s):
**
** Returns:
**   0 if OK, 1 on error.
**
*******************************************************************************
******************************************************************************/

int writeLabels(new_pool_t *np)
{

  char		       	*buf;
  int			len = SIZEOF_POOL_LABEL;
  int	       	 	fd;
  uint		      	i,j;
  char		       	*fname;
  time_t	      	seed;
  pool_label_t		label;
  new_spool_t		*ns;
  new_dev_t		*nd;
  int			devid_cnt = 0; 
  int			dlid_cnt = 0;

  /* Create a unique id for this pool */

  memset(&label, 0, sizeof(pool_label_t));
  buf = (char *)malloc(len);
  memset(buf, 0, len);

  time(&seed);

  label.pl_magic = POOL_MAGIC;
  label.pl_pool_id=(((uint64)np->np_spools[0].ns_devs[0].nd_dev) << 32) + seed;
  label.pl_version = POOL_VERSION;
  label.pl_subpools = np->np_num_spools;
  strncpy(label.pl_pool_name, np->np_name, POOL_NAME_SIZE);

  if (!isQuiet)
    printf("\nPool ID: 0x%"FMT_64"x\n", label.pl_pool_id);


  for (i=0; i < np->np_num_spools; i++) {
    
    ns = &(np->np_spools[i]);

    devid_cnt = dlid_cnt = 0;

    label.pl_sp_id = i;
    label.pl_sp_devs = ns->ns_blk_devs;
    label.pl_sp_dldevs = ns->ns_dlk_devs;
    label.pl_striping = ns->ns_stripe;
    
    for (j=0; j < ns->ns_num_devs; j++) {
      
      nd = &(ns->ns_devs[j]);

      label.pl_blocks = nd->nd_blocks;
      label.pl_dlocks = nd->nd_dlocks;

      if (label.pl_blocks)
	label.pl_sp_devid = devid_cnt++;

      if (label.pl_dlocks)
	label.pl_sp_dlid = dlid_cnt++;
      
      pool_label_out(&label, buf);

      fname = get_path(nd->nd_dev);

      if ((fd = open(fname, O_WRONLY)) < 0) {
	fprintf(stderr, "%s open: %s.\n", fname, strerror(errno));
	exit(1);
      }

      if (!isQuiet) {
	printf("Writing %s label...\n", fname);
      }

      if (write(fd, buf, len) < len) {
	fprintf(stderr, "%s: write: %s\n", fname, strerror(errno));
	exit(1);
      }
      
      close(fd);

      memset(buf, 0, len);
    }
  }

  free(buf);
  return(0);
}





















